(** First steps of Wainer's hierarchy : 
    From f_0 to f_\omega 


Pierre Castéran, 
 *)

Require Import ArithRing Omega  Max.

(** ** Definitions *)

Definition nat_fun := nat -> nat.

Definition id_nat (n:nat) := n.

Fixpoint iterate {A:Type}(f : A -> A) (n: nat)(x:A) :=
  match n with
    | 0 => x
    | S p => f (iterate  f p x)
  end.

(** Functionals associated with successor and limit ordinals *)

Definition Fsuc F n :=  iterate F (S n) n.

Definition Flim (seq : nat -> nat_fun) n := seq n n.


(** First steps of the hierarchy *)

Fixpoint f_ (i: nat) : nat_fun :=
  match i with
      0 => S
    | S j => Fsuc (f_ j)
  end.

Definition f_omega  := Flim f_.


(** *** examples 
*)


Compute f_ 0 41.

Compute f_ 1 21.

Compute f_ 2 5.

Compute f_omega 2.

 Lemma f1_eqn : forall n, f_ 1 n = S (2 * n).
  Proof. 
    intro n.  cbn.
    generalize n at 1 3 as p. 
    induction p.
    - cbn.  ring. 
    - cbn; intros. rewrite IHp. trivial. 
  Qed.


(** ** Abstract properties on nat_fun *)

Definition strict_mono f := forall n p,  n < p -> f n < f p.


Definition dominates_from n g f  := forall p, n <= p -> f p < g p.

Definition dominates_strong g f  := {i : nat | dominates_from i g f}.

Definition dominates g f := exists i : nat, dominates_from i g f .

Definition fun_le f g  := forall n:nat,  f n <=  g n.

Infix ">>" := dominates (at level 60).

Infix ">>s" := dominates_strong (at level 60).

Infix "<<=" := fun_le (at level 60).

(**  Notice that (S <<= f) means that x < f x for every x *)



Lemma fun_le_trans f g h : f <<= g ->  g <<= h -> f <<= h.
Proof. 
   intros; red.
   intro n; transitivity (g n); auto.
Qed.


Lemma mono_le f (Hf : strict_mono f) :  forall n, n <= f n.
Proof.
  induction n.
  - auto with arith. 
  -  apply le_lt_trans with (f n);  auto with arith.  
Qed. 

Lemma mono_injective f (Hf : strict_mono f) :  forall n p , f n = f p -> n = p.
Proof.
  intros n p H.
  destruct (Nat.lt_total n p).
  - specialize (Hf _ _ H0);   rewrite H in Hf;  destruct (lt_irrefl _ Hf).
  -  destruct H0; trivial. 
     +  specialize (Hf _ _ H0); rewrite H in Hf; destruct (lt_irrefl _ Hf).
Qed. 

Lemma mono_weak f (H: strict_mono f) :
  forall n p, n <= p -> f n <= f p.
Proof.
  induction 1.
  -  left.
  -   apply Nat.lt_le_incl; apply H; auto with arith. 
Qed.




Lemma dominates_from_max :
  forall f g h i j, dominates_from  i g f  ->
                    dominates_from j  h g  ->
                    dominates_from  (max i j) h f .
Proof.
  intros f g h i j H H0 k Hk;  transitivity (g k).
  +  apply H; eauto with arith.
  + apply H0; eauto with arith.
Qed.


Lemma dominates_trans : forall f g h,  dominates g f ->
                                       dominates h g ->
                                       dominates h f.
Proof.
  intros f g h [i Hi] [j Hj]; exists (max i j);
  eapply dominates_from_max with g; eauto.
Qed.

Lemma dominates_trans_strong : forall f g h,
                                 dominates_strong g f ->
                                 dominates_strong h g ->
                                 dominates_strong h f.
Proof.
  intros f g h [i Hi] [j Hj];exists (max i j);
  eapply dominates_from_max with g; eauto.
Qed.

(** ** Abstract properties of iterate *)

(** *** fixpoint rewriting *)

Lemma iterate_S_eqn {A:Type}(f : A -> A) (n: nat)(x:A):
  iterate f (S n) x = f (iterate f n x).
Proof. reflexivity.  Qed.


(** *** Equation associated with tail recursion *)

Lemma iterate_S_eqn2 {A:Type}(f : A -> A) (n: nat)(x:A):
  iterate f (S n) x =  (iterate f n (f x)).
Proof.
  induction n. 
  - reflexivity.
  - rewrite (iterate_S_eqn f (S n)), IHn;  reflexivity. 
Qed.


Lemma iterate_le f (Hf : strict_mono f) :
  forall x y, x <= y -> forall z, iterate f x z <= iterate f y z.
Proof.
  induction 1.
  - trivial.   
  - intros; rewrite iterate_S_eqn.
     transitivity (iterate f m z); auto. 
     apply mono_le; auto. 
Qed.


Lemma iterate_lt  f (Hf : strict_mono f) (Hf' :  S  <<= f):
  forall x y, x < y -> forall z, iterate f x z < iterate f y z.
Proof.
  induction 1.
  - intros.  rewrite iterate_S_eqn; auto. 
  - intros; rewrite iterate_S_eqn.
     transitivity (iterate f m z); auto. 
Qed.

Lemma iterate_mono f (Hf : strict_mono f) (Hf' :  S  <<= f):
  forall n, strict_mono (iterate f n).
Proof.
  induction n.
  -  red; intro i; cbn;auto.
  - cbn; intros i j H;  apply Hf;auto.
Qed.

Lemma iterate_ge : forall f , S  <<= f -> 
                               forall  j n, j <= iterate f n j.
Proof.
  induction n.
  cbn.
  auto with arith.
  apply Nat.lt_le_incl.
  rewrite iterate_S_eqn.
  apply le_lt_trans with (iterate f n j).
  auto.
  apply H.
Qed.  


Lemma iterate_dom_prop :
  forall f g i (Hgt : S <<= f)
         (Hm : strict_mono f) (Hm': strict_mono g),
    dominates_from i g f ->
    forall k, 0 < k ->
              dominates_from i (iterate g k) (iterate f k).
Proof.
  induction k.
  -   intro H0; inversion H0.
  -   destruct k.
   +   simpl;  intros _ l Hl;  apply H; auto.
   +   intros _ l Hl; repeat rewrite iterate_S_eqn.
       transitivity (g (f (iterate f k l))).
   *   apply H; transitivity (f l).
    {   transitivity l;  auto. 
        apply Nat.lt_le_incl.
        eapply Hgt; auto.
    }
    apply mono_weak;  auto.
    eapply iterate_ge;  auto. 
   *  apply Hm';  assert (0 < S k) by auto with arith.
      apply IHk in H0; specialize (H0 l).
      repeat rewrite iterate_S_eqn in H0; auto.   
Qed. 

Lemma dominates_from_le  i j g f : i <= j ->
                                   dominates_from i g f -> 
                                   dominates_from j g f .
Proof. 
  induction 1; auto.
  intros H0 x H1; apply IHle; auto.
  auto with arith.
Qed. 


Section On_Fsuc.
Variable f : nat_fun.
Hypotheses (Hmono: strict_mono f)
           (Hgt :  S <<= f ).

Lemma Fsuc_mono : strict_mono (Fsuc f).
Proof.
  intros n p H ; unfold Fsuc.
   repeat rewrite iterate_S_eqn.
   apply Hmono.
   transitivity (iterate f n p).
   -  now apply iterate_mono.   
   -  now apply iterate_lt. 
Qed. 


Lemma Fsuc_ge_S : S  <<= Fsuc f.
Proof. 
 intro n; unfold Fsuc; change n at 1 with (iterate f 0 n). 
 apply iterate_lt; auto with arith. 
Qed. 


Lemma Fsuc_le : f <<= Fsuc f.
Proof. 
  unfold Fsuc; intro n.
  change (f n) with (iterate f 1 n).
  apply iterate_le; auto with arith. 
Qed.

Lemma Fsuc_dom : dominates_from 1 (Fsuc f) f.
Proof. 
 unfold Fsuc; intros n H.
 change (f n) with (iterate f 1 n).
 apply iterate_lt; auto with arith. 
Qed. 

End On_Fsuc.


(** Let's now study the properties of all the (f_ i) s *)
(** used in tactics like  "rewrite (F_S_rw i) "  *)
  
Lemma F_S_rw : forall i , f_ (S i)  = Fsuc (f_ i) .
Proof.   reflexivity. Qed.


(** Like in Ketonen-Solovay, a simultaneous proof of four properties 
    of the (f_ i) s *)

Section Properties.

    Let P (i:nat) := strict_mono (f_ i) /\
                   S  <<= f_ i  /\
                   (f_ i) <<= (f_ (S i)) /\
                   dominates_from 1 (f_ (S i)) (f_ i).
    
Section The_induction.

  (** Base step : (sequential) proof of (P 0) *)
      
  Lemma mono_0 : strict_mono (f_ 0).
  Proof. intros n p H; simpl f_; auto with arith. Qed. 

  Lemma f_0_ge_S : S  <<= (f_ 0).
  Proof. intros n ; simpl f_; auto with arith. Qed.

 
  

  Lemma f_1_0_dom : dominates_from 1 (f_ 1) (f_ 0).
  Proof.
    red;intros.
    simpl (f_ 0).
    rewrite f1_eqn.
    omega.
  Qed.
  
  Hint Resolve f_1_0_dom mono_0 f_0_ge_S.

  Lemma f_1_0_ge :  (f_ 0) <<= (f_ 1).
  Proof.
    intro n; destruct n.
    - compute; auto with arith. 
    - eauto with arith. 
  Qed. 

  Hint Resolve  f_1_0_ge.

  Lemma P0 : P 0.  
  Proof. 
    repeat split; auto. 
  Qed.   

  (* inductive step *)

  Variable i : nat.
  Hypothesis Hi : P i.

  Ltac hdecomp := unfold P in Hi;decompose [and] Hi.

   Lemma mono_Si : strict_mono (f_ (S i)).
   Proof. hdecomp; simpl. now apply Fsuc_mono.   Qed.

   Hint Resolve mono_Si.
   
   Lemma fSi_ge_S : S  <<= f_ (S i).
    Proof. hdecomp; simpl. now apply Fsuc_ge_S.   Qed.

     Hint Resolve fSi_ge_S.

   Lemma le_Si : f_ (S i) <<=  f_ (S (S i)).
   Proof. hdecomp.
          rewrite (F_S_rw (S i)); auto.
          apply Fsuc_le;auto. 
   Qed.


   Lemma dom_Si : dominates_from 1 (f_ (S (S i))) (f_ (S i)).
   Proof.
     hdecomp;  rewrite (F_S_rw (S i));  apply Fsuc_dom;auto.
   Qed. 

   Lemma PSi : P (S i).
   Proof. 
     unfold P; repeat split.
     - apply mono_Si.
     - apply fSi_ge_S. 
     - apply le_Si.
     -  apply dom_Si. 
 Qed. 

End The_induction.

About PSi.


Lemma AllP: forall i:nat, P i.
Proof. 
 induction i. 
 -  apply P0.
 -  apply PSi;auto. 
Qed.


End Properties.


(** Introduces in the context the four components of (P n) *)
Ltac get_properties  n := 
  let H := fresh "H" in
  generalize (AllP n); intro H; decompose [and] H.


Example mono_42 : strict_mono (f_ 42).
Proof.
  now get_properties 42. 
Qed. 


Lemma f_mono_weak : forall n p, n <= p -> f_ n <<=  f_ p.
Proof. 
  induction 1.
   - red;auto. 
   -  apply fun_le_trans with (f_ m);auto. 
      rewrite F_S_rw .
      apply Fsuc_le.
      get_properties m;auto. 
Qed. 

Lemma f_lt_lt : forall i j, i < j ->
                            dominates_from 1 (f_ j) (f_ i).

Proof. 
  induction 1.
  -   get_properties i; auto. 
  -   intros j Hj.  
      transitivity (f_ m j);auto. 
      get_properties m; auto. 
Qed. 


(** Let's now play with f_0, f_1, etc . *)

Lemma LF0' : forall i j, iterate (f_ 0) i j = (i + j).
Proof. 
  induction i.
  -   reflexivity.
  -  intro; cbn; auto. 
Qed.

(** Tactics for undiagonalization *)

(** performs an induction only on the occ1-th occurrence of n *)

Tactic Notation "undiag1" constr(n) integer(occ1) :=
  let n' := fresh "n" in
  generalize n at occ1 ; intro n'; induction n'.

(** performs an induction only on the occ1-th and occ2_th occurrences of n *)

Tactic Notation "undiag2" constr(n) integer(occ1) integer(occ2) :=
  let n' := fresh "n" in
  generalize n at occ1 occ2; intro n'; induction n'.


Lemma LF1 : forall n,  f_ 1 n = S (2 * n).
Proof.
  intro;  cbn.    
  undiag2 n 1 3.  
  - cbn; ring. 
  - cbn;  f_equal; auto. 
Qed. 



Fixpoint exp2 (n:nat) : nat :=
  match n with
      0 => 1
    | S i => 2 * exp2 i
  end.


Lemma exp2_ge_S : S  <<= exp2.
Proof. 
  red; induction n. 
  - cbn; auto with arith. 
  - cbn;  omega. 
Qed. 


Lemma exp2_positive : forall i, 0 < exp2 i.
Proof.
 intro i; apply le_lt_trans with i; auto with arith. 
 apply exp2_ge_S.
Qed. 

Lemma exp2_mono : strict_mono exp2.
Proof.
  red; induction 1.
  - cbn.
      generalize (exp2_positive n); generalize (exp2 n);intros;omega.
  -  cbn;
      generalize IHle, (exp2_positive m); generalize (exp2 n), (exp2 m);
      intros;omega.
Qed. 



Lemma LF1' : forall i j, exp2 i * j < iterate (f_ 1) (S i) j .
Proof.
  induction i.
  - cbn.  intros; cbn. ring_simplify.
    rewrite LF0'.
    omega.
  - intros.
    simpl exp2.  ring_simplify.
    simpl (2+i).
    rewrite iterate_S_eqn.
    rewrite LF1.
    specialize (IHi j).
    rewrite mult_assoc_reverse.
    generalize IHi; generalize  (exp2 i * j);
    generalize  (iterate (f_ 1) (S i) j).  
    intros. change (N.to_nat 2) with 2; omega.
Qed.



Lemma LF2_0 : dominates_from 0 (f_ 2) (fun i => exp2 i * i).
Proof.
  red; intros. rewrite F_S_rw. unfold Fsuc.
  clear H. 
  undiag2 p 1 3.
  -  cbn. ring_simplify.  rewrite LF0'. omega.
  -  simpl exp2.  ring_simplify.
     simpl (2+n); rewrite iterate_S_eqn.
    rewrite LF1;    rewrite mult_assoc_reverse.
    change (N.to_nat 2) with 2; omega.
Qed.

Lemma LF2 :  f_ 2  >>s  (fun i => exp2 i * i).
Proof.
  exists 0; apply LF2_0.
Qed.


Lemma minoration_f_2_2 :  dominates_from 2 (f_ 2) exp2.
Proof.
  change 2 at 1 with (max 2 0).
  eapply dominates_from_max.
  2 : apply LF2_0.
  -   intros i;  induction 1.
      +   compute;   auto with arith.
      +   simpl. ring_simplify.  
        rewrite <- mult_n_Sm.
        rewrite mult_assoc_reverse.
        generalize IHle; generalize  (exp2 m * m); generalize (exp2 m).
        intros. change (N.to_nat 2) with 2. omega.
Qed. 


Lemma minoration_f_2  : f_ 2 >>s exp2.
Proof.
  exists 2; apply minoration_f_2_2.
Qed.


(**  2 ^ 2 ... ^2 ^ n *)
Fixpoint bell_tower (n top: nat) :=
  match n with
      0 => top
    | S p => exp2 (bell_tower p top )
  end.


Lemma L2 : forall n t, iterate exp2 n t = bell_tower n t.
  induction n;simpl; auto.
Qed.


Lemma L3 : forall x,
              dominates_from x (f_ 2) exp2 ->
              forall k,
                0 < k ->
                dominates_from x 
                               (iterate (f_ 2)  k)
                               (iterate exp2 k).
Proof.
  intros x Hx;  intros;   apply iterate_dom_prop; auto.
  -   red;intros.
      apply exp2_ge_S.
  -   apply exp2_mono.
  -   now get_properties 2.
Qed.



Section S4.
Variables  (n:nat)(f g : nat_fun)
      (Hf': strict_mono f)
      (Hg: S <<= g)
      (H : forall p, n<=p -> f p <= g p).


Remark R2: forall p x,  n <= x ->  x <= iterate g p x.
Proof. 
  induction p.      
  simpl; auto.  
  intros; rewrite iterate_S_eqn.
  apply le_trans with (iterate g p x); auto with arith.
Qed.

Remark R22: forall p ,  p <= iterate g p 1.
Proof. 
  induction p.      
  simpl; auto.  
  intros; rewrite iterate_S_eqn.
  apply le_trans with (S (iterate g p 1)); auto with arith.
Qed.

Remark R23: forall p ,  p < iterate g p 1.
Proof. 
  induction p.      
  simpl; auto.  
  intros; rewrite iterate_S_eqn.
  apply lt_le_trans with (S (iterate g p 1)); auto with arith.
Qed.

Lemma R3 : forall i x,  n <= x -> iterate f i x <= iterate g i x.
Proof. 
  induction i. 
  -  simpl; auto with arith. 
  - intro x; repeat rewrite iterate_S_eqn; intros.
    transitivity (f (iterate g i x)).
   +  destruct (le_lt_or_eq _ _  (IHi _ H0)).
     *  apply Nat.lt_le_incl; auto. 
     *  rewrite H1;left;auto. 
  +   apply H; transitivity x;auto. 
     apply R2;  auto. 
Qed. 

End S4.

(* Simplifier cette preuve ! *)

Lemma LF3_2  : dominates_from 2  (f_ 3) (fun  n => iterate exp2 n n).
Proof.  
  red; assert  (dominates_from 0 (f_ 2) (fun n => n)). 
  { red;   intros;  get_properties 2; auto.  }
  red; intros; generalize ( minoration_f_2_2); intro.
  generalize (L3 2 H1);  intros.
  specialize (H2 2).
  assert (0 < 2) by   eauto with arith.
  specialize (H2 H3).
  rewrite F_S_rw.
  apply le_lt_trans with ( iterate (f_ 2)  p p).
  -   generalize  (iterate_dom_prop  exp2 (f_ 2) p);  intro.
      assert (S <<= f_ 2) by (now get_properties 2). 
      apply R3 with p.  
      +   apply exp2_mono;  auto.                         
      + auto.
      + destruct 1.
        *  specialize (H1 p);   apply Nat.lt_le_incl; auto with arith. 
        *  apply Nat.lt_le_incl;  apply H1;  eauto with arith. 
       +            eauto with arith. 
  - unfold Fsuc;  rewrite iterate_S_eqn;  get_properties 2; auto. 
Qed. 



Lemma LF3 : f_ 3 >>s (fun  n => iterate exp2 n n).
Proof.   
 exists 2; apply LF3_2.
Defined.


Lemma LF3' : bell_tower 4 4  <  f_ 3 4.
Proof.
  specialize  (LF3_2 4); rewrite <- L2; intro; apply H; auto with arith. 
Qed. 

(** ** On F omega *)

Remark f_omega_rw : forall n, f_omega n = f_ n n.
Proof.
  unfold f_omega, Flim; trivial. 
Qed.


Fact f_omega_S (n:nat) : f_omega (S n) > f_ n (S n).
Proof.
  rewrite f_omega_rw;  get_properties n.
  apply H4;  auto with arith .
Qed.

Lemma f_omega_gt (n : nat): forall p : nat, n < p ->
                                            f_omega p > f_ n p.
Proof .
  induction 1.
 -   apply f_omega_S.
 -  rewrite f_omega_rw;  get_properties m.
    apply le_lt_trans with  (f_ m (S m)).
    +   apply f_mono_weak; auto with arith. 
    +  apply H5;  auto with arith. 
Qed. 

Lemma f_omega_depasse_tout : forall n, dominates_from (S n) f_omega (f_ n).
Proof.
  red; intros;  apply f_omega_gt;   auto .
Qed .

Lemma f_omega_mono : strict_mono f_omega.
Proof .
  intros n p H;  repeat rewrite f_omega_rw.
  get_properties n;  get_properties p.
  transitivity (f_ n p); auto with arith. 
  apply f_lt_lt;  eauto with arith. 
Qed. 

Lemma f_omega_ge_S :  S <<= f_omega.
Proof.
  intros n;rewrite f_omega_rw;  get_properties n; auto .
Qed .




Lemma f_i_weak_mono : forall i n p, n <= p -> f_ i n <= f_ i p.
Proof. 
  intros; get_properties i.
   destruct (Lt.le_lt_or_eq _ _ H).
   apply Nat.lt_le_incl; auto. 
   subst; auto. 
Qed. 

Lemma iterate_mono_1 (f g: nat_fun) (k:nat) (Hf: strict_mono f)(Hf' : S <<= f)
      (H : forall n, k <= n -> f n <= g n) :
  forall i n, k <= n -> iterate f i n <= iterate g i n.
Proof. 
  induction i. 
  simpl. 
 auto with arith. 
 intros; repeat rewrite iterate_S_eqn. 
   transitivity (f (iterate g i n)); auto. 
   apply mono_weak; auto. 
   apply H.
   transitivity (iterate f i n).
   2: auto.
   induction H0.
   clear IHi.   
    induction i. 
  simpl. 
  auto with arith. 
  rewrite iterate_S_eqn. 
  transitivity (iterate f i k).
    auto. 
    apply Nat.lt_le_incl.
  auto. 
  transitivity (S m).
 auto with arith. 
  apply iterate_ge. 
  auto. 
Qed. 

