Require Import Omega.
Require Import RelationClasses.
Require Import Relations Relation_Definitions.
Require Import  Ensembles  Epsilon .
Require Import Sets.Image.
Require Import Sets.Infinite_sets.


Require Import inductive_wqo.

(* Context {A: Type}. *)
Class Wqo {A:Type} (R : relation A) : Prop :=
  {
    well_order : forall {u:nat -> A}, exists i j: nat, ( i<j /\ R (u i) (u j));
    preorder_wqo :>  PreOrder R;
  }.

(* nat wqo *)
(***********)
Section nat_wqo.
Require Import Classical.
  
Theorem pigeonhole :
  forall m n, m < n -> forall f, (forall i, f i < m) ->
                      { i : nat & { j : nat | i < j < n /\ f i = f j } }.
Admitted.

Check pigeonhole.

Definition translate (k:nat) := ( fun  n => n+k).  
Definition shift (k:nat) (v:nat -> nat) := ( fun  n => v (n+k)).

Instance N_Wqo: Wqo le.
 split.
 + intro.
   generalize pigeonhole; intros.
   assert (~(forall i, forall j, i < j -> u j <u i)).
   * intro.
     generalize (H0 1);intro.
     generalize (H0 0 1);intro.
     assert (u 1 < u 0).
     apply H0;omega.
     specialize (H (u 1) (u 0) H3 (shift 2 u)).
     unfold shift in H.
     assert ( forall j : nat, u (j+2) < u 1).
     intros;apply H1;omega.
     specialize (H H4).
     destruct H as [i [j [H H5]]].
     specialize (H0 (i+2) (j+2)).
     rewrite H5 in H0;omega.
   * apply not_all_ex_not in H0.
     destruct H0 as  [i H0].
     apply not_all_ex_not in H0.
     destruct H0 as  [j H0].
     exists i;exists j.
     apply imply_to_and in H0;omega.
 + split.
   * unfold RelationClasses.Reflexive;intros;omega.
   * unfold RelationClasses.Transitive;intros;omega.
Qed.

End nat_wqo.

Variable A : Type.
Generalizable Variables A leA.

(** when applied to a pre-order relation leA , return the equivalence,
    the strict order and the incomparable relation associated to leA *)

Definition preorder_equiv {leA:relation A} `{P:@PreOrder A leA }(a b : A) :=
  leA a b /\ leA b a.

Definition ltA {leA:relation A} `{P:@PreOrder A leA }(a b : A) :=
  leA a b /\ ~ leA b a.

Definition preorder_incomp {leA:relation A} `{P:@PreOrder A leA }(a b : A) :=
  ~ (leA a b) /\ ~ (leA b a).

Definition compl_leA  {leA:relation A} `{P:@PreOrder A leA }(a b : A) :=
  ~ (leA a b).

Definition good_sequence  {leA:relation A} `{P:@PreOrder A leA } (u:nat -> A) :=
  exists i j:nat, leA (u i) (u j).


Lemma preorder_equiv_symetric {leA:relation A}{P:PreOrder leA}:
   forall a b, preorder_equiv a b -> preorder_equiv b a.
Proof.
  intros.
  destruct H;split;auto.
Qed.


(*============================================================================*)
(*accessibility *)
(****************)
(* Variable A : Type.
   Variable R : A -> A -> Prop.
 *)

Check Acc.
 (* forall A : Type, (A -> A -> Prop) -> A -> Prop *)
Print Acc.
(* Acc_intro : (forall y : A, R y l -> Acc R y) -> Acc R l *)
Check Acc (compl_leA).

Check well_founded.
 (* forall A : Type, (A -> A -> Prop) -> Prop *)
Print well_founded.
(* forall a : A, Acc R a *)
Check continues.
(*  forall A : Set, (A -> A -> Prop) -> list A -> list A -> Prop *)
Print continues.
(* Ct0 : forall (a : A) (l : list A),
          ~ greater leA a l -> continues leA (a :: l) l    *)
Check wqo_acc.
(* forall A : Set, (A -> A -> Prop) -> Prop  *)
Print wqo_acc.
(* forall l : list A, bad leA l -> Acc (continues leA) l *)


(*============================================================================*)
(* subsequences *)
(*********************)

Class Increasing  {D:Type} {leD:relation D} `{P:@PreOrder D leD } (f: nat -> D) : Prop :=
  {
     increase : forall i j:nat, i < j -> leD (f i) (f j) /\ ~ leD (f j) (f i);
}.


Definition subsequence (u : nat -> A) (n:nat -> nat) `{Incr: Increasing nat le n} := fun i:nat => u (n i).

Check subsequence.
(* (nat -> A) -> forall n : nat -> nat, Increasing n -> nat -> A *)


Instance incr_translate (k:nat): Increasing (translate k).
split;intros.
unfold translate.
omega.
Qed.

Instance incr_shift (k:nat) (v:nat -> nat) : Increasing v -> Increasing (shift k v).
split;intros.
unfold shift.
destruct H.
specialize increase0 with (i+k) (j+k).
apply increase0.
omega.
Qed.

Check incr_shift.
(* forall k : nat, Increasing (shift k) *)

Definition shift_seq (k:nat) (u : nat -> A):= 
  subsequence u (translate k).

Check shift_seq.
(* nat -> (nat -> A) -> nat -> A *)

(*============================================================================*)
(* upward closed sets *) 
(***********************)
Section upward_closed_sets.
Require Import Setoid.
Require Import   Sets.Ensembles.
Require Import Init.Wf.
Check In.

Class Upward_closed {D:Type} {leD:relation D} `{P:@PreOrder D leD } (B:Ensemble D)  :=
  {
     upw_closed : forall x y:D, In D B x/\ leD x y -> In D B y;
}.


Class Upward_closed_sequence {D:Type} {leD:relation D} `{P:@PreOrder D leD } (u:nat-> Ensemble D)   := 
  {
    upw_closed_seq :forall i:nat, Upward_closed (u i);
}.

Class Downward_closed {D:Type} {leD:relation D} `{P:@PreOrder D leD } (B:Ensemble D)  :=
  {
     dnw_closed : forall x y:D, In D B x/\ leD y x -> In D B y;
}.


Class Downward_closed_sequence {D:Type} {leD:relation D} `{P:@PreOrder D leD } (u:nat-> Ensemble D)   := 
  {
    dnw_closed_seq :forall i:nat, Downward_closed (u i);
}.


Inductive segment  (k:nat) : (Ensemble nat):=
    In_segment : forall n:nat, k <= n -> In nat (segment k) n.

Definition segments_seq := fun k:nat => (segment  k).
 
Check (Upward_closed_sequence (segments_seq)).


Instance preOrder_inclusion : PreOrder (Included nat).
split.
unfold Reflexive.
intro X.
unfold Included.
auto.
unfold Transitive.
intros X Y Z.
intros.
unfold Included.
intro.
unfold Included in H; specialize H with x.
unfold Included in H0; specialize H0 with x.
intro.
apply H0;apply H; trivial.
Qed.


Instance incr_segments_seq : Increasing segments_seq.

Check Upward_closed.


Instance upward_closed_segments : Upward_closed_sequence  segments_seq.
split.
intro.
constructor.
intros.
destruct H.
unfold segments_seq in H.
destruct H.
unfold segments_seq.
split.
omega.
Qed.

End upward_closed_sets.


(*============================================================================*)
(* equivalent definitions of wqo *)
(*********************************)
Definition wqo_2  `{P:@PreOrder A leA }: Prop :=
  forall u : nat -> A,  exists n: nat -> nat,  exists Incr: Increasing n,
    forall i j: nat,  i<j -> leA ((subsequence u n) i) ((subsequence u n) j).
    (* forall i j: nat,  i<j -> (n i)<(n j) /\ leA (u (n i)) (u (n j)). *)

                     
Definition wqo_3  `{P:@PreOrder A leA }: Prop :=
  ~ ( exists u : nat -> A, forall (i j: nat), ltA (u (j)) (u i) )
  /\ ~ ( exists u : nat -> A, forall (i j: nat), i<>j -> preorder_incomp (u i) (u j)).
    (* ~(exists u : nat -> A, forall (i j: nat), leA (u i) (u j) -> i=j ) *)


Definition  wqo_4 `{P:@PreOrder A leA }: Prop :=
  forall u: nat -> Ensemble A, 
    Increasing u ->Upward_closed_sequence u -> exists i:nat, 
                                  forall j:nat, i<=j -> (u i) = (u j).

Check well_founded.

Definition  wqo_4' `{P:@PreOrder A leA }: Prop :=
  forall u: nat -> Ensemble A, 
    Increasing u ->Downward_closed_sequence u -> exists i:nat, 
                                  forall j:nat, i<=j -> (u i) = (u j).


Lemma wqo1_wqo3 (leA : relation A) (WQO: Wqo leA):  wqo_3.
Proof.
  red; split.
  + intro.
    destruct H as [u H].
    destruct (WQO).
    destruct (well_order0 u) as [i [j H1]].
    destruct H1.
    destruct (H i j);auto.
  + intro.
    destruct H as [u H].
    destruct (WQO).
    destruct (well_order0 u) as [i [j H1]].
    destruct H1.
    specialize H with i j.
    unfold preorder_incomp in H.
    apply H in H1;omega.
Qed.


Lemma wqo2_wqo1 `{P:@PreOrder A leA }:  wqo_2 -> (Wqo leA).
Proof.
  intro.
  unfold wqo_2 in H.
  split.
  intro; destruct (H u)  as [f H0].
  destruct H0  as [incr H0].
  unfold subsequence in H0.
  exists (f 0); exists (f 1).
  specialize (H0 0 1).
  split.
  + destruct incr.
    specialize (increase0 0 1).
    omega.
  + apply H0.
    auto.
  + assumption.
Qed.

Lemma wqo3_wqo4' `{P:@PreOrder A leA }: wqo_3 -> wqo_4' .
Proof.
  intro.
  unfold wqo_3 in H.
  unfold wqo_4'.
  intros.
Admitted.  

Require Import Logic.Classical_Prop.
Require Import Classical.

Lemma not_and_or_not: forall P Q : Prop, ~ (P /\ Q) <-> ~ P \/ ~ Q.
Proof.
intros.
split.
apply not_and_or.
apply or_not_and.
Qed.

Set Implicit Arguments.
Arguments In {U} _ _.


Lemma wqo3_wqo2 `{P:@PreOrder A leA }: ~ wqo_2 -> ~wqo_3.
Proof.
  intro.
  unfold wqo_2  in H.
  generalize (not_all_ex_not _ _ H);intro.
  clear H.
  destruct H0 as [u0 H0].
  unfold wqo_3;intro.
  generalize (not_ex_all_not _ _ H0);intro.
  clear H0.
  assert ( forall n : nat -> nat, forall Incr : Increasing n, exists i j: nat,  i < j /\ (~ (n i < n j) \/ ~(leA (u0 (n i)) (u0 (n j))))).
  intros.
  specialize (H1 n).
  generalize (not_ex_all_not _ _ H1);intro.
  specialize (H0 Incr).
  apply not_all_ex_not in H0.
  destruct H0.
   apply not_all_ex_not in H0.
  destruct H0.
  exists x; exists x0.
  apply imply_to_and in H0.
  unfold subsequence in H0.
  destruct H0.
  split; trivial.
  right;trivial.
  clear H1.
Admitted.

Lemma wqo1_wqo4 (leA : relation A) (WQO: Wqo leA):  wqo_4.
Proof.
  intro.
  destruct WQO.
  intros.
  unfold compl_leA in H.
Admitted.

Lemma wqo4_wqo1 `{P:@PreOrder A leA }:  wqo_4 -> (Wqo leA).
Proof.
  intro.
  unfold wqo_4 in H.
  split.
  intros.
Admitted.  



  


  
