(*  Pierre Casteran 
    LaBRI, Universit\u00e9 de Bordeaux 
*)

Require Import Arith Max Bool Omega  Compare_dec  Relations Ensembles
               Wellfounded.
Require Import Operators_Properties  Prelude.More_nat  Prelude.Restriction
        not_decreasing.
Require Import ArithRing   DecPreOrder Logic.Eqdep_dec  Prelude.Proofunicity.
Set Implicit Arguments.

Coercion is_true: bool >-> Sortclass.



(** * Cantor "pre" Normal form 
 After Manolios and Vroon work on ACL2 

 T1 is a type of terms used for representing Cantor normal form

ocons a n b is intended to represent
  the ordinal  omega^a *(S n)  + b 
*)

Inductive T1 : Set  :=
| zero : T1
| ocons : T1 -> nat -> T1 -> T1.



(** An alternative notation for Cantor Normal Form, using strictly positive 
    coefficients
*)
Require Import PArith.

Inductive T1p :=
  zerop | consp : T1p -> positive -> T1p -> T1p.

Fixpoint unshift (alpha : T1p) : T1 :=
  match alpha with
      zerop => zero
    | consp beta p gamma => ocons (unshift beta)
                                  (Nat.pred (Pos.to_nat p))
                                  (unshift gamma)
      end.

Fixpoint shift (alpha : T1) : T1p :=
  match alpha with
      zero => zerop
    | ocons beta n gamma => consp (shift beta)
                                  (Pos.of_nat (S n))
                                  (shift gamma)
      end.

Notation "{w ^ alpha '*' i '+' beta }" := (consp alpha i beta) (at level 0).




(** ** boolean tests *)

Fixpoint is_succ alpha :=
  match alpha with
      zero => false
    | ocons zero _ _ => true
    | ocons alpha n beta => is_succ beta
  end.

Fixpoint is_limit alpha :=
  match alpha with
      zero => false
    | ocons zero _ _ => false
    | ocons alpha n zero => true
    | ocons alpha n beta => is_limit beta
  end.


(** *** Additive principals 
 *)

Inductive ap : T1 -> Prop :=
  ap_intro : forall alpha, ap (ocons alpha 0 zero).

(** *** some abreviations *)

(** omega^alpha * (S k) 
*)

Definition omega_term (alpha:T1)(k:nat) :=
   ocons alpha k zero.

Definition phi0 alpha := ocons alpha 0 zero.

(**
 FS i  denotes the successor of i 
*)
Notation "'FS' n" := (ocons zero n zero) (at level 29) : cnf_scope.

Open Scope cnf_scope.

Definition fin (n:nat) := match n with 0 => zero | S p => FS p end.

Notation omega := (ocons (ocons zero 0 zero) 0 zero).

Notation one := (ocons zero 0 zero).

Coercion fin  : nat >-> T1.

(** omega-based  logarithm 
*)

Definition log alpha :=
  match alpha with
    | zero => zero 
    | ocons a _ _ => a
  end.

(**  omega towers 
*)

Fixpoint tower (height:nat) : T1 := 
 match height with 
| 0 =>  FS 0 
| S h' => phi0 (tower h')
 end.

Definition omega_ c := tower (S c).

Fixpoint height alpha :=
match alpha with
    zero => 1
  | ocons zero _ _ => 1
  | ocons alpha _ _ => S (height alpha)
end.


Lemma height_of_tower : forall n, 0 < n -> height (tower n) = S n.  
Proof.
  induction n.
  -  reflexivity.
  -   destruct n.
     + now  simpl.
     +     cbn;  simpl in IHn;    destruct (tower n).    
       *     discriminate IHn;auto with arith.
       *  rewrite IHn;auto with arith.
 Qed.


(**  ** A total strict order on T1 
 *)


Fixpoint compare (alpha alpha':T1):comparison :=
  match alpha, alpha' with
    zero, zero => Eq
  | zero, ocons a' n' b' => Lt
  | _   , zero => Gt
  | (ocons a n b),(ocons a' n' b') =>
      (match compare a a' with 
          | Lt => Lt
          | Gt => Gt
          | Eq => (match lt_eq_lt_dec n n'
                   with
                       inleft  (left _) => Lt
                     | inright _ => Gt
                     |   _ => compare b b'
                   end)
       end)
  end.


Definition lt_b alpha beta : bool :=
  match compare alpha beta with
      Lt => true
    | _ => false
  end.

Definition lt alpha beta : Prop := lt_b alpha beta.

Example E1 : lt (ocons omega 56 zero) (tower 3).
Proof. reflexivity. Qed.

Example E2 : ~ lt (tower 3) (tower 3).
 discriminate. 
Qed.


Lemma compare_refl : forall alpha, compare alpha alpha =  Eq.
Proof.
 induction alpha; simpl;auto.
 rewrite IHalpha1, IHalpha2.
 destruct (Compare_dec.lt_eq_lt_dec n n).
 - destruct s; auto.
  + destruct (Lt.lt_irrefl _ l).
 - destruct (Lt.lt_irrefl _ l).
Qed.

Lemma zero_lt : forall alpha n beta, lt  zero (ocons alpha  n beta).
Proof.  reflexivity. Qed. 

Lemma head_lt :
  forall alpha alpha' n n' beta beta',
       lt alpha  alpha' -> 
       lt (ocons alpha n beta)  (ocons alpha' n' beta').
Proof.
  unfold lt, lt_b ;  intros.   
  cbn;  destruct (compare alpha alpha') ; auto;   discriminate. 
Qed.

Lemma coeff_lt : forall alpha n n' beta beta',
                (n < n')%nat -> 
                lt (ocons alpha n beta)  (ocons alpha n' beta').
Proof. 
  unfold lt, lt_b ;  intros;  cbn;   rewrite compare_refl.
  destruct (lt_eq_lt_dec n n').
   - destruct s;  auto.
     + subst; elimtype False; omega. 
   - destruct (Nat.lt_irrefl n);   eauto with arith. 
Qed. 

Lemma tail_lt : forall alpha n beta beta',
               lt beta  beta' -> 
               lt (ocons alpha n beta)  (ocons alpha n beta').
Proof.
 intros; unfold lt, lt_b; cbn; rewrite compare_refl.
  destruct (lt_eq_lt_dec n n).
  - destruct s; auto. 
  - elimtype False; omega. 
Qed. 

Hint Resolve zero_lt head_lt coeff_lt tail_lt  : T1.

Open Scope cnf_scope.
Delimit Scope cnf_scope with cnf.

(* the total non-strict order associated with lt   *)

Definition le (alpha beta :T1) := match compare alpha beta with
                                      Gt => False | _ => True end. 
Hint Unfold le : T1. 


(**   The predicate "to be in normal form" 

Cantor normal form needs the exponents of omega to be
   in strict decreasing order *)


Fixpoint nf_b (alpha : T1) : bool :=
  match alpha with
    | zero => true
    | ocons a n zero => nf_b a
    | ocons a n ((ocons a' n' b') as b) =>
      (nf_b a && nf_b b && lt_b a' a)%bool
  end. 

Definition nf alpha :Prop := nf_b alpha .

Lemma zero_nf : nf zero. 
Proof. reflexivity. Qed.


Lemma single_nf :  forall a n, nf a ->  nf (ocons a n zero).
Proof.   unfold nf; now cbn. Qed. 

Lemma ocons_nf : forall a n a' n' b, lt a'  a -> 
                           nf a -> 
                           nf(ocons a' n' b)-> 
                           nf(ocons a n (ocons a' n' b)).
Proof. 
unfold nf; cbn; intros a n a' n' b H H0 H1;  rewrite H0.
destruct b. 
- red in H; rewrite H1, H; reflexivity.
- simpl; simpl in H1;
  destruct (nf_b a'), (lt_b b1 a'), (nf_b b1), (nf_b b2), b2; simpl in *; auto.
  all : destruct (lt_b b2_1 b1); simpl in *;auto.
Qed. 

Hint Resolve zero_nf single_nf ocons_nf.

Lemma nf_inv1 : forall a n b, nf (ocons a n b) -> nf a.
Proof.
  unfold nf;destruct b; cbn.
  - auto.
  - destruct (nf_b a); auto.
Qed.

Lemma nf_inv2 : forall a n b, nf (ocons a n b) -> nf b.
Proof.
  unfold nf; cbn; destruct a, b;  auto with T1.
  destruct (nf_b (ocons b1 n0 b2)); auto. 
  destruct (nf_b (ocons b1 n1 b2)); auto.
  destruct (nf_b (ocons a1 n a2)); auto.
Qed. 

Lemma nf_inv3 : forall a n  a' n' b', nf (ocons a n (ocons a' n' b')) ->
                                    lt a' a.
Proof.
  unfold nf; cbn;
  destruct a, a', b';  try discriminate; auto with T1.
  -   intro H; red in H;repeat rewrite andb_true_iff in H. 
      decompose [and] H;discriminate.  
  -   intro H;  red in H;  repeat rewrite andb_true_iff in H; decompose [and] H;discriminate.  
  -   intro H;   red in H; repeat rewrite andb_true_iff in H; decompose [and] H;discriminate.  
  -   unfold lt;auto.
      intro H;  red in H;  repeat rewrite andb_true_iff in H.
      decompose [and] H; auto. 
  - intros H; red in H; repeat rewrite andb_true_iff in H;  decompose [and] H; auto. 
Qed.

Lemma lt_proof_unicity :
  forall alpha beta (H H': lt alpha beta), H = H'.
Proof.
 intros; red in H, H';  apply eq_proofs_unicity_on.
  destruct y;  rewrite H; auto. 
 right; discriminate. 
Qed.

Lemma nf_proof_unicity :
  forall alpha  (H H': nf alpha), H = H'.
Proof.
  intros;  red in H, H';  apply eq_proofs_unicity_on.
  destruct y. 
  rewrite H; auto. 
  rewrite H; right; discriminate. 
Qed.

Class Ord (alpha : T1) : Prop  :=  {HNF : nf alpha }.


Lemma Ord_unicity : forall alpha (O O' : Ord alpha), O = O'.
Proof. 
  destruct O, O';now rewrite (nf_proof_unicity HNF0 HNF1). 
Qed. 


(** epsilon0 as a set *)

Definition epsilon_0 : Ensemble T1 := nf.

(*  "lt_phi0 beta alpha "means "beta less than (phi0 alpha)" *)

Inductive lt_phi0 : T1 -> T1 -> Prop :=
| lt_phi0_z : forall alpha, lt_phi0 zero alpha
| lt_phi0_c : forall alpha alpha' n' beta',
                lt alpha' alpha -> 
                lt_phi0 (ocons alpha' n' beta') alpha.

Hint Constructors lt_phi0 : T1.

(** *  Ordinal arithmetics *)


Definition max alpha beta := match compare alpha beta
                             with Lt => beta | _ => alpha end.



Fixpoint succ (c:T1) : T1 :=
  match c with zero => FS 0
             | ocons zero n _ => ocons zero (S n) zero
             | ocons a n b => ocons a n (succ b)
  end.


Fixpoint pred (c:T1) : option T1 :=
  match c with zero => None
             | ocons zero 0 _ => Some zero
             | ocons zero (S n) _ => Some (ocons zero n zero)
             | ocons a n b => 
                  match (pred b) with None => None
                                    | Some c => Some (ocons a n c)
                  end
  end.

Fixpoint plus (c1 c2 : T1) {struct c1}:T1 :=
  match c1,c2 with
 |  zero, y  => y
 (*|  ocons zero n _, zero => ocons zero n zero *)
 |  x, zero  => x
 (*|  ocons zero n _, ocons zero n' _ => ocons zero (S (n+n')) zero *)
 |  ocons a n b, ocons a' n' b' =>
      (match compare a a' with
        | Lt => ocons a' n' b'
        | Gt => (ocons a n (plus b (ocons a' n' b')))
        | Eq  => (ocons a (S(n+n')) b')
       end)
 end
 where "a + b" := (plus a b) : cnf_scope.


Fixpoint minus (c1 c2 : T1) {struct c1}:T1 :=
  match c1,c2 with
 |  zero, y  => zero
 |  ocons zero n _, ocons zero n' _ =>
                  if (le_lt_dec n n') 
                  then zero
                  else  ocons zero (Peano.pred (n-n')) zero
 |  ocons zero n _, zero =>  ocons zero n zero
 |  ocons zero n _, y =>  zero
 |  ocons a n b, zero =>  ocons a n b
 |  ocons a n b, ocons a' n' b' =>
      (match compare a a' with 
            | Lt => zero
            | Gt => ocons a n b
            | Eq  => (match (lt_eq_lt_dec n n') with
                            | inleft (left _) => zero
                            | inright _ => (ocons a (Peano.pred (n-n')) b')
                            | _ =>  b - b'
                      end)

       end)
 end
 where  "c1 - c2" := (minus c1 c2):cnf_scope.


Compute omega + omega + FS 5.

Fixpoint mult (c1 c2 : T1) {struct c2}:T1 :=
  match c1,c2 with
 |  zero, y  => zero
 |  x, zero => zero
 |  ocons zero n _, ocons zero n' _ => 
                 ocons zero (Peano.pred((S n) * (S n'))) zero
 |  ocons a n b, ocons zero n' b' =>  
                 ocons a (Peano.pred((S n) * (S n'))) b
 |  ocons a n b, ocons a' n' b' =>
     ocons (a + a') n' ((ocons a n b) * b')
 end
where  "c1 * c2" := (mult c1 c2) : cnf_scope.

Fixpoint exp_F (a:T1)(n:nat){struct n}:T1 :=
 match n with 
 | 0 =>  FS 0
 | S p => a * (exp_F a p)
 end.


Fixpoint exp  (a b : T1) {struct b}:T1 :=
  match a,b with
 |  x, zero => ocons zero 0 zero
 | ocons zero 0 _ , _ => ocons zero 0 zero
 | zero, y            => zero
 | x , ocons zero n' _ =>  exp_F x (S n')
 | ocons zero n _, ocons  (ocons zero 0 zero) n'  b' =>
        ((ocons (ocons zero n' zero) 0 zero) * 
                ((ocons zero n zero) ^  b'))
 | ocons zero n _, ocons  a' n' b' =>
            (omega_term 
                    (omega_term (a' - (FS 0)) n')
                    0) *
                 ((ocons zero n zero) ^ b')
 | ocons a  n b, ocons  a' n' b' =>
    ((omega_term   (a * (ocons a' n' zero)) 
                            0) * 
            ((ocons a n b) ^ b'))
end 
where "a ^ b" := (exp a b) : cnf_scope.


(* properties of the T1 data type *)
(************************************)

(* A tactic for decomposing a non zero ordinal *)
(* use : H : lt zero ?c ; a n b : names to give to the constituents of 
   c *)

Definition get_decomposition : forall c:T1, lt zero c ->
                           {a:T1 & {n:nat & {b:T1 | c = ocons a n b}}}.
Proof.
 intro c; case c.
 - intro H; elimtype False; inversion H. 
 -  intros c0 n c1; exists c0;exists n;exists c1;auto.
Defined.

Ltac decomp_exhib H a n b e:= 
 let Ha := fresh in
 let Hn := fresh in
 let Hb := fresh in
  match type of H
  with lt zero ?c =>
    case (get_decomposition  H);
     intros a Ha;
     case Ha;intros n Hn; case Hn; intros b e;
     clear Ha Hn 
  end.

     
(* Properties of lt *)
(**********************)


Lemma compare_ok_1: 
  forall a a', (compare a a' = Lt <-> lt a  a') /\ 
          (compare a a' = Eq <-> a = a') .
Proof.
  induction a;simpl.
  -  destruct a'; cbn.
     + split;split.
       *  discriminate 1.
       * discriminate.
       *  reflexivity.
       *  reflexivity.
     +  split;split.
        * reflexivity. 
        * reflexivity.  
        *  discriminate 1.
        * discriminate 1.
  - destruct a'; simpl.
    +  split; split.
       *  discriminate 1.
       *  discriminate. 
       * discriminate. 
       *  discriminate 1.
    +  case_eq (compare a1 a'1).
       * intros H;  case (Compare_dec.lt_eq_lt_dec n n0).
         {  destruct s.
            -  repeat split.
               replace a1 with a'1.
               {   auto with T1.  }
               case (IHa1 a'1).
               + intros H0 H1; destruct H1.
                 * symmetry; auto.  
               +  discriminate 1.
               + injection 1.
                 intros;  subst n0; destruct (Lt.lt_irrefl n);auto with arith.
            - repeat split.
              intro H0; subst n0;  replace a1 with a'1.
              + apply tail_lt; auto. 

                *  case (IHa2 a'2);intros;auto.
                   case H1;auto.
              +  elim (IHa1 a'1);intros H1 H2; case H2;auto.
                 symmetry;auto. 
              + subst n0; replace a1 with a'1.
                *  intro H0;  assert (H1 : lt a2  a'2).
                   { unfold lt, lt_b in H0. cbn in H0.
                 rewrite compare_refl in H0.
                 destruct (lt_eq_lt_dec n n). 
                 destruct s.
                   destruct (Nat.lt_irrefl _ l).
                   case_eq (compare a2 a'2).
                   { intro H1; rewrite H1 in H0. discriminate. }
                   { intro; unfold lt, lt_b.  auto.  }
                   { intro H1; rewrite H1 in H0. discriminate. }
                  discriminate. 
              }
                   unfold lt, lt_b in H1. destruct (compare a2 a'2);
                     try discriminate; auto.
             *     case (IHa1 a'1);intros H0 H1;  case H1;symmetry; auto.              
           +          
                   case (IHa2 a'2);intros H2 H3; case H2;auto.
                *  case (IHa1 a'1);intros H0 H1;  case H1;auto.
                   intros;symmetry;auto.
                replace a1 with a'1.
                  subst n0.  
                   replace a2 with a'2.
                   reflexivity.
                   case H2;intros.
                   symmetry;auto.
                   rewrite <- H3;auto. 
                  symmetry; rewrite <- H1; auto. 
              +  injection 1;  intro;subst a'2.
                case (IHa2 a2).
                intros; case H2;auto.
         }
         {
           case (IHa1 a'1);intros.
           split. 
           split.
           discriminate 1.
           intro H2.
           
           replace a1 with a'1 in H2.
           { unfold lt, lt_b in H2; simpl in H2.
              rewrite compare_refl in H2.
            destruct (lt_eq_lt_dec n n0).
     destruct s.
  elimtype False; omega. 
 subst n0; elimtype False; omega.
   discriminate H2.
           }
   symmetry; rewrite <- H1; auto. 
       split. 
       discriminate 1.
 intro. 
   destruct H1.
 rewrite (H1 H) in H2.
  injection H2.
  intros; subst. 
   destruct (Nat.lt_irrefl n0 l).
         }
        
       *    intros; repeat split.
            { intro; apply head_lt. unfold lt, lt_b. rewrite H. auto. }
            { discriminate 1. }
            {   injection 1;intros.
                subst a'1; rewrite compare_refl in H. discriminate H. }
       * intros; repeat split.
         {   discriminate 1. }
         { 
           -   destruct (IHa1 a'1).
               destruct (IHa2 a'2).
               clear IHa1 IHa2.
               unfold lt, lt_b; cbn. 
               rewrite H. discriminate. 
         }
        discriminate. 
         { injection 1; intros;subst.
           destruct (IHa1 a'1).
           generalize (refl_equal a'1); intro H3; 
    rewrite <- H2 in H3; rewrite H3 in H; discriminate.
         }
Qed.




Lemma compare_rev : forall a b, compare b a = CompOpp (compare a b).
Proof with auto.
  induction a; destruct b; cbn; auto.
  rewrite IHa1. destruct  (compare a1 b1).
  destruct (lt_eq_lt_dec n n0).
   destruct s.
   destruct (lt_eq_lt_dec n0 n). 
   destruct s. 
   elimtype False; omega.
   subst;  elimtype False; omega.
    reflexivity. 
    destruct (lt_eq_lt_dec n0 n). 
    destruct s.
  subst;  elimtype False; omega.
   rewrite IHa2.
 simpl CompOpp.
 reflexivity. 
  subst;  elimtype False; omega.
   destruct (lt_eq_lt_dec n0 n). 
 destruct s.
 reflexivity. 
 subst;  elimtype False; omega.
  elimtype False; omega.
   simpl CompOpp.
  reflexivity. 
  reflexivity.
Qed.




Lemma compare_reflect : forall a a',
                          match compare a a' with
                            |   Lt => lt a  a'
                            |   Eq => a = a'
                            |   Gt => lt a'  a
                          end.
Proof.
  intros c c';  case (compare_ok_1 c c');intros H0 H1; case H0; case H1;
  intros;  clear H0 H1.

  case_eq (compare c c'); auto.
  rewrite compare_rev.

  intro. 
unfold lt, lt_b.
rewrite CompOpp_iff in H0.
now rewrite H0.
Qed.



Theorem not_lt_zero : forall a, ~ lt a  zero.
Proof.  destruct a; compute; discriminate. Qed.

Hint Resolve not_lt_zero : T1.






 Inductive lt_cases (a  b : T1) (n :nat) (a' b':T1) (n':nat) : Type :=
    lt_left (H : lt a a')
  | lt_middle (H : a = a')(H1 : (n < n')%nat)
  | lt_right (H : a = a')(H1 : n = n')(H2 : lt b b').


Lemma lt_inv_strong : forall a n b a' n' b', 
                        lt (ocons a n b) (ocons a' n' b') ->
                        lt_cases a b n a' b' n'. 
Proof. unfold lt, lt_b.
       intros; simpl in H.
       case_eq (compare a a').
       intro Ha; rewrite Ha in *.
       replace a' with a.
        destruct  (lt_eq_lt_dec n n'). 
        destruct s. 
        constructor 2. reflexivity. auto. 
  case_eq  (compare b b'); auto.
 intro H0. rewrite H0 in H. discriminate. 
 intro; constructor 3. reflexivity. 
 auto.
 rewrite H0 in H.
 red. unfold lt_b;auto. 
 now rewrite H0. 
 intro H0; rewrite H0 in H. discriminate.  discriminate. 
 generalize (compare_reflect a a'); now rewrite Ha.
intro H0; rewrite H0 in H.
constructor 1.
red. unfold lt_b;auto.
rewrite H0;auto.
intro H0; rewrite H0 in H.
  discriminate.
 Qed.

  Theorem lt_inv : forall a n b a' n' b', 
                      lt (ocons a n b) (ocons a' n' b') ->
                      lt a  a' \/
                      a = a' /\ (n < n')%nat \/
                      a = a' /\ n = n' /\ lt b  b'.
  Proof.
   intros; destruct (lt_inv_strong H).
   - now left. 
   - right; left;auto. 
   -  right; right; auto. 
 Qed.      


Theorem lt_irrefl : forall a, ~ lt a  a.
Proof. 
 induction a.
 -  red;inversion_clear 1.
 -  intro H; case (lt_inv_strong  H); intuition.
Qed.

Hint Resolve lt_irrefl : T1.

Lemma lt_inv_nb : forall a n n' b b',
                       lt (ocons a n b) (ocons a n' b') ->
                       {(n<n')%nat} + { n=n' /\ lt b  b'}.
Proof.
 intros.
 destruct (lt_inv_strong H).
  destruct (lt_irrefl H0).
  destruct H0.
 intuition.
 intuition.
Qed.


Lemma lt_inv_b : forall a n b b',
       lt (ocons a n b) (ocons a n b') -> lt b  b'.
Proof.
 intros a n b b' H; destruct (lt_inv_strong H).
  destruct (lt_irrefl H0).
  destruct H.
 destruct H0.
  destruct (Nat.lt_irrefl n);tauto.
 tauto. 
Qed.

(* lt is a (strict) order *)

Theorem lt_trans : forall a b, lt a  b -> 
                        forall c, lt b c -> lt a  c.
Proof.
  induction  a.
  - destruct b.
    discriminate. 
    destruct c.
    discriminate. 
    auto with T1.
  -
    destruct b.
    discriminate. 
   destruct c.
    discriminate. 
   destruct (lt_inv_strong H).
  intro H2; destruct (lt_inv_strong H2).

  eauto with T1.
  subst. auto with T1.
  subst. auto with T1.
  subst. clear H. destruct H1.
 
   intro H2; destruct (lt_inv_strong H2).
  auto with T1.
 subst. 
  eauto with arith T1.
 subst. 
 auto with arith T1. 
 intro H2; destruct (lt_inv_strong H2).
  auto with T1. 
 subst; eauto with arith T1. 
 subst; eauto with arith T1.  
 subst. 
  clear H. 
intro H3; destruct (lt_inv_strong H3).
  auto with T1.
   subst; auto with T1.
  subst; eauto with T1.
Qed.


Theorem lt_not_gt : forall a b, lt a  b  -> ~ lt b  a.
Proof.
 intros one o2 H H0;
  generalize (lt_trans  H H0);  intro H2; case (lt_irrefl H2).
Qed.

Lemma finite_lt : forall n p : nat, (n < p)%nat ->
                                    lt (FS n)   (FS p).
Proof.
 intros; auto with T1.   
Qed.

Lemma finite_ltR :
  forall n p : nat, lt (FS n) (FS p) -> (n < p)%nat.
Proof.
  intros; simpl in H. unfold lt, lt_b in H. 
  simpl in H.   destruct (lt_eq_lt_dec n p).  destruct s; auto. discriminate.
  discriminate. 
Qed.

Lemma le_refl : forall alpha, le alpha alpha.
Proof. 
 unfold le.
 intro; now rewrite compare_refl.
Qed.

Lemma lt_le_incl : forall alpha beta, lt alpha beta -> le alpha beta.
Proof.
 unfold le, lt, lt_b; intros.
  destruct (compare alpha beta); auto; try discriminate.
Qed. 


Lemma le_eq_lt_dec : forall alpha beta, le alpha beta ->
                                       {alpha = beta} + {lt alpha beta}.
Proof.
 unfold le, lt, lt_b.
  intros; case_eq (compare alpha beta).
 left.
 generalize (compare_reflect alpha beta); now rewrite H0.
  right; auto. 
  intro H0; rewrite H0 in H; contradiction. 
Qed.

(* to remove ... *)
Lemma le_lt_or_eq : forall alpha beta, le alpha beta ->
                                       alpha = beta \/ lt alpha beta.
Proof.
 intros alpha beta H; destruct (le_eq_lt_dec _ _ H);auto. 
Qed. 





(* The following property is specific to the ordinal epsilon0: *) 

Lemma lt_a_phi0_a : forall a, lt a  (phi0 a).
Proof.
 induction a;simpl.
 - unfold phi0;  auto with T1. 
 -  unfold phi0 in *. 
    assert (H : le (ocons a1 0 zero) (ocons a1 n a2)).
    {  case n.
      -  case a2.
        +  apply le_refl. 
        +  intros; apply lt_le_incl. auto with T1 arith. 
      -  intros; apply lt_le_incl; auto with T1 arith. 
     }
  apply head_lt.

  destruct (le_lt_or_eq _ _ H).
 injection H0; intros;  now subst. 
  eapply lt_trans; eauto.  
Qed.


Lemma phi0_lt : forall a b, lt a  b -> lt (phi0 a) (phi0 b).
Proof.
 unfold phi0;intros c c' H; auto with T1.  
Qed.


Lemma phi0_ltR : forall a b, lt (phi0 a)  (phi0 b) -> lt  a  b.
Proof.
  unfold phi0;intros c c' H; case (lt_inv H).
  -  trivial.
  -  intros [(_,i)|(_,(_,i))]; inversion i.
Qed.

Theorem zero_le : forall a, le zero a.
Proof.
 unfold le; destruct  a;  reflexivity. 
Qed.

Theorem le_trans : forall a b c, le a  b -> le b  c -> le a  c.
Proof.
 intros a b c H H0; destruct (le_lt_or_eq _ _ H).
 subst; auto.
  destruct (le_lt_or_eq _ _ H0).
  now subst. 
 apply lt_le_incl;eapply lt_trans;eauto.
Qed.



Theorem le_lt_trans : forall a b c, le a  b -> lt b  c -> lt a  c.
Proof.
   intros a b c H H0; destruct (le_lt_or_eq _ _ H).
 -  subst b;auto.
 -  intros;eapply lt_trans;eauto.
Qed.


Theorem lt_le_trans : forall a b c, lt a  b -> le b  c -> lt a  c.
Proof.
  intros a b c H H0; destruct (le_lt_or_eq _ _ H0).
 -  subst b;auto.
 -  eapply lt_trans;eauto.
Qed.


Theorem le_inv : forall a n b a' n' b', 
                      le (ocons a n b)  (ocons a' n' b') ->
                      lt a  a' \/
                      a = a' /\ (n < n')%nat \/
                      a = a' /\ n = n' /\ le b  b'.
Proof.
 intros a n b a' n' b' H; destruct (le_lt_or_eq _ _  H). 
 -  injection H0; intros; subst; right.
 right.
 repeat split;auto. 
apply le_refl. 
 -  destruct  (lt_inv H0).
 now left. 
 destruct H1.
 right; auto. 
 right; auto. decompose [and] H1; right; repeat split; auto.
 apply lt_le_incl; auto. 
Qed.


Lemma lt_phi0_inv :
  forall a n b a', lt (ocons a n b) (phi0 a') -> lt a a'.
Proof. 
  intros a n b a' H; destruct (lt_inv H); auto. 
  destruct H0; decompose [and] H0; clear H H0.
  inversion H2.
 destruct (not_lt_zero H4).
Qed.

Lemma lt_not_le: forall a b,  lt a b -> ~  le b  a.
Proof.
 intros a b H H0; destruct (le_lt_or_eq _ _ H0).
 - subst; destruct (lt_irrefl H).
 -    absurd (lt a a); eauto with T1.
     eapply lt_trans;eauto.
Qed.

Hint Resolve lt_le_incl le_refl : T1.


Lemma lt_inv_le : forall a n b a' n' b', 
                     lt (ocons a n b)  (ocons a' n' b') -> 
                     le a  a'.
Proof.
 intros a n b a' n' b' H; case (lt_inv H).
 -  auto with T1.
 -  intros [(e,i)|(e,(e',i))].
    + subst a; auto with T1.
    +  subst a; auto with T1.
Qed.


Theorem le_zero_inv : forall a, le a  zero -> a = zero.
Proof.
  intros a H; destruct (le_lt_or_eq _ _ H).
  auto.
   destruct (not_lt_zero H0).
Qed.

Theorem le_tail : forall a n b b', le b  b' -> 
                                   le (ocons a n b) (ocons a n b').
Proof.
 intros;  destruct (le_lt_or_eq _ _ H).
 -  subst b; auto with T1.
 -  auto with T1.
Qed.

Hint Resolve zero_le le_tail : T1.

Theorem le_phi0 : forall a n  b , 
                    le (phi0 a ) (ocons a n b).
Proof.
 induction  n.
 -   unfold phi0.  intro; apply le_tail. auto with T1.
 -   intros.  apply lt_le_incl;  apply le_lt_trans with (ocons a n b).
     + auto.
     + auto with T1 arith.   
Qed.

Lemma head_lt_ocons : forall a n b, lt a  (ocons a n b).
Proof.
 induction a.
 -  auto with T1. 
 -  auto with T1. 
Qed.


Definition T1_eq_dec : forall (a b : T1), {a = b}+{a <> b}.
Proof.  decide equality; apply eq_nat_dec. Defined.  


Definition lt_eq_lt_dec  : forall a b, {lt a  b}+{a = b}+{lt b  a}.
Proof.
 induction a; destruct b; auto with T1.
 case (IHa1 b1);intros s.
 -  case s;intros; auto with T1.
     subst b1; case (Compare_dec.lt_eq_lt_dec n n0).
  +  destruct 1.
     auto with T1.
     subst n;  case (IHa2 b2); auto with T1.
     destruct 1;[idtac| subst b2];auto with T1.
  +  right;auto with T1.
 -  auto with T1.
Defined.

Definition lt_le_dec : forall a b, {lt a  b}+{le b  a}.
Proof.
 intros a b; case (lt_eq_lt_dec a b).
 -  destruct 1.
    + left;auto.
    + subst;   right; auto with T1.
 -   right; auto with T1.
Defined.

Lemma le_lt_eq (a b : T1) : le a  b -> {lt a  b} + { a = b}.
Proof.
  intros H; destruct (lt_le_dec a b);auto.
  right.   destruct (le_lt_or_eq _ _ H),(le_lt_or_eq _ _ l); auto. 
   destruct (@lt_irrefl a);
  now apply lt_trans with b. 
Qed. 


Lemma lt_phi0_log : forall alpha beta, alpha <> zero -> lt_phi0 beta alpha ->
                                       lt (log beta)  alpha.
Proof.
 destruct 2; auto.
 -  simpl; destruct (@le_lt_eq zero alpha); auto.
    + apply zero_le;  auto.
    + subst alpha; now destruct H.
Qed.

Lemma log_lt_phi0 : forall a b, lt (log b)  a -> lt_phi0 b a. 
Proof. 
destruct b.
- constructor.
- simpl; now constructor.
Qed.

Instance epsilon0_pre_order : TotalDecPreOrder le.
Proof.
split.
- split.
 + intro x; apply le_refl.
 +  red;apply le_trans.
-  split.
   + intros a b; destruct (lt_le_dec a b).
     *  left; auto with T1.
     *  destruct (lt_le_dec a b).
        left. auto with T1. now right.
   + intros a b; destruct (lt_eq_lt_dec a b) as [[H1 | H2] | H3].
     *  left;auto with T1.
     * subst a ; left; auto with T1.
     * right;intro H4; now destruct (lt_not_le H3).
Defined.


Ltac auto_nf :=
  match goal with |- nf ?alpha =>
                  ( repeat (apply ocons_nf || apply single_nf || apply zero_nf)) || (eapply nf_inv1 || eapply nf_inv2); eauto
                                                                                 end.

Lemma nf_tail_lt_nf : forall  b b':T1 , lt b'  b -> nf b' ->
                                    forall a n,   nf (ocons a n b) ->
                                                  nf (ocons a n b').
Proof.
 intros. 
 destruct b. 
   destruct (not_lt_zero  H).
  destruct b'. 
  auto_nf. 
  auto_nf. 
  destruct (lt_inv H).
 
  apply ocons_nf. 
  apply lt_trans with b1;  auto.
  eapply nf_inv3; eauto.
  eapply nf_inv1;eauto. 
 auto. 
  destruct H2.
 decompose [and] H2;subst. 

 apply ocons_nf;auto.
 eapply nf_inv3;eauto.
 
  eapply nf_inv1;eauto.
decompose [and] H2;subst. 
 clear H2.
 apply ocons_nf;auto.
 eapply nf_inv3;eauto.
 eapply nf_inv1;eauto.
Qed.


Lemma tail_lt_ocons : forall b n a, 
     nf (ocons a n b)-> 
     lt b  (ocons a n b).
Proof.
 induction b.
 -  eauto with T1. 
 - intros n0 a H.   apply head_lt. eapply nf_inv3; eauto.  
Qed.

Lemma lt_phi0_inv1 : forall a n b a', lt_phi0 (ocons a n b) a' -> lt a  a'.
Proof. now inversion 1. Qed.

Lemma nf_intro : forall a n b, nf a -> 
                               nf b ->
                               lt_phi0 b a -> 
                               nf (ocons a n b).
Proof.  destruct 3;  eauto with T1. Qed. 

Lemma nf_intro' : forall a n b, nf a -> 
                               nf b ->
                               lt b  (ocons a 0 zero) -> 
                               nf (ocons a n b).
Proof.
  destruct b.
  - eauto with T1.
  - intros.
 apply ocons_nf; auto.

    now apply lt_phi0_inv in H1.
Qed.


Lemma lt_phi0_intro : forall a n b, nf (ocons a n b) -> 
                                    lt_phi0 b a.
Proof. intros; unfold nf in H; cbn in H.
 destruct b.
 constructor. 
 constructor; auto.
 red in H; decompose [and] H; repeat rewrite andb_true_iff in H.
 tauto. 
Qed.

Lemma nf_coeff_irrelevance : forall a b n p, nf (ocons a n b) ->
                                             nf (ocons a p b).
Proof.
 intros; apply nf_intro.
 -  eapply nf_inv1;eauto.
 -  eapply nf_inv2;eauto.
 -  eapply lt_phi0_intro;eauto.
Qed.



Lemma lt_phi0_phi0 : forall a b, lt_phi0 b a ->
                                 lt b ( phi0 a).
Proof.
  induction 1;  unfold phi0;    auto with T1. 
Qed.

Lemma lt_phi0_phi0R : forall a b, lt b  (phi0 a) -> lt_phi0 b a.
Proof.
  induction b.
  -  constructor.
     -  intro H; apply lt_phi0_inv in H. 
  constructor;auto. 
Qed.

Lemma lt_phi0_iff :
  forall a b, nf a -> nf b ->
              (lt_phi0 b a <-> forall n, nf(ocons a n b)).
Proof.
  split.
  - intros; now apply nf_intro.
  - intro; now apply lt_phi0_intro with 0.
Qed.

  
Lemma log_nf : forall a, nf a -> nf (log a).
Proof.
 destruct a;unfold log;simpl.
 -  constructor.
 - intro; eapply nf_inv1;eauto.
Qed.

Lemma nf_of_finite : forall  n b, nf (ocons zero n b) -> 
                                  b = zero.
Proof.
  intros n b H; destruct b. auto.
   apply nf_inv3 in H.
  case (not_lt_zero H);auto. 
Qed.

Lemma nf_finite : forall n:nat, nf (FS n). 
Proof.    auto with T1. Qed.


Lemma nf_omega : nf omega.
Proof.  auto with T1. Qed.


Theorem nf_phi0 : forall a, nf a -> nf (phi0 a).
Proof. compute;auto with T1. Qed.


Instance O_zero : Ord zero.
Proof. split; constructor. Defined. 

Instance O_finite (i:nat) : Ord (FS i).
Proof. split. apply nf_finite. Defined.

Instance O_omega : Ord omega.
split.
apply nf_omega.
Defined. 


Instance Omega_omega : Ord (phi0 omega).
split; apply  nf_phi0.
apply nf_omega.
Defined.

Lemma nf_tower : forall n, nf (tower n).
Proof.  induction n; simpl; unfold phi0;  auto with T1.
Qed.

(** A strong induction scheme for nf *)

Definition nf_rect : forall P : T1 -> Type,
   P zero ->
   (forall n: nat,  P (ocons zero n zero)) ->
   (forall a n b n' b', nf (ocons a n b) -> 
                        P (ocons a n b)-> 
                        lt_phi0 b' (ocons a n b) -> 
                        nf b' -> 
                        P b' ->
                         P (ocons (ocons a n b) n' b')) ->
   forall a, nf a -> P a.
Proof.
 intros P H0 Hfinite Hcons; induction a.
 -  trivial.
 -  generalize IHa1;case a1.
  +  intros IHc0 H; rewrite (nf_of_finite H);  apply Hfinite.
  +  intros c n0 c0 IHc0 H2; apply Hcons.
     *  eapply nf_inv1;eauto.
     *  apply IHc0; eapply nf_inv1; eauto.
     *  eapply lt_phi0_intro;  eauto.
     *  eapply nf_inv2;eauto.
     * apply IHa2; eapply nf_inv2;eauto.
Defined.





 Lemma compare_Eq_impl : forall a b, compare a b = Eq -> a = b.
 Proof.
   intros; generalize (compare_reflect a b); rewrite H;auto.
 Qed.
 
(* the following theorems were proved in another time and
 place ; It seems that some cleaning is necessary *)

Lemma compare_rw1 : forall a b, lt a  b -> compare a b = Lt.
Proof.
 intros c1 c2; generalize (compare_reflect c1 c2).  
 case (compare c1 c2);auto.
 - intros e H;subst c2;case (lt_irrefl H).
 - intros H1 H2;case (lt_not_gt H2);auto.
Qed.



Lemma compare_rw3 : forall a b:T1 , lt b  a  -> compare a b = Gt.
Proof.
 intros a b; generalize (compare_reflect a b);  
 case (compare a b);auto.
 - intros e H;subst b;case (lt_irrefl H).
 -  intros H1 H2;case (lt_not_gt H2);auto.
Qed.


Theorem compare_reflectR : forall a b : T1,
  (match lt_eq_lt_dec a b with
    inleft  (left _) => Lt
  | inleft  (right _) => Eq
  | inright _ => Gt
  end)
  = compare a b.
Proof.
 intros c1 c2;case (lt_eq_lt_dec c1 c2).
 -  destruct s.
  +  rewrite compare_rw1;auto.
  +  subst c1;rewrite compare_refl;auto.
 -  intro; rewrite compare_rw3;auto.
Qed.

Lemma max_le_1 : forall a b, le a  (max a b).
Proof.
 unfold max; intros a b;case (lt_eq_lt_dec a b).
 -  destruct 1.
    +  rewrite compare_rw1;auto.
        auto with T1.
    +  subst b;rewrite compare_refl.
 auto with T1. 
 - intro;rewrite compare_rw3;  auto with T1. 
Qed.

Lemma max_comm : forall a b, max a b = max b a.
Proof.
 unfold max; intros a b; case (lt_eq_lt_dec a b).
 -  destruct 1.
  +  rewrite compare_rw1;auto.
     rewrite compare_rw3;auto.
  +  subst b;auto.
 -  intro; rewrite compare_rw3 ;auto; rewrite compare_rw1;auto.
Qed.

Lemma max_dec : forall a b, {max a b = a} + {max a b = b}.
Proof.
 unfold max;  intros a b; case (lt_eq_lt_dec a b);auto.
 -  destruct 1.
  +  repeat rewrite compare_rw1;auto.
  +  subst b;repeat rewrite compare_refl;auto.
 - intro; repeat rewrite compare_rw3;auto.
Qed.

Lemma max_nf : forall a b, nf a -> nf b -> nf (max a b).
Proof.
 intros a b; case (max_dec a b); intro H;rewrite H;auto.
Qed.


Lemma max_assoc : forall alpha beta gamma,
                    max (max alpha beta) gamma =
                    max alpha (max beta gamma).
Proof.
  intros a b c; unfold max;  case (lt_eq_lt_dec a b).
  -  destruct 1.
     +  repeat (rewrite (compare_rw1 l)).
        case (lt_eq_lt_dec b c).
        * destruct 1.
          { repeat (rewrite (compare_rw1 l0)).
            assert (H : lt a  c) by ( eapply lt_trans;eauto).
            rewrite (compare_rw1 H);auto.
          }  
          subst c; repeat rewrite compare_refl.
          rewrite compare_rw1;auto.
        *  intro H; repeat (rewrite (compare_rw3 H)).
           rewrite compare_rw1;auto.
     +  subst b; repeat (rewrite (compare_refl));auto.
        case (lt_eq_lt_dec a c).
        *  destruct 1.
           {  repeat (rewrite (compare_rw1 l));auto. }
           subst c;repeat rewrite compare_refl;auto.
        *  intro H;repeat (rewrite (compare_rw3 H));auto.
           repeat rewrite compare_refl;auto.
  -  intro H; repeat rewrite (compare_rw3 H);auto.
     case (lt_eq_lt_dec a c).
     +  destruct 1.
        * rewrite (compare_rw1 l);auto.
          assert (H0: lt b  c) by (eapply lt_trans;eauto).
          repeat rewrite (compare_rw1 H);auto.
          rewrite (compare_rw1 H0);auto.
          rewrite (compare_rw1  l);auto.
        *  subst c; repeat rewrite (compare_rw1 c);auto.
           rewrite compare_refl;auto.
           repeat rewrite (compare_rw1 H);auto.
           now rewrite (compare_refl).
     +    intro H0;   rewrite (compare_rw3 H0).
          case (lt_eq_lt_dec b c).
          * destruct 1.
            repeat rewrite (compare_rw1 l);auto.
            repeat rewrite (compare_rw3 H0);auto.
            subst c;repeat rewrite (compare_refl);auto.
            repeat rewrite (compare_rw3 H);auto.
          *  intro H1;repeat rewrite (compare_rw3 H1);auto.
             repeat rewrite (compare_rw3 H);auto.
Qed.


Lemma lt_intro : forall alpha beta, compare alpha beta = Lt <-> lt alpha  beta.
Proof.
  intros a b ; rewrite <- compare_reflectR.
  destruct (lt_eq_lt_dec a b) as [[H1 | H2] | H3]; split; auto;
  try discriminate.
  - subst b; intro H; destruct (lt_irrefl H).
  - intro H4; destruct (@lt_irrefl a); now apply lt_trans with b.
Qed.

Lemma gt_intro : forall alpha beta, compare alpha beta = Gt <-> lt beta  alpha.
Proof.
 intros a b ; rewrite <- compare_reflectR.
 destruct (lt_eq_lt_dec a b) as [[H1 | H2] | H3]; split; auto; try discriminate.
 - intro H4; destruct (@lt_irrefl a); now apply lt_trans with b.
 -  subst b; intro H; destruct (lt_irrefl H).
Qed.





(********************)
(* Well foundedness *)
(********************)

(* Unfortunately, lt is NOT well founded on T1 *)

Section lt_not_well_founded.

Check omega ^ 2.
  
  (* let us build the sequence of terms :
        omega + omega + .... + omega ^ 2   *)
  Let f := (fix f (i:nat): T1 :=
            match i with 0 => phi0 2
                       | S i => ocons 1 1 (f  i)
            end).

 
 Lemma  f_decreases : forall i, lt (f (S i)) (f i).
 Proof.
  induction i; compute; auto with T1.
 Qed.

 Theorem lt_not_wf : ~ (well_founded lt).
 Proof. 
  red; intro wf; case (not_decreasing _ lt);auto.
   exists f; apply f_decreases.
 Qed.

End lt_not_well_founded.
(* end of inclusion *)


(*
 Lt's well foundedness is proved by a Tait-like  induction :

 Acc_Zero and Acc_Cons allow a structural induction using Ord_rect.
 Acc_Cons's proof requires a well_founded induction (using Acc_ind)
 on ordinals less than a, and two inductions for considering
 ordinals of the form alpha * n + beta 
*)


Definition LT := restrict nf lt.
Infix "<" := LT : cnf_scope.
Definition LE := restrict nf le.
Notation "alpha <= beta" := (LE alpha beta) : cnf_scope.

Definition Elements alpha : Ensemble T1 :=
  fun beta => beta < alpha.


Coercion Elements : T1 >-> Ensemble.

Check In T1 zero omega.



Lemma LT_proof_unicity :
  forall alpha beta (H H' : alpha < beta), H = H'.
 Proof. 
   unfold LT, restrict; intros.
   generalize (andU  H H').
 intros.    
 apply H0. 
  apply nf_proof_unicity.
  intros. 
   apply (andU H1 H'0).
  intros; apply lt_proof_unicity.
   apply nf_proof_unicity.
 Qed.

 Lemma LE_zero : forall alpha, nf alpha -> zero <= alpha.
Proof.  
  split; auto with  T1.
Qed. 


Lemma LE_refl : forall alpha, nf alpha -> alpha <= alpha. 
 split; auto with  T1.
Qed. 


Lemma LT_trans : forall a b c:T1, a < b -> b < c -> a < c.
Proof.
  unfold LT, restrict; intros a b c H H0; decompose [and] H  ;
  decompose [and] H0; repeat split; auto.
  now apply lt_trans with b.
Qed.

Theorem LE_trans : forall a b c,  a <=  b -> b <=  c ->  a  <= c.
Proof.
   unfold LE, restrict; intros a b c H H0; decompose [and] H  ;
   decompose [and] H0; repeat split; auto.
    now apply le_trans with b.     
Qed. 

Lemma LE_antisym (alpha beta : T1):  alpha <= beta ->
                                     beta <= alpha ->
                                     alpha = beta.
Proof.
  unfold LE; intros [H [H0 H1]] [H2 [H3 H4]].
  unfold le in *; generalize (compare_reflectR alpha beta).
  destruct  (lt_eq_lt_dec alpha beta).
  -  destruct s; trivial.
     + rewrite compare_rev in H3;  intro e; rewrite <- e in *; contradiction.
  - rewrite compare_rev in H3;  intro e; rewrite <- e in *; contradiction.  
 Qed. 


Lemma LT1 : forall alpha n beta, nf (ocons alpha n beta) ->
                                 zero < ocons alpha n beta.
Proof.  repeat split;auto; constructor. Qed.

Lemma LT2 : forall alpha alpha' n n' beta beta',
              nf (ocons alpha n beta) ->
              nf (ocons alpha' n' beta') ->
              alpha < alpha' ->
              ocons alpha n beta < ocons alpha' n' beta'.
Proof. repeat split; auto.  apply head_lt; auto. destruct H1. tauto. Qed.


Lemma LT3 : forall alpha  n n' beta beta',
              nf (ocons alpha n beta) ->
              nf (ocons alpha n' beta') ->
              (n < n')%nat  ->
              ocons alpha n beta < ocons alpha n' beta'.
Proof. repeat split; auto. apply coeff_lt. auto.   Qed.

Lemma LT4 : forall alpha  n  beta beta',
              nf (ocons alpha n beta) ->
              nf (ocons alpha n beta') ->
              beta < beta'  ->
              ocons alpha n beta < ocons alpha n beta'.
Proof.   repeat split; auto; apply tail_lt.  destruct H1; tauto. Qed.

Hint Resolve LT1 LT2 LT3 LT4: T1.



Lemma LT_irrefl (alpha : T1) : ~ alpha < alpha.
Proof. 
 destruct 1 as [H [H0 H1]]; destruct (lt_irrefl H0).
Qed.

Lemma LE_LT_trans : forall alpha beta gamma, alpha <= beta -> beta < gamma ->
                                             alpha < gamma.
Proof.
  intros alpha beta gamma [H1 [H2 H3]] [H4 [H5 H6]]; repeat split; auto. 
  apply le_lt_trans with beta;auto.
Qed.

Lemma LT_LE_trans (alpha beta gamma : T1) : alpha < beta ->
                                             beta <= gamma  ->
                                             alpha  < gamma.
Proof.
  intros [H [H0 H1]] [H' [H'0 H'1]]; repeat split; auto with T1.
  apply lt_le_trans with beta;auto.
Qed.

Lemma LT_zero : forall alpha, ~ alpha < zero.
Proof. intros alpha [H [H0 H1]]; inversion H0.
destruct (not_lt_zero H0).
Qed. 


Ltac not_neg H :=
  match goal with
    | H : (?i < 0)%nat |- _ => destruct (Nat.nlt_0_r i H)
    | H : ?alpha < zero |- _ => destruct (LT_zero  H)
    | H : lt ?alpha  zero |- _ => destruct (not_lt_zero  H)
  end.


Lemma nf_ocons_LT : forall (a : T1) (n : nat) (a' : T1) (n' : nat) (b : T1),
               a' < a ->
               nf a -> nf (ocons a' n' b) -> nf (ocons a n (ocons a' n' b)).
Proof.  intros; apply ocons_nf;auto;     destruct H;tauto. Qed.

Hint Resolve nf_ocons_LT: T1.

Hint Resolve nf_inv1 nf_inv2 nf_inv3 : T1.

Lemma head_LT_cons : forall alpha n beta, nf (ocons alpha n beta) ->
                                          alpha < ocons alpha n beta.
Proof.
  split; eauto with T1.
  split.
  -  apply head_lt_ocons.
  -  auto.
Qed. 

Lemma tail_LT_cons : forall alpha n beta, nf (ocons alpha n beta) ->
                                          beta < ocons alpha n beta.
Proof. 
 split;  eauto with T1.
 split; auto. 
  apply tail_lt_ocons; auto.
Qed. 

(** *  Towards transfinite induction *)

 Lemma  LT_inv : forall a n b a' n' b', 
                    ocons a n b <  ocons a' n' b' ->
                    a < a' \/
                    a = a' /\ ((n < n')%nat \/ n = n' /\   b < b').
  Proof.
    intros a n b a' n' b' H; case H.
    -   clear H;intros H (H0,H1);  case (lt_inv H0).
        +   left; split.
            *  eapply nf_inv1;eauto.
            *  split;auto;eapply nf_inv1;eauto.
        + intros [(e,i)|(e,(e',i))]. 
          *   auto. 
          *   right;split;auto.
              right;split;auto.
              split.
              eapply nf_inv2;eauto. 
              split;[auto|eapply nf_inv2;eauto].
  Qed.

  Inductive LT_cases (a  b : T1) (n :nat) (a' b':T1) (n':nat) : Type :=
    LT_left (H : a < a')
  | LT_middle (H : a = a')(H1 : (n < n')%nat)
  | LT_right (H : a = a')(H1 : n = n')(H2 : b < b').


  Lemma  LT_inv_strong : forall a  b n a'  b' n', 
                    ocons a n b <  ocons a' n' b' -> LT_cases a b n a'  b' n'. 
  Proof.
    intros a  b n a' b' n' H. case H.
    -   clear H;intros H (H0,H1);  case (lt_inv_strong H0).
        +   constructor 1. split.
            *  eapply nf_inv1;eauto.
            *  split;auto;eapply nf_inv1;eauto.
        + intros; subst.  constructor 2; auto. 
        + constructor 3 ; auto. split; eauto with T1. 
  Qed.

  Lemma remove_first_sumand : 
  forall a n b  b', 
    ocons a n b <  ocons a n b' -> b < b'.
  Proof.
    intros a n b b' H; apply LT_inv in H.
    destruct H.
    - destruct H as [_ [H _] ].
      destruct (lt_irrefl H).    
    - destruct H as [_ [H | [ _ H]]]; auto.
      destruct (Nat.lt_irrefl _ H).   
   Qed.


  Lemma LT_ocons_0 : forall a n b a' b', ocons a n b <  ocons a' 0 b' ->
                                   a < a' \/ n = 0 /\ a = a' /\  b < b'.
  Proof.
    intros a n b x c H; case (LT_inv H).
    - now left.
    -  intros [e H0];subst x; case H0;intro H1. 
       +   not_neg  H1.
       +   case H1;right;auto.
  Qed.


(** ** The order < on Cantor Normal forms is well  founded *)

Module Direct_proof.
Section well_foundedness_proof.
  
  Hint Unfold restrict LT.

  Lemma Acc_zero : Acc LT  zero.
  Proof.
    split; intros y (H1, (H2,H3)); not_neg H2.
  Qed.


 
 
 Section First_attempt.
 Lemma wf_LT : well_founded_restriction _ nf lt.
 Proof.
  red.
  induction a as [| a IHa n b IHb].
  - split.
    inversion 1.
    destruct H2 as [H3 _];not_neg H3.
  -  split.
     intros y [H1 [H2 H3]].
 (*
  a : T1
  n : nat
  b : T1
  IHa : nf a -> Acc (restrict nf lt) a
  IHb : nf b -> Acc (restrict nf lt) b
  H : nf (ocons a n b)
  y : T1
  H1 : nf y
  H2 : y < ocons a n b
  H3 : nf (ocons a n b)
=============================================
   Acc (restrict nf lt) y
  *)
  Abort.    

  End First_attempt.
 
 (** Let us consider a stronger predicate :
     a satisfies Acc_strong if any well formed  ordinal of the form
     ocons a n b is accessible
  *)
 
 Let Acc_strong (alpha:T1) :=
      forall n beta, 
        nf (ocons alpha n beta) -> Acc LT (ocons alpha  n beta).

 Lemma Acc_strong_stronger : forall alpha, 
     nf alpha -> Acc_strong  alpha -> Acc LT  alpha.
 Proof.
  intros alpha H H0; apply acc_imp with (phi0 alpha).
  - repeat split; trivial.
    + now apply lt_a_phi0_a.
  -  apply H0;  now apply single_nf.
Qed.

         
 Lemma Acc_implies_Acc_strong : forall alpha, 
   Acc LT  alpha -> Acc_strong alpha.
 Proof.
  (* main induction (on a's accessibility)   *)
  unfold Acc_strong; intros alpha Aalpha; pattern alpha.
  eapply Acc_ind with (R:= LT);[| assumption].
  clear alpha Aalpha; intros alpha Aalpha IHalpha. 

  (*   subgoal 1 

  a : T1
  Aa : forall y : T1, LT y a -> Acc LT y
  IHa : forall y : T1,
       LT y a ->
       forall (n : nat) (b : T1),
       nf (ocons y n b) -> Acc LT (ocons y n b)
  ============================
   forall (n : nat) (b : T1),
   nf (ocons a n b) -> Acc LT (ocons a n b)
*)
 (* end of subgoal1 *)

  (** for any n and b, such that ocons a n b is well formed
    b is accessible *)


  assert(beta_Acc : 
       forall beta, lt_phi0 beta alpha -> nf alpha -> nf beta 
-> Acc LT beta).
 
 (* Proof of beta_Acc

   Since beta is  less than omega^alpha, 
beta  is of the form omega^alpha'*(p+1)+beta' where
   LT alpha' alpha , so the inductive hypothesis IHalpha can be used 
 *)
 (******************)

  {  intros b H H' H'';  assert (H0 : LT b (phi0 alpha)).
     { repeat split;auto.
       apply lt_phi0_phi0; auto.
       
     }
 
     generalize H0;pattern b; case b.
     - intro;apply Acc_zero.
     -  intros t n t0 H1; case H1;  destruct 2;  unfold phi0 in H3;
        case (lt_inv H3).
        + intro H5;generalize H2;case n.
          { inversion 1.
            - intros; apply IHalpha.
              + split.
                * apply nf_inv1 in H2. auto. 
                * split;auto.
              + auto. 
          }
  intros;apply IHalpha.
  split;auto.
  eapply nf_inv1;eauto.
  auto.

  + destruct 1.
  case H5;intros H6 H7; not_neg H7.
  case H5;intros _ (_,H6);not_neg H6.
  }
  
 (* end of proof of beta_Acc *)
 (*************************)


 (*
  a : T1
  Aa : forall y : T1, LT y a -> Acc LT y
  IHa : forall y : T1,
       LT y a ->
       forall (n : nat) (b : T1),
       nf (ocons y n b) -> Acc LT (ocons y n b)
  b_Acc : forall b : T1, lt_phi0 b a -> nf a -> nf b -> Acc LT b
  ============================
   forall (n : nat) (b : T1),
   nf (ocons a n b) -> Acc LT (ocons a n b)
*)
  
  (* we can now use a nested induction on n (Peano induction)
    then on b (well_founded induction using b_Acc) *)

  induction n.
  -    (* n=0 let's use b's accessibility for doing an induction on b *)
    intros b Hb; assert (H : Acc LT  b).
    {  apply beta_Acc.
      -  eapply lt_phi0_intro; eauto.
      -  eapply nf_inv1;eauto.
      -  eapply nf_inv2;eauto.
    }
        
  (* let's prove that every predecessor of (ocons a 0 b) is accessible *)
  { 
    pattern b;eapply Acc_ind;[|eexact H].
    intros; split; intro y; case y.
     - intro; apply Acc_zero.
     -   intros c n c0 H3; case (LT_ocons_0 H3).
         intro; unfold LT; case H3;auto.
         intros (e,(e',r)); subst n c; auto.
  }
  (**
  induction step for (S n)
 
  An LT_antecedent of (ocons a (S n) b) is of one of the following forms :
    - ocons a' p b' : use IHa
    - ocons a p b' where p <= n : use IHn 
    - ocons a (S n) b' where b'<b : use b's accessibility
 *)
  -
    { intros b H; generalize H; pattern b; eapply Acc_ind with (R:= LT).
      - split; intro y;pattern y; case y.
        intro;apply Acc_zero.
        intros c n0 c0 H3.
        (*
         H3 : LT (ocons c n0 c0) (ocons a (S n) x)
         ============================
         Acc LT (ocons c n0 c0)
         *)
        case (LT_inv H3).
        + (* c < alpha *)
           intro; apply IHalpha;auto.
           case H3; auto. 
        + (* c = alpha and n0 < S n *)
          intros (e,[i|(e',r')]).
          assert (disj:n0 = n \/ (n0 < n)%nat).
          { inversion i;auto with arith. }
          case disj;intro H4.
          * subst n c; apply IHn.
            case H3;auto.
          * apply Acc_inv with (ocons alpha  n zero).
            apply IHn.
             apply single_nf;auto. 

            subst c; eapply nf_inv1;eauto.
            subst c; split. 
            case H3;auto. 
            split. 
              auto with T1.
    apply single_nf;auto.

            case H3;intros;eapply nf_inv1;eauto.
            (* c = a, n0 = S n and c0 < x *)
           *  subst n0 c;apply H1; auto.
              case H3;auto.
      - apply beta_Acc. 
        + eapply lt_phi0_intro;eauto.
        + eapply nf_inv1;eauto.
        + eapply nf_inv2;eauto.
}

Qed.

 (* A (last) structural induction *)

Theorem nf_Acc : forall alpha, nf alpha -> Acc LT  alpha.
Proof.
 induction alpha.
 -  intro; apply Acc_zero.
 -  intros; eapply Acc_implies_Acc_strong;auto.
    apply IHalpha1;eauto.
    apply nf_inv1 in H; auto. 
Defined.

End well_foundedness_proof.
End Direct_proof.






Definition nf_Acc := Direct_proof.nf_Acc.


Corollary nf_Wf : well_founded_restriction _ nf lt.
Proof.  red; intros; now apply nf_Acc. Qed.


Corollary T1_wf : well_founded LT.
Proof.
 intros alpha.
 case_eq(nf_b alpha).
  intro.
   now generalize (nf_Wf H).
   intros  H .
   split. 
   destruct 1 as [H1 [H2 H3]].
  red in H3; rewrite H in H3; discriminate.    
 Qed. 


Definition transfinite_recursor :
 forall (P:T1 -> Type),
   (forall x:T1, 
                   (forall y:T1, nf x -> nf y ->  lt y  x -> P y) -> P x) ->
    forall a, P a.
Proof.
 intros; apply well_founded_induction_type with LT.
 -  exact T1_wf;auto.
 - intros. apply X. intros; apply X0. repeat split;auto. 

Defined.

(** Transfinite induction for a subset of T1 *)
(* useful ???? *)
Definition transfinite_induction_subset :
  forall (P : T1 -> Type) (E : Ensemble T1),
      (forall x:T1, E x -> nf x -> 
           (forall y:T1, E y -> nf y ->  lt y  x -> P y) -> P x) ->
   forall a, nf a -> E a -> P a.
Proof.
 intros P E X a  H  H0;
  eapply well_founded_restriction_rect with
  (R:=lt)(E:=fun a => nf a /\ E a)(a:=a).
 - red;intros a0 H1; apply Acc_incl with LT. 
  + red;intros; unfold restrict in *;  red. red. tauto.
  +  apply nf_Acc; case H1;auto.
 -  destruct 1; intros; eapply X; eauto.
    intros; apply X0.
    + split;auto.
    + assumption. 
 - split;auto.
Defined.

Definition transfinite_recursor_LT :
 forall (P:T1 -> Type),
   (forall x:T1, 
                   (forall y:T1, y  < x -> P y) -> P x) ->
    forall a,  P a.
Proof.
 intros; apply transfinite_recursor; auto.
 intros;  apply X;auto.
 destruct 1 as [H1 [H2 H3]];auto.
Defined.


Import  Direct_proof.

Ltac transfinite_induction alpha :=
  pattern alpha; apply transfinite_recursor.

Ltac transfinite_induction_LT a :=
  pattern a; apply transfinite_recursor_LT. 

(** ***  Properties of successor *)

Lemma succ_lt_phi0 : forall c a n b, lt_phi0 c (ocons a n b) -> 
                                 lt_phi0 (succ c) (ocons a n b).
Proof.
 induction c.
 -  simpl; repeat constructor.
 - simpl; case c1.
  +  repeat constructor.
  + intros t n0 t0 a n1 b H; inversion_clear H; constructor; auto.
Qed.

Lemma succ_nf : forall a, nf a -> nf (succ a).
Proof.
  induction a.
  -  simpl;repeat constructor; auto  with arith.
  -  simpl; generalize IHa1 ; case a1.
     +  simpl;repeat constructor; auto.
     + intros c n0 c0 H H0;  apply nf_intro.
       * eapply nf_inv1; eauto.
       *  apply IHa2;  eapply nf_inv2 ; eauto.
       * apply succ_lt_phi0; eapply lt_phi0_intro; eauto.
Qed.

Instance O_succ alpha {Oalpha: Ord alpha} : Ord (succ alpha). 
split.
  apply succ_nf; apply Oalpha.
Defined. 



Lemma succ_not_zero : forall alpha, succ alpha <> zero.
Proof.
induction alpha.
 -  discriminate.
 - cbn; destruct alpha1; discriminate.
Qed.


(* properties of lt and le *)

Lemma lt_succ :  forall a, lt a  (succ a).
Proof.
  induction a; cbn; auto with T1.
  - destruct a1; cbn ; auto with T1.   
Qed.


Definition zero_succ_limit_dec :
  forall alpha, nf alpha ->
                ({alpha = zero} + {is_limit alpha }) + 
                {beta : T1 |  nf beta /\ alpha = succ beta} .
Proof with eauto with T1.
induction alpha as [| gamma Hgamma n beta Hbeta].
- intros _; left; now left.                                           
- intro H;  destruct gamma.
  + assert (beta = zero). eapply nf_of_finite; eauto. 
   
    subst beta; right.
    destruct n.
    { exists zero. split; auto with T1. }
    { exists (FS n).    split.
     apply nf_finite.
     reflexivity. 
   }
  +  destruct Hbeta ...
     *   destruct s.
         subst beta.
         left;right.
         reflexivity.
         left;right.
         simpl; destruct beta.
         reflexivity.
         auto.
     *  destruct s as [beta0 [H1 H2]]; subst beta.
     right;  exists (ocons (ocons gamma1 n0 gamma2) n beta0).
     split.
     { apply nf_intro; auto.
       inversion H; auto.
       apply nf_inv1 in H. auto. 
       apply lt_phi0_phi0R; apply lt_trans with (succ beta0);auto.
       apply lt_succ.
       apply lt_phi0_phi0;  eapply lt_phi0_intro; eauto.
     }
     reflexivity.
Defined. 

Lemma phi0_log : forall a, lt a  (phi0 (succ (log a))).
Proof.
 destruct a.
-  constructor.
- simpl; unfold phi0.  apply head_lt; apply lt_succ.
Qed.
 

(* properties of addition *)


Lemma plus_zero : forall a, zero + a  = a.
Proof.  simpl; intro a; case a; auto. Qed.

Lemma plus_a_zero : forall a,  a + zero = a.
Proof.
 intro a; case a;simpl.
 -  trivial.
 - intro a0;case a0;simpl;auto.
Qed.


Lemma plus_fin_omega : forall n:nat ,  FS n + omega = omega.
Proof. simpl;auto.  Qed.

Notation One := (FS 0).

Lemma succ_is_plus_one: forall a, nf a -> succ a = a + One.
Proof.
 induction a as [|a IHa n b IHb].
 - reflexivity. 
 - destruct  a.
   + simpl; now rewrite <- plus_n_O.
   +  intros H; simpl. 
      rewrite IHb;[ reflexivity  | eapply nf_inv2;eauto].
Qed.  


(* plus and lt *)


Lemma plus_ocons_ocons_rw1 : forall a n b a' n' b', 
     lt a  a' ->
     ocons a n b + ocons a' n' b' = ocons a' n' b'.
Proof.
 simpl; destruct a.
 -  destruct a'. 
   +  inversion 1.
   +  intros; rewrite compare_rw1; auto with T1.
 -  destruct a'; inversion 1.
    + now rewrite compare_rw1.
    
Qed.


Lemma plus_ocons_ocons_rw2 : forall a n b n' b', 
      nf (ocons a n b) ->
      nf (ocons a n' b') ->
      plus (ocons a n b) (ocons a n' b')= 
      ocons a (S (n + n') ) b'.  
Proof.
 simpl; destruct a.
 - intros n b n' b' H H0; now rewrite (nf_of_finite H0).
 - now  rewrite compare_refl.
Qed.


Lemma plus_ocons_ocons_rw3 : forall a n b a' n' b', 
                               lt a'  a ->
                               nf (ocons a n b) ->
                               nf (ocons a' n' b') ->
                               ocons a n b + ocons a' n' b'= 
                               ocons a n (b + (ocons a' n' b')).
Proof.
  simpl;intros a n b a' n' b' H H0 H1; rewrite compare_rw3.
  - destruct a; intros; auto.
  -  assumption.
Qed.



Lemma ap_plus : forall a, ap a ->
    forall b c, nf b -> nf c -> lt b  a -> lt c  a -> lt (b + c)  a.
Proof.
 destruct 1. 
 intro b; elim b. 
 - intro c; elim c;intros.
  +  simpl; auto with T1. 
  +  simpl;auto.
 -  intros c H0 n t H c0 H1 H2 H3 H4;  generalize c0 H2 H4.
    destruct c1.
    + rewrite (plus_a_zero);auto.
    + intros H5 H6;  case (lt_eq_lt_dec c c1_1).
      *  destruct 1.
         { rewrite (plus_ocons_ocons_rw1 n t n0 c1_2 l); auto. }
         subst c1_1; rewrite (plus_ocons_ocons_rw2 H1 H5).
         apply lt_phi0_inv  in H6.
         auto with T1.
      *  intro H7; rewrite plus_ocons_ocons_rw3;auto.
          apply lt_phi0_inv  in H3.
          auto with T1. 
Qed.


Lemma ap_plusR : forall c, nf c -> c <> zero ->
                  (forall a b, nf a -> nf b ->  lt a  c ->
                          lt b  c -> lt (a + b) c)  ->
                  ap c.
 destruct c as [|c1 n c2].
 -  now destruct 2. 
 - case c2.
   + case n.
   *  constructor.
   *  intros n0 H H0 H1;  generalize (H1 (ocons c1 0 zero) (ocons c1 n0 zero)).
      clear H1;intros H1.
      assert (H2 : nf (ocons c1 0 zero)) by 
       (eapply nf_coeff_irrelevance;eauto).
      assert (H3 : nf (ocons c1 n0 zero)) by 
          (eapply nf_coeff_irrelevance;eauto).
      assert  (H4 : lt (ocons c1 0 zero) (ocons c1 (S n0) zero)).
      {auto with T1 arith. } 
      assert (H5 : lt (ocons c1 n0 zero)  (ocons c1 (S n0) zero )) by
          (auto with T1 arith  ).
      generalize (H1 H2 H3 H4 H5); intro H6.
      rewrite plus_ocons_ocons_rw2 in H6; simpl in H6.
      case (lt_irrefl H6);auto.
      auto.
      auto.
 + intros t n0 t0 H H0 H1.
   assert (H2 : nf (ocons c1 n zero)). {
    
     apply  single_nf. 
     eapply nf_inv1;eauto.
   } 
   assert (H3: nf (ocons t n0 t0)).
   {  eapply nf_inv2;eauto. }
   assert (H4 : lt (ocons c1 n zero) (ocons c1 n (ocons t n0 t0))).
   { apply tail_lt; auto with T1. }
   assert (H5 : lt (ocons t n0 t0) (ocons c1 n (ocons t n0 t0))).
   {  apply nf_inv3 in H.  auto with T1. }
   generalize (H1 _ _ H2 H3 H4 H5).
   clear H1 H4 H5;intro H1; rewrite plus_ocons_ocons_rw3 in H1; auto.
  *   simpl in H1; case (lt_irrefl H1); auto.
  *   apply nf_inv3 in H; auto. 
Qed. 



(** Technical lemma for proving plus_nf *)

Lemma plus_nf0 : forall a, nf a -> forall b c, lt b  (phi0 a) ->
                                    lt c  (phi0 a) ->
                                    nf b -> nf c ->
                                    nf (b + c).
Proof.
 intros a . pattern a;  transfinite_induction a.
 intros x Cx Hx.
 destruct b; destruct c.
 -  simpl;constructor.
 -  simpl;auto.
 - intros;rewrite plus_a_zero; auto.
 -  intros; case (lt_eq_lt_dec b1 c1).
  +  destruct 1.
    *  rewrite plus_ocons_ocons_rw1; auto. 
    *  subst c1; rewrite plus_ocons_ocons_rw2; auto.

  +  intro; rewrite plus_ocons_ocons_rw3;auto.
     apply nf_intro.
     * eapply nf_inv1;eauto.
     * eapply Cx with b1.
       eapply nf_inv1;eauto.
       eauto with T1.
       apply lt_phi0_inv in H;auto.
       eauto with T1.
       apply lt_phi0_phi0. 
       eapply lt_phi0_intro with n; auto. 
       apply lt_phi0_phi0. 
       eapply lt_phi0_intro with n0; auto. 
       apply ocons_nf; auto. 
       eauto with T1.
       eauto with T1.
       auto. 
     *  apply lt_phi0_phi0R; apply ap_plus.
        unfold phi0;constructor.
        eapply nf_inv2;eauto.
        auto.
        apply lt_phi0_phi0.
        eapply lt_phi0_intro; eauto.
        unfold phi0;   auto with T1.
  Unshelve. 
 exact 0.
Qed.


Lemma plus_nf : forall a,  nf a -> forall b, nf b -> nf (a + b).
Proof.
  intros a Ha b Hb; case (lt_eq_lt_dec a b).
  - destruct 1.
    + (** a < b *)  apply plus_nf0 with b; auto.
                    apply lt_trans with b; auto.
                    apply lt_a_phi0_a.
                    apply lt_a_phi0_a.
    + (** a = b *)  subst b;
      apply plus_nf0 with a; try assumption; 
                    apply lt_a_phi0_a.
                    
  - (** b < a *)
    intro;  apply plus_nf0 with a; auto.
    apply lt_a_phi0_a.
    apply lt_trans with a;auto.
    apply lt_a_phi0_a.
Qed.


Lemma omega_term_plus_rw:
  forall a n b, 
    nf (ocons a n b) ->
    omega_term a n + b = ocons  a n b.
Proof.
simpl; destruct a.
-  intros n b H; rewrite (nf_of_finite H); auto.
-  destruct b.
 + reflexivity.
 +  intro H.
    apply nf_inv3 in H. 
   red in H.
 unfold lt_b in H. 
 rewrite compare_rev . 
 destruct  (compare b1 (ocons a1 n a2)); try discriminate. 
 reflexivity. 
Qed.

Lemma plus_is_zero : forall a b, nf a -> nf b ->
                               a + b  = zero -> a = zero /\  b = zero.
Proof.
 destruct a;destruct b.
 - now split.  
 - cbn;  discriminate 3.
 -  intro;rewrite plus_a_zero.
  +  discriminate 2.
 -  simpl; case a1;case b1.
    +  discriminate 3.
    + intros c n1 c0 H H0; rewrite compare_rw1.
      *  discriminate 1.
      *  constructor.
     +  intros c n1 c0 H H0; rewrite compare_rw3.
      *  discriminate 1.
      *  constructor.
     + intros c n1 c0 c3 n2 c4 H H0;
       case (compare  (ocons c3 n2 c4) (ocons c n1 c0));discriminate 1.
Qed.

Lemma plus_not_strict_mono_l :
  exists a b c:T1, nf a /\ nf b /\ nf c /\ lt a  b /\ a + c = b + c.
Proof.
  exists  One, (FS 1), omega;   cbn;  repeat split; auto with T1.
Qed. 


Lemma succ_strict_mono : forall a b, 
                    lt a  b -> nf a -> nf b -> 
                    lt (succ a) (succ b).
Proof.
    induction a; destruct b. 
    - intros; destruct (not_lt_zero H).
    -  cbn. destruct b1.
         intros.   apply coeff_lt; auto with arith. 
     intros _ _ H.
   apply head_lt; auto with T1. 
    - 
       intro H; destruct (not_lt_zero H).
 - cbn. 
    destruct a1.
      destruct b1.
      intros.
 destruct (lt_inv H).
 destruct (not_lt_zero H2).  
 destruct H2.
 destruct H2; auto with arith T1.
 decompose [and] H2.
assert (b2 = zero).      
      eapply nf_of_finite; eauto. 
 subst b2. 
destruct (not_lt_zero H6).  
intros; apply head_lt. 
auto with T1.
intros.
destruct b1.
destruct (lt_inv H).
destruct (not_lt_zero H2).  
destruct H2.
 destruct H2; discriminate. 
 destruct H2; discriminate.
 
destruct (lt_inv H).
  auto with T1.
destruct H2.
destruct H2.
rewrite H2.
 auto with T1.
 decompose [and] H2.
rewrite H5, H3.
apply tail_lt. 
apply IHa2 ; auto. 
 eauto with T1.
 eauto with T1. 
Qed. 

Lemma succ_strict_mono_LT : forall alpha beta, 
                              alpha < beta -> succ alpha < succ beta.
Proof.  
  intros alpha beta H; destruct H as [H [H0 H1]]; repeat split;auto.
  - now apply succ_nf.
  - now apply succ_strict_mono.   
  - now apply succ_nf.
Qed.



   
Lemma phi0_mono_strict : forall a b, lt a  b -> lt (phi0 a) (phi0 b).
Proof.  unfold phi0; auto with T1.  Qed.

Lemma LT_nf_l : forall alpha beta , alpha < beta -> nf alpha.
Proof. now  destruct 1.
Qed.

Lemma LT_nf_r : forall alpha beta , alpha < beta -> nf beta.
Proof. now  destruct 1 as [_ [_ H]]; assumption.
Qed.

Hint Resolve LT_nf_r LT_nf_l : T1.


Lemma phi0_mono_strict_LT : forall alpha beta, alpha < beta ->
                                               phi0 alpha < phi0 beta.
Proof. intros.  apply LT2; eauto with T1.
Qed.


Lemma phi0_mono : forall a b, le a  b  -> le (phi0 a) ( phi0 b).
Proof.
  intros a b H; destruct (le_lt_or_eq _ _ H).
 -  subst b; auto with T1. 
 -  apply lt_le_incl; unfold phi0; auto with T1.
Qed.


   

Lemma succ_mono : forall a b, nf a -> nf b ->
   le a  b -> le (succ a)  (succ b).
Proof.
   intros a b H H0 H1; destruct (le_lt_or_eq _ _ H1).
 -  subst a; auto with T1.
 - apply lt_le_incl;  now  apply succ_strict_mono.
Qed.




 
Lemma lt_succ_le_R : forall a,  nf a -> forall b, nf b ->
                      le (succ a) b ->  lt a  b .
Proof.
  intros c Hc; elim Hc using nf_rect.
  - intros b H H0.  destruct b. destruct H0.   auto with T1. 
  - intros n b H H0.  destruct (le_lt_or_eq _ _ H0).
     subst. 
     cbn. 
     auto with T1 arith. 
  apply lt_trans with (succ (FS n)); auto with T1.
    cbn; auto with T1 arith. 
 - intros a  n b n' b' H H0 H1 H2 H3 b0 H4 H5.
    destruct (le_lt_or_eq _ _ H5).
    subst b0.
    apply lt_succ;auto.
    apply lt_trans with (succ (ocons (ocons a n b) n' b')).
    apply lt_succ;auto.
    auto. 
Qed.

Lemma LT_succ_LE_R : forall alpha  beta, nf alpha ->
                                         succ alpha  <= beta -> alpha  < beta. 
Proof. 
 intros. 
  destruct H0 as [H1 [H2 H3]].
 repeat split; auto. 
  apply  lt_succ_le_R; auto. 
Qed.


Lemma lt_succ_le_2 : forall a,  nf a -> forall b, nf b ->
                      lt a  (succ b)  ->  le a  b.
Proof.
  intros c Hc; elim Hc using nf_rect.
  - intros;apply zero_le.
  -  intros n b H; case b; cbn.
     + intros H0; generalize (lt_inv_nb  H0).
      intros [H2 | [_ H2]]; inversion H2.
     +   destruct t.
      * intros.  destruct (lt_inv_nb H0).
         assert (H8:n = n0 \/ (n < n0)%nat) by omega.
         destruct H8.
        destruct t.    
 subst n0. 
 auto with T1.
 subst n0; auto with T1.
  auto with T1.
 destruct a.
 
destruct (not_lt_zero H2).  
* 
   intros.
apply lt_le_incl.
 auto with T1.
  -  destruct b0.
 intros. 
  cbn in H5. 
destruct (lt_inv  H5).
destruct (not_lt_zero H6).
  destruct H6. 
 destruct H6; discriminate. 
  destruct H6; discriminate. 
  intros. 
 cbn in H5. 
 destruct b0_1.
destruct (lt_inv  H5).
 destruct (not_lt_zero H6).
 destruct H6; discriminate. 
  destruct (lt_inv  H5).
apply lt_le_incl;auto  with T1.

destruct H6.
  destruct H6.
 injection H6;intros. 
subst. 
 auto with T1. 
 decompose [and] H6.
  injection H7; intros; subst.  
apply le_tail.
apply H3. 
eauto with T1. 
auto. 
Qed.  



Lemma lt_succ_le : forall a,  nf a -> forall b, nf b ->
                     lt a  b -> le (succ a)  b.
Proof.
 induction a.
 -  intros H0 c'; case c'.
   +  intros H1 H2; destruct (not_lt_zero H2). 
   +  destruct t.
    *  destruct n; intros t H H1.
      generalize (nf_of_finite H).
      intro; subst.
       now compute. 
       now compute.
    *  intros; simpl. now  compute. 
 -  destruct b.
    intros H0 H1; destruct (not_lt_zero H1).
    intros H0 H1; destruct (lt_inv H1).
    destruct a1;
      simpl.
     apply lt_le_incl.    
     auto with T1.
     apply lt_le_incl.    
     auto with T1.
     destruct H2.
     destruct H2; subst b1.
     simpl succ. 
     destruct a1.
      generalize (nf_of_finite H0).
     intro; subst. 
     generalize (lt_le_S _ _ H3).

     intro H2; destruct (Lt.le_lt_or_eq _ _ H2).
   auto with T1.
   subst; auto with T1.
   auto with T1.
   decompose [and] H2; subst. 
   clear H2.
    simpl succ. 
    destruct b1. 
    generalize (nf_of_finite H).
   intro; subst. 
   destruct (lt_inv H1). 
    destruct (not_lt_zero H2).
    destruct H2.
   destruct H2.
   destruct (Nat.lt_irrefl _ H3).
   decompose [and] H2; subst. 
   generalize (nf_of_finite H0).
   intro; subst. 
   destruct (not_lt_zero H7).
 apply le_tail. 
  eauto with T1.
Qed.

Lemma LT_succ_LE : forall alpha beta ,
                       alpha < beta -> succ alpha <= beta.
Proof.   
 intros. 
  destruct H as [H1 [H2 H3]]; repeat split; auto. 
  apply succ_nf;auto. 
  apply  lt_succ_le; auto. 
Qed. 


Lemma succ_strict_monoR :
  forall alpha beta, nf alpha -> nf beta -> lt (succ alpha) (succ beta) ->
                     lt alpha beta.
  Proof.
intros alpha beta Halpha Hbeta H.
 apply lt_le_trans with (succ alpha).
 - now apply lt_succ.
 - apply lt_succ_le_2; auto.
   now apply succ_nf.
Qed.


Lemma succ_monomorphism : forall alpha (H:nf alpha) beta (H' : nf beta),
                             lt alpha beta <-> lt (succ alpha) (succ beta).
Proof.
  split.
  - intro; now apply succ_strict_mono.
  -  intro; now apply succ_strict_monoR.
Qed.


Lemma succ_injective : forall alpha (H:nf alpha) beta (H' : nf beta),
                         succ alpha = succ beta -> alpha = beta.
Proof.
  intros.
  destruct (lt_eq_lt_dec alpha beta); auto.
  destruct s; auto. 
  all:  rewrite succ_monomorphism  in l; auto;  rewrite H0 in l;
  now destruct (lt_irrefl l).
Qed. 


Lemma plus_left_absorb : 
  forall a n b c, lt c  (phi0 a) ->
                  c + ocons a n b = ocons a n b.
Proof.
  cbn;  destruct c.
  -   reflexivity.
  -   cbn. intro H; destruct (lt_inv H). 
      red in H0; unfold lt_b in H0 .
      destruct (compare c1 a); auto; try discriminate.
      destruct H0 as [[H0 H1] |[H0 [H1 H2]]].
       destruct (Nat.nlt_0_r _ H1).
      destruct (not_lt_zero H2).
Qed.


(* About minus *)

Lemma minus_lt : forall a b, lt a  b -> a - b = zero.
Proof.
  induction a; destruct b.
  - reflexivity. 
  -  reflexivity.
  - intro H; destruct (not_lt_zero H).
  - intro H; destruct (lt_inv H).
 + cbn.
    destruct a1.
    * destruct b1.
      destruct (not_lt_zero H0).
      auto. 
  * red in H0; unfold lt_b in H0; destruct (compare (ocons a1_1 n1 a1_2) b1).
   discriminate.
  auto. 
  discriminate. 
 +  destruct H0 as [[H0 H1] |[H0 [H1 H2]]].
    cbn.   
     subst b1.
     destruct a1.
    destruct (le_lt_dec n n0); auto. 
     destruct (Lt.lt_irrefl n);eauto with arith.
   rewrite compare_refl.
   destruct (Compare_dec.lt_eq_lt_dec n n0).
   destruct s; auto.
     subst n0; destruct   (Lt.lt_irrefl n); auto. 
   destruct (Lt.lt_irrefl n);eauto with arith.
  cbn. 
   destruct a1.
  destruct b1.
   destruct (le_lt_dec n n0); auto. 
   subst n0; destruct (Lt.lt_irrefl n ); auto. 
   auto. 
   subst. 
 rewrite compare_refl.
  destruct (Compare_dec.lt_eq_lt_dec n0 n0); auto. 
 destruct s. 
 auto. 
 auto. 
  destruct (Lt.lt_irrefl n0 ); auto.
Qed.


Lemma minus_a_a : forall a, a - a = zero.
Proof.
 induction a;simpl;auto.
 -  case a1.
  +  case (le_lt_dec n n).
     * auto.
     *  intros; absurd (n < n)%nat; auto with arith.
  + case (Compare_dec.lt_eq_lt_dec n n ).
   *  destruct s.
      absurd (n < n)%nat;auto with arith.
      intros;rewrite compare_refl.
      rewrite IHa2;auto.
   * intro; absurd (n<n)%nat;auto with arith.
Qed.
 
 Lemma minus_le : forall a b, le a b -> a - b = zero.
Proof.
 intros a b H; destruct (le_lt_or_eq _ _ H). 
 -  subst b; apply minus_a_a.
 -  apply minus_lt;auto.
Qed.

Lemma mult_fin_omega : forall n: nat, 
                          FS n * omega = omega.
Proof.  now cbn. Qed.

Lemma phi0_plus_mult : forall a b, nf a -> nf b -> 
                       phi0 (a + b) = phi0 a * phi0 b.
Proof.
 cbn; intro a; case a.
 -  intro b; case b;reflexivity. 
 -  intros until b;case b;simpl.
  +  case t;cbn;auto.
  +  reflexivity. 
Qed.


(** operations on T1 extend operations on nat *)

Lemma succ_compat : forall n:nat, succ (FS n) = FS (S n).
Proof.  reflexivity.  Qed.


Lemma plus_compat: forall n p,  FS n +  FS p = FS (S n + p).
Proof.
  reflexivity. 
Qed.

Lemma mult_compat : forall n p,  FS n * FS p =
                            FS (n * p + n + p)%nat.
Proof.
 cbn;   intros; f_equal;   ring. 
Qed.

(*
Lemma exp_F_compat : 
 forall p n : nat , exp_F  (FS n) p =
             F (n ^ p)%nat.
Proof.
 induction p;simpl.
 -  trivial.
 - intro n; rewrite (IHp n); now  rewrite mult_compat.
Qed.
*)

(*
Lemma exp_compat : forall p n, F n ^ F p =
                          F (n ^  p)%nat.
Proof.
 induction p.
 -  destruct n;simpl.
 + reflexivity.
 +  destruct n;auto.
 -  destruct n;simpl;auto.
    case n.
  +  cbn ;  rewrite power_of_1; reflexivity.
  +  intros;
     replace (ocons zero (S n0) zero) with (F (S (S n0))).
     { rewrite exp_F_compat.
       rewrite mult_compat.
       cbn; f_equal;  ring.
     }       
     reflexivity.
Qed.
 *)

Lemma mult_0_a : forall a, zero * a  = zero.
Proof.  induction a;simpl;auto. Qed.

Lemma mult_a_0 : forall a, a * zero = zero.
Proof.
 induction a as [|a1 IHa1 n a2 IHa2];[reflexivity| now destruct a1].  
Qed.

Lemma mult_1_a : forall a, nf a ->  One * a = a.
Proof.
  induction a. 
  - reflexivity.
  - cbn; simpl in IHa2.
    intro H; case_eq a1.
    + intro; subst a1;  rewrite (nf_of_finite H);
      now rewrite <- (plus_n_O n).
 + intros t n0 t0 H0; subst a1; rewrite IHa2.
   * reflexivity.
   *  eapply nf_inv2;eauto.
Qed.

Lemma mult_a_1 : forall a, nf a -> a * One  = a.
  Proof.
 induction a.
 - reflexivity.
 - cbn;  simpl in IHa2.
   intro H;  case_eq a1.
   intro; subst a1; rewrite mult_1_r.
   now rewrite (nf_of_finite H).
   intros; subst a1; now  rewrite mult_1_r.
Qed.

Lemma exp_fin_omega : forall n, FS (S n) ^ omega = omega.
Proof.  reflexivity. Qed.

(** The next three lemmas express the consistency between
     the intuitive meaning given to the constructor cons and
     its derivates : phi0, omega-term, and the arithmetic 
     operations on ordinals which belong to epsilon0 *)


Lemma phi0_def : forall a, nf a -> omega ^ a =  phi0 a.
Proof.
  unfold  phi0;  cbn;simple  induction a; cbn.
 - reflexivity.
 -  destruct t; cbn.
    +  intros H n t H0 H1;  generalize (nf_of_finite H1);
       intro; subst;
       case n; cbn.  
       * reflexivity.
       * induction n0; cbn.
         reflexivity.
         rewrite IHn0;  reflexivity.
    +  intros  H n0 t H0 H1; unfold omega_term;
       rewrite H0; cbn.
       fold (phi0 t);
         fold (phi0 (ocons (ocons t1 n t2) n0 t));
         fold (phi0 (ocons (ocons t1 n t2) n0 zero));
         rewrite <- (omega_term_plus_rw H1);
         rewrite  phi0_plus_mult.
       * unfold omega_term;auto.
       * unfold omega_term. apply single_nf.  eauto with T1. 
       *  eapply nf_inv2;eauto.

       * eapply nf_inv2;eauto.
Qed.


Lemma omega_term_def : forall a n, nf a -> omega_term a n =
                                           omega ^ a *  FS n.
 Proof. 
  intros a n H; rewrite phi0_def; auto.
  simpl; case a; simpl; unfold omega_term; auto.
  - rewrite <- (plus_n_O n); auto.
  - rewrite <- (plus_n_O n);  auto.
 Qed.
 

Lemma ocons_def : forall a n b, nf(ocons a n b) ->
                       ocons a n b =  omega ^ a * FS n + b.
Proof.
 intros; rewrite <- omega_term_plus_rw; auto.
 - rewrite omega_term_def; auto.
   eapply nf_inv1;eauto.
Qed.

Theorem unique_decomposition : forall a n b a' n' b',
                    nf (ocons a n b) -> nf (ocons a' n' b') ->
                    omega ^ a *  FS n + b =
                    omega ^ a' * FS n' + b' ->
                    a = a' /\ n = n' /\ b = b'.
Proof.
  intros a n b a' n' b' N N'; rewrite <- (ocons_def N);
  rewrite <- (ocons_def N'); now injection 1.
Qed.

 
Theorem Cantor_normal_form :
  forall o, lt zero  o -> nf o ->
            {a:T1 & {n: nat & {b : T1 |
                              o = omega ^ a * FS n + b  /\
                              nf (ocons a n b) }}}.
Proof.
 intro ; case o.
 - intro H; case (lt_irrefl H). 
 - intros a n b H H0; exists a;exists n;exists b; split. 
   + now apply ocons_def.
   + assumption.
Defined.


Lemma LT_one : forall alpha, alpha < one -> alpha = zero.
Proof.
 intros alpha [H1 [H2 _]].
 destruct alpha; auto. 
 destruct (lt_inv H2).
 - destruct (not_lt_zero H). 
 -   decompose [and or] H.
    destruct (Nat.nlt_0_r _ H4).
   destruct (not_lt_zero H5). 
Qed.


Lemma lt_omega_inv : forall alpha,  alpha < omega ->
                                     alpha = zero \/ exists n, alpha = FS n.
Proof.                                    
  intros alpha [H1 [H2 _]]; destruct alpha; auto. 
  destruct (lt_inv H2). 
  -  assert(alpha1 = zero).
    {   apply LT_one.
        split.
         subst. 
         eauto with T1.
 split.
    auto.
    now compute. 
    }
    subst alpha1;  right; exists n.
   generalize (nf_of_finite H1).
 intro; now subst. 
  -     destruct H as [[H3 H4] |[H3 [H4 H5]]].
       destruct (Nat.nlt_0_r _ H4).
        destruct (not_lt_zero H5). 
Qed.

Ltac T1_inversion H :=
  match type of H with lt _ zero => destruct (not_lt_zero H)
                    | Nat.lt _ 0 => destruct (Nat.nlt_0_r _ H)
                    | Nat.lt ?x ?x => destruct (Lt.lt_irrefl _ H)
                    | lt ?x ?x => destruct (lt_irrefl  H)
                    | lt (ocons _ _ _) (ocons _ _ _) =>
                      destruct (lt_inv H)
                    | nf (ocons zero ?n ?y) => let e := fresh "e" in
                                               generalize (nf_of_finite H);
                                                 intros  e
  end.

Lemma LT_of_finite :
  forall alpha n, alpha < FS n -> alpha = zero \/
                                  exists p, (p <= n)%nat  /\ alpha = FS p.
Proof.
  intros alpha n [H1 [H2 H3]].
  destruct alpha.
  - now left.
  - T1_inversion H2.
    T1_inversion H.
    destruct H as [[H4 H5] |[H4 [H5 H6]]].
    subst. 
    generalize (nf_of_finite H1).
    intro;subst. 
    right; exists n0.
    split; auto with T1 arith.  
    T1_inversion H6.
Qed.
   
  (* About omega ^ omega *)

Hint Resolve nf_phi0 : T1.

Definition omega_omega := phi0 omega.

Lemma nf_omega_omega : nf omega_omega.
Proof.  repeat constructor. Qed.   



Lemma is_limit_cases : forall alpha n beta,
                         nf (ocons alpha n beta) ->
                         is_limit (ocons alpha n beta)  ->
                         { alpha <> zero /\ beta = zero} +
                         {alpha <> zero /\ is_limit beta }.
Proof.
 intros alpha n beta H;  simpl;  destruct alpha.
  - discriminate.
  - destruct beta.
   +  left; split;auto.
      discriminate.
   + right; split; auto.
     * discriminate.  
Qed.


Lemma pred_of_succ : forall  beta,  nf beta -> pred (succ beta) = Some beta.
Proof.
   induction beta.
  - reflexivity.
  - simpl; destruct beta1.
   + intros; replace beta2 with zero.   
    reflexivity. 
    { T1_inversion H. auto.      }
  +  intros; simpl;  destruct n.
   *   destruct beta2.
       { reflexivity. }
       { simpl;  destruct (beta2_1).
         - f_equal.
           cbn. 
           apply nf_inv2 in H. 
           T1_inversion H.
           subst;auto. 
         -
           simpl; simpl in *; destruct (pred (succ beta2_2)).
           f_equal.
           injection   IHbeta2.
           intro;subst;auto.
           eapply nf_inv2;eauto. 
           discriminate IHbeta2.
           eapply nf_inv2;eauto. 
       }           
   * rewrite IHbeta2;   auto.
     apply nf_inv2 in H; eauto. 
Qed.

Lemma pred_of_limit : forall alpha, nf alpha ->
                                    is_limit alpha ->
                                    pred alpha = None.
Proof.
 intros; induction alpha.
 - reflexivity. 
 -    simpl; destruct alpha1.
   +   destruct n;  simpl in H0; discriminate.
   +   destruct (is_limit_cases  H H0).
    *  destruct a.    
       subst.
       reflexivity.
    * destruct a; rewrite IHalpha2;auto.
      eauto with T1.
Qed.


Lemma pred_of_limitR : forall alpha, nf alpha -> alpha <> zero ->
                                     pred alpha = None -> is_limit alpha.
Proof.
  intros alpha Halpha; destruct (zero_succ_limit_dec Halpha).
  - destruct s; auto.
  - destruct s  as [x [H H0]]; subst.
    rewrite pred_of_succ; auto.
    + discriminate.
Qed.


Lemma pred_LT : forall alpha beta, nf alpha -> pred alpha = Some beta ->
                                   beta < alpha .
Proof.
  intros; destruct (zero_succ_limit_dec H).
  - destruct s.
   +  subst; discriminate.
   +   rewrite pred_of_limit in H0; trivial.
       * discriminate.
 -   destruct s.
     destruct a.
     subst;     rewrite pred_of_succ in H0; trivial.
   +  injection H0; intros;  subst x; split; auto.
      split; auto.
      now apply lt_succ.
Qed.

Lemma pred_nf : forall alpha beta, nf alpha -> pred alpha = Some beta ->
                                   nf  beta.
Proof.
 intros alpha beta H H0; now destruct (pred_LT H H0).
Qed.


(** More properties of < *)

Lemma is_limit_succ : forall alpha, nf alpha -> ~ is_limit (succ alpha) .
  intros alpha ; transfinite_induction alpha.
  intros.    
  destruct x.

  - discriminate.
  - intros.
    simpl. 
     destruct x1.
  discriminate. 
  simpl.
  case_eq (succ x2).
   intros.
  now destruct (succ_not_zero x2).

 intros. 
  rewrite <- H1.
  apply H.
eauto with T1.
 eauto with T1.
  apply tail_lt_ocons.
 auto. 
 eauto with T1.
Qed. 


Lemma is_limit_succ_tail :
  forall alpha n beta, nf beta ->  ~ is_limit (ocons alpha n (succ beta)).
Proof.  
simpl;intros; destruct alpha.
- discriminate.
-  case_eq (succ beta).
   + intro; now destruct (succ_not_zero beta).
   +  intros gamma p delta H0;  rewrite <- H0; now  apply is_limit_succ.
Qed.


Lemma LT_succ : forall alpha, nf alpha -> alpha < succ alpha.
Proof. 
  repeat split; auto.
  -  apply lt_succ.
  -  now apply succ_nf.
Qed.

Lemma LE_r : forall alpha beta, alpha < beta -> alpha <= beta.
Proof.
  intros alpha beta [H1 [H2 H3]]; repeat split; eauto with T1.
Qed.


Lemma nf_LT_iff :
  forall alpha n beta, nf (ocons alpha n beta) <->
                       nf alpha /\ nf beta
                       /\ beta < phi0 alpha.
Proof.
  split.
  intros H.
  repeat split; eauto with T1.
  apply    lt_phi0_phi0.
  eapply lt_phi0_intro; eauto.
  intro H; decompose [and] H.
  apply nf_intro; eauto.
  apply lt_phi0_phi0R. destruct H3; tauto.
Qed.

Lemma nf_LT_right : forall alpha n beta beta',
                      nf (ocons alpha n beta) ->
                      beta' < beta ->
                      nf (ocons alpha n beta').
Proof.
   intros alpha n beta beta'; repeat rewrite nf_LT_iff.
 intros [H1 [H2 H3]] H4;  split; auto.
  split.   eauto with T1.
  apply LT_trans with beta; auto.
Qed.














Lemma is_limit_not_zero : forall alpha, nf alpha -> is_limit alpha  ->
                                        alpha <> zero.
Proof.
  unfold not; intros; subst;discriminate.
Qed. 


Lemma eq_succ_LT : forall alpha beta, nf beta -> alpha = succ beta ->
                                      beta < alpha.
Proof. 
intros; subst;  apply LT_succ; auto.
Qed. 

Lemma eq_succ_lt : forall alpha beta, nf beta -> alpha = succ beta ->
                                      lt  beta  alpha.
Proof. 
 intros alpha beta H H0; destruct (eq_succ_LT H H0); tauto.
Qed. 

  
Definition limit_strict (s : nat -> T1) (l : T1) :=
  (forall i, s i < l) /\
  (forall l', (forall i, s i < l') -> l <= l').


Definition limit (s : nat -> T1) (l : T1) :=
  (forall i, s i <= l) /\
  (forall l', (forall i, s i <= l') -> l <= l').


Definition limit_strict_lub : forall s l alpha,  limit_strict s l ->
                                                 (forall i, s i < alpha) ->
                                                 l <= alpha.
Proof. destruct 1; auto.  Qed.



Definition limit_lub : forall s l alpha,  limit s l ->
                                          (forall i, s i <= alpha) ->
                                          l <= alpha.
Proof.  destruct 1;eauto.  Qed.

Definition limit_strict_maj : forall s l ,  limit_strict s l ->
                                            forall i, s i < l.
Proof. destruct 1; trivial.  Qed.


Definition limit_maj : forall s l ,  limit s l ->
                                      forall i, s i <= l.
Proof. now destruct 1. Qed.



Lemma limit_strict_unique : forall s l l', limit_strict s l ->
                                           limit_strict s l' ->
                                           l = l'.
Proof.
  intros s l l' H H'; destruct H, H';  apply LE_antisym;eauto.
Qed.

Lemma limit_unique : forall s l l', limit s l ->
                                           limit s l' ->
                                           l = l'.
Proof.
  intros s l l' H H'; destruct H, H';  apply LE_antisym;eauto.
Qed.



Lemma not_zero_lt : forall alpha, nf alpha -> alpha <> zero -> zero < alpha.
Proof.
  split.
   constructor. 
   split;auto. 
   destruct alpha.
   destruct H0;auto.
   constructor. 
Qed.

Lemma not_LT_zero : forall alpha,  ~ alpha < zero.
Proof.
  red; destruct 1.
  destruct H0. inversion H0. 
  T1_inversion H0.
Qed.


Lemma LT_succ_le_2_s :
  forall alpha beta, nf alpha -> nf beta ->
                     alpha < succ beta -> {alpha < beta} + {alpha = beta}.
Proof. 
 intros.
  destruct H1 as [H2 [H3 H4]].
  generalize (lt_succ_le_2 H H0 H3).
  intro;  destruct (le_eq_lt_dec _ _ H1).
  now right.
  left; repeat split; auto. 
Qed.

Lemma LT_succ_le_2 :
  forall alpha beta, nf alpha -> nf beta ->
                     alpha < succ beta -> {alpha < beta} + {alpha = beta}.
Proof. 
 intros.
  destruct H1 as [H2 [H3 H4]].
  generalize (lt_succ_le_2 H H0 H3).
  intro.
  destruct (le_eq_lt_dec _ _ H1).
  now right.
  left; repeat split; auto. 
Qed.




Lemma lt_succ_le_2':
  forall a : T1, nf a -> forall b : T1, nf b -> a <  succ b  ->
                                        a < b \/ a=b.
Proof.
  intros.
  destruct H1 as [H2 [H3 H4]].
  generalize (lt_succ_le_2  H2 H0 H3);auto.
  intro H5;  destruct (le_lt_or_eq _ _ H5); auto. 
  left; split; auto.   
Qed.   


Lemma LE_LT_eq_dec : forall alpha beta, alpha <= beta ->
                                        {alpha < beta} + {alpha = beta}.
Proof. 
 unfold LE, restrict; intros alpha beta H; decompose [and] H.
 destruct (le_eq_lt_dec _ _ H2).
 now right.  
 left; repeat split; auto. 
Defined. 

Lemma succ_lt_limit alpha (Halpha : nf alpha)(H : is_limit alpha ):
  forall beta, beta < alpha -> succ beta < alpha.
Proof. 
  intros.
  assert (succ beta <= alpha).
  apply  LT_succ_LE; auto. 
 destruct (LE_LT_eq_dec H1); auto. 
  subst alpha. 
  SearchAbout (is_limit (succ _)).
  destruct  (@is_limit_succ beta ); eauto with T1. 
Qed. 


Lemma succ_ocons alpha n beta : beta <> zero -> nf (ocons alpha n beta) -> 
                                succ (ocons alpha n beta) =
                                ocons alpha n (succ beta).
destruct alpha.
simpl. 
intros.
  destruct H; apply nf_of_finite with n; auto. 
  reflexivity.
Qed.


(* Demo *)

Example alpha_0 : T1 :=
  ocons (ocons (ocons zero 0 zero)
               0
               zero)
        0
       (ocons (ocons zero 2 zero)
              4
              (ocons zero 1 zero)).

Compute alpha_0.

(*
   = ocons omega 0 (ocons (FS 2) 4 (FS 1))
     : T1
*)



Example alpha_0_eq : alpha_0 = phi0 omega  +
                               phi0 (fin 3)  * fin 5 +
                               fin 2.
reflexivity.
Qed.






  