(** *  Pierre Cast\u00e9ran, 
       LaBRI and University of Bordeaux 

       After J. Ketonen and R. Solovay's paper
  " Rapidly Growing ramsey Functions"
    Annals of mathematics, Mar. 1981 

 *)

Require Import Arith Epsilon0.
Import Relations Relation_Operators.

Set Implicit Arguments.

(* Ketonen and Solovay study the operators "{alpha}(i)" and the relation 
   they generate

   In our adaptation to Coq, we use other notation  :
   "d i alpha" instead of {alpha}(i)  (for i>= 1) (like in Sladek's paper
   "the Termite and the Tower"

   For simplicity's sake, our main functions use "dS i alpha" instead
  of "d (i+1) alpha" (for i<= 0)
 *)


(** Canonical sequences *)

(** Contrary to KS, We didn't define {alpha}0) *)

Fixpoint dS (i:nat) alpha :=
  match alpha with
      zero => zero
    | ocons zero 0 zero => zero
    | ocons zero (S k) zero => FS k
    | ocons gamma 0 zero =>
      match pred gamma with
          Some gamma' => ocons gamma' i zero
        | None => ocons (dS i gamma) 0 zero
      end
    |  ocons gamma (S n) zero =>
       match pred gamma with
           Some gamma' => ocons gamma n (ocons gamma' i zero)
         | None => ocons gamma n (ocons (dS i gamma) 0 zero)
       end
    | ocons alpha n beta => ocons alpha n (dS i beta)
  end.

(*  (d 0 alpha) not compatible with KS 
    but (d i alpha)  OK for any i > 0 
*)

Definition d i alpha := match i with 0 => zero | S j => dS j alpha end.

(** {omega^omega}(3) *)

Example d3 :  d 3 (omega ^ omega) = omega ^  3.
Proof. reflexivity. Qed.

(** ** Rewrite lemmas for simplification of canonical sequences *)

Lemma dS_zero : forall i, dS i zero = zero.
Proof. reflexivity. Qed.




Lemma dS_lim1 : forall i lambda, nf lambda -> is_limit lambda = true
                                -> dS i (ocons lambda 0 zero) =
                                   ocons (dS i lambda) 0 zero.
Proof.
  intros; unfold dS at 1; destruct lambda.
  - simpl; discriminate.
  - replace ( pred (ocons lambda1 n lambda2)) with (@None T1).
    + f_equal.
    + rewrite pred_of_limit;auto.
Qed.


Lemma dS_lim2 : forall i n lambda, nf lambda -> is_limit lambda = true
                                -> dS i (ocons lambda (S n) zero) =
                                   ocons  lambda n (ocons  (dS i lambda) 0 zero).
Proof.
  intros; unfold dS at 1;  destruct lambda.
- simpl; discriminate.
- replace ( pred (ocons lambda1 n lambda2)) with (@None T1);   f_equal.
  + rewrite pred_of_limit;auto.
  +  rewrite pred_of_limit;auto.
Qed.

Lemma dS_succ_eqn1 : forall i gamma, nf gamma -> 
                                 dS i (ocons (succ gamma) 0 zero) =
                                 ocons gamma i zero.
Proof.
  intros; unfold dS at 1;  rewrite pred_of_succ;
  case_eq (succ gamma); trivial.
 -  intro; destruct (succ_not_zero _ H0); auto 2.
Qed.


Lemma dS_succ_eqn2 : forall i n gamma, nf gamma -> 
                                 dS i (ocons (succ gamma) (S n) zero) =
                                 ocons (succ gamma) n (ocons gamma i zero).
Proof.
  intros; unfold dS at 1;   rewrite pred_of_succ;
  case_eq (succ gamma); trivial.
  -  intro; destruct (succ_not_zero _ H0).
Qed.

Lemma dS_tail : forall i alpha n beta,
                  nf (ocons alpha n beta) ->
                  beta <> zero ->
                  dS i (ocons alpha n beta) =
                  ocons alpha n (dS i beta).
Proof.
  intros; unfold dS at 1;destruct alpha, n, beta.
  - now destruct H0.
  - f_equal.
  - simpl dS; now destruct H0.
  -  f_equal; now destruct  H0.
  -  f_equal; now destruct  H0.  
  -  f_equal.
  - now destruct H0.
  -  f_equal.
Qed.

Lemma dSSn (i:nat) :
  forall alpha n  ,
    nf alpha -> 
    dS i (ocons alpha (S n) zero) =
    ocons alpha n (dS i (ocons alpha 0 zero)).
 Proof. 
  intros; destruct (@zero_succ_limit_dec alpha).
  - eauto with T1.
  - destruct s.
    + destruct n.
      subst alpha; cbn.  reflexivity.       
      subst alpha; simpl; reflexivity.
    +      rewrite dS_lim2 ; auto.
           rewrite dS_lim1; auto.
  - destruct s as [beta [Hdeta e]]; subst alpha.
    rewrite dS_succ_eqn2, dS_succ_eqn1; auto.
 Qed.

 
 Lemma dS_succ : forall i alpha,  nf alpha -> dS i (succ alpha) = alpha.
 Proof.
   induction alpha.
   -   reflexivity.
   -   destruct alpha1.
       + intros; replace alpha2 with zero.   
         reflexivity. 
         {   T1_inversion H; auto. }  
       +  intros; simpl;  destruct n.
          destruct alpha2.
          * reflexivity.   
          *  simpl;destruct (alpha2_1).
             f_equal.
             apply nf_inv2 in H.
             T1_inversion H. 
             now subst. 
             f_equal; rewrite <- IHalpha2.
             f_equal.  
             eauto with T1.
          *  case_eq (succ alpha2).
             intros;  now destruct (@succ_not_zero alpha2).
             intros;   f_equal.
             rewrite <- IHalpha2.
             now rewrite <- H0.
             eauto with T1. 
 Qed.

 Lemma dS_zero_inv (alpha:T1) (i:nat) : 
   dS i alpha = zero -> alpha = zero \/ alpha = one.
 Proof.
   destruct alpha; cbn; auto.
   destruct alpha1, n, alpha2;auto; try discriminate.
   all:  destruct (pred (ocons alpha1_1 n0 alpha1_2)); try discriminate.
 Qed.



 Lemma dS_LT : forall i alpha, nf alpha -> alpha <> zero ->
                               dS i alpha <  alpha.
Proof with eauto with T1.
  intros i alpha ; transfinite_induction alpha.
  clear alpha . intros alpha Hrec Halpha; destruct (zero_succ_limit_dec Halpha).
  
  - destruct s.
    + subst alpha; now destruct 1.
    + 
      destruct alpha.
      * now destruct 1.
      * intros.
        destruct (T1_eq_dec alpha1 zero).
        {      subst.
               destruct n.
               - replace alpha2 with zero.
                 split;auto. 
                 simpl. 
                 constructor.
                 split;simpl;constructor.
                 constructor.
                 T1_inversion Halpha; auto.
               -       replace alpha2 with zero.
                       +        simpl.
                                split.
                                constructor.
                                constructor.
                                apply finite_lt; auto with arith. 
                                apply nf_finite.
                       +   T1_inversion Halpha; auto. 
        }

        destruct alpha2.
        {  destruct n.
           {   destruct  (@zero_succ_limit_dec alpha1) ...
               destruct s.
               subst;  now destruct n0. 
               rewrite dS_lim1.
               split.
               apply single_nf. 
               destruct (Hrec alpha1) ...
               apply head_lt_ocons ;  auto.
               split.
               apply head_lt. 
               apply Hrec ...
               apply head_lt_ocons; auto.
               auto. 
               eauto with T1.
               auto. 
               destruct s.
               {   destruct a.
                   { subst .
                     rewrite dS_succ_eqn1. split.
                      apply single_nf; auto. 
                      split.
                      apply head_lt. 
                      apply lt_succ; auto. 
                     auto.
                     eauto with T1.
                   }
               }
           }
           destruct  (@zero_succ_limit_dec alpha1) ...
           { destruct s.
             subst; now destruct n0.   
             rewrite dS_lim2.
             split.
             { apply ocons_nf. 
               destruct (Hrec alpha1) ...
               apply head_lt_ocons.
               tauto.
               eauto with T1.
               apply single_nf. 
               destruct (Hrec alpha1); eauto with T1.
               apply head_lt_ocons.
             }
             split.
             apply coeff_lt.
             auto with arith.
             eauto with T1.
             auto.
             eauto with T1.
           }
           destruct s; subst.
           {     destruct a.
                 subst; rewrite dS_succ_eqn2.
                 split.
                 { 
                   apply ocons_nf. 
                   apply lt_succ.
                   apply succ_nf; auto.
                   apply single_nf; auto.
                 }
                 split.
                 {     apply coeff_lt;   auto with arith.
                 }
                 auto with T1.
                 auto.
           }
        }
        rewrite dS_tail.
        {
          split.
          {
            apply nf_intro ...
            destruct (Hrec  (ocons alpha2_1 n1 alpha2_2)) ...
            discriminate.
            apply lt_phi0_phi0R.
            apply lt_trans with (ocons alpha2_1 n1 alpha2_2).
            destruct (Hrec (ocons alpha2_1 n1 alpha2_2)) ...
            discriminate.
            tauto.
            {
              apply lt_phi0_phi0.
              apply lt_phi0_intro with n;auto.
            }
          }
          split.
          apply tail_lt. 
          destruct (Hrec (ocons alpha2_1 n1 alpha2_2)); auto.
          eauto with T1.
          apply tail_lt_ocons;  auto.
           eauto with T1.
           discriminate.
          destruct (Hrec (ocons alpha2_1 n1 alpha2_2)) ...
          discriminate.
          tauto. 
          auto.
        }
        eauto with T1.
        discriminate.
  - destruct s.
    {   destruct a; subst.
        intros; rewrite dS_succ; auto.
        split;    auto.
        split; auto; apply lt_succ; auto.
    }
Qed. 


Lemma nf_dS : forall i alpha, nf alpha -> nf (dS i alpha).
Proof.
    intros i alpha Hnf.
    destruct (T1_eq_dec alpha zero).
    - subst. constructor.
    - destruct  (dS_LT i  Hnf); auto.
Qed.



  
Lemma dS_lt : forall i alpha, nf alpha -> alpha <> zero ->
                              lt (dS i alpha) alpha.
Proof.
  intros i alpha Hnf.
  destruct (T1_eq_dec alpha zero).
  - subst. now destruct 1.
  - destruct  (dS_LT i Hnf); tauto.
Qed.


Lemma dS_limit_mono alpha i j : nf alpha -> is_limit alpha = true ->
                                (i < j)%nat -> dS i alpha < dS j alpha.
Proof with eauto with T1.
  pattern alpha ; apply transfinite_recursor.
  clear alpha;  intros alpha Hrec alpha_nf.
  intros.  
  destruct alpha.
  - discriminate.
  -  destruct (is_limit_cases alpha_nf H).
     +  destruct a;subst.
        destruct (@zero_succ_limit_dec alpha1) ...
        *   destruct s.   
            { subst;   discriminate H. }
            {   destruct n.
                - repeat rewrite dS_lim1 ...
                  {  apply LT2 ...
                     apply single_nf. 
                     apply nf_dS ...
                     apply single_nf. 
                     apply nf_dS ...
                     apply Hrec ...
                     apply head_lt_ocons.
                  }
                - repeat rewrite dS_lim2 ...
                  {
                    apply LT4 ...
                    apply nf_ocons_LT ...
                    apply dS_LT ...
                    apply single_nf. 
                    apply nf_dS ...
                    apply nf_ocons_LT ...
                    apply dS_LT ...
                    apply single_nf; apply nf_dS ...
                    apply LT2 ...
                    apply single_nf;
                    apply nf_dS ...
                    apply single_nf;
                    apply nf_dS ...
                    apply Hrec ...
                    apply head_lt_ocons.
                  }
            }
        * destruct s;subst.
          destruct a.
          subst alpha1.
          destruct n.
          { repeat rewrite dS_succ_eqn1 ... }
          {
            repeat rewrite dS_succ_eqn2 ...
            apply LT4 ...
            apply nf_ocons_LT ...
            apply LT_succ ...
            apply nf_ocons_LT ...
            apply LT_succ ...
          }
     + destruct a.
       repeat rewrite dS_tail ...
       * apply  LT4 ...
         apply nf_intro ...
         apply nf_dS ...
         apply lt_phi0_phi0R ...
         apply lt_trans with alpha2 ...
         apply dS_lt ...
         intro;subst;discriminate.
         apply lt_phi0_phi0 ...
         apply lt_phi0_intro with n ...
         apply nf_intro ...
         apply nf_dS ...
         apply lt_phi0_phi0R ...
         apply lt_trans with alpha2 ...
         apply dS_lt ...
         intro;subst;discriminate.
         apply lt_phi0_phi0 ...
         apply lt_phi0_intro with n ...
         apply Hrec ...
         apply tail_lt_ocons ...
     *  intro;subst;discriminate.
     *  intro;subst;discriminate.
Qed.



Lemma is_limit_dS_not_zero : forall i alpha, nf alpha ->
                                             is_limit alpha  ->
                                             dS i alpha <> zero.
Proof.
destruct alpha.
- discriminate.
- intros; simpl; destruct alpha1.
 +   destruct n.
   *  destruct alpha2 ; discriminate.
   *  destruct alpha2; discriminate.
 +  destruct n.
   *  destruct alpha2.
    {  destruct (pred (ocons alpha1_1 n0 alpha1_2));   discriminate. }
       discriminate.
   *   destruct alpha2.
      destruct (pred (ocons alpha1_1 n0 alpha1_2)); discriminate.
      discriminate.
Qed.


(** Let alpha be a limit ordinal.
    For any beta < alpha, we can compute some item of the canonical sequence
    associated with alpha which is greater than beta.
   
   It is a constructive way of expressing that alpha is the limit of its 
   own canonical sequence *)

Lemma dS_limit_S_strong lambda : 
  nf lambda ->
  is_limit lambda  ->
  forall beta, beta < lambda ->
               {i:nat | beta < dS i lambda}.
Proof.
  transfinite_induction_LT lambda.
  clear lambda ; intros lambda Hrec Hlambda.
  intros   H beta H1.
  assert (Hbeta: nf beta) by eauto with T1.
  destruct (zero_succ_limit_dec Hlambda).
  -   destruct s.   
      { (* lambda = zero *)
        subst lambda. inversion H1.
        destruct (not_LT_zero H1).
      }
      destruct lambda.
      discriminate.
      destruct (is_limit_cases Hlambda i).
      { (* first case : beta = zero *)
        destruct a; subst lambda2.
        destruct beta.
        { exists 0; apply not_zero_lt. 
          apply nf_dS; auto.
          apply is_limit_dS_not_zero;auto.
        }
        destruct (LT_inv_strong H1).
        destruct n.
        destruct (@zero_succ_limit_dec   lambda1); eauto with T1.
        destruct s.
        subst; assert (False) by (eapply not_LT_zero; eauto).
        contradiction. 
        assert {i :nat | beta1 < dS i lambda1}.         
        {   apply Hrec.
            apply head_LT_cons; auto with T1.
            all: auto.
        }
        destruct H3.
        exists x; rewrite dS_lim1; eauto with T1.
        destruct s   as [x [H4 H5]].
        subst lambda1; assert ({beta1 < x} + {beta1 = x}).
        { apply  LT_succ_le_2 in H2 ;eauto with T1.  }
        destruct H3.
        exists 0; rewrite dS_succ_eqn1;auto with T1.
        subst; exists (S n0);  rewrite dS_succ_eqn1; auto. 
        apply LT3; auto with arith.
        clear i; exists 0;  rewrite dSSn.
        apply LT2; eauto with T1.
        apply nf_intro; eauto with T1.
        apply nf_dS; eauto with T1.
        apply lt_phi0_phi0R.
        unfold phi0; apply dS_lt.
        eapply nf_coeff_irrelevance;eauto. 
        discriminate. 
        eauto with T1.
        - subst.
          +  destruct n.
             not_neg H3.
             apply lt_n_Sm_le in H3.
             destruct (le_lt_eq_dec _ _ H3).
             exists 0; rewrite dSSn. 
             apply LT3; auto.
             {
               apply nf_intro; eauto with T1.
               apply nf_dS; eauto with T1.
               apply lt_phi0_phi0R.
               unfold phi0.
               apply dS_lt.
               eapply nf_coeff_irrelevance;eauto.     
               discriminate.  
             }
             { eauto with T1. }
             
             assert {i: nat |  beta2 < (dS i (ocons lambda1 0 zero))}.
             { apply Hrec.
               apply LT3;auto with arith.
               cbn;  destruct lambda1.
               destruct H1; auto.
               auto.
                cbn.
                destruct lambda1.
                cbn in H.
                discriminate.
                auto. 
                {
                  split.
                  eauto with T1.
                  split. 
                  apply lt_phi0_phi0.     
                  apply lt_phi0_intro with n0;auto with T1.
                  eapply nf_coeff_irrelevance;eauto.     
                }
             }
             subst;
               assert {i: nat |  beta2 < (dS i (ocons lambda1 0 zero))}.
             { apply Hrec.
               apply LT3;auto with arith.
               cbn;  destruct lambda1.
               destruct H2; auto. 
               auto.
               cbn;  destruct lambda1.
               destruct H2; auto. 
               auto.                
               {
                 
                 split.
                 eauto with T1.
               split. 
               apply lt_phi0_phi0.    
               apply lt_phi0_intro with n;auto with T1.
               eapply nf_coeff_irrelevance;eauto.   
             }
            }
             destruct H4.
             exists x; rewrite dSSn. 
             apply LT4;auto.
             { 
               apply nf_intro;  eauto with T1.
               apply lt_phi0_phi0R.   
               apply dS_lt.
               apply nf_phi0; eauto with T1.
               unfold phi0; discriminate.
             }
             auto.
        -  
          assert False.
          destruct (LT_zero H4).
          contradiction. 
      }
      destruct a ;  assert (lambda2 <> zero).
      intro; subst; discriminate.   
      destruct beta.
      exists 0;  apply not_zero_lt.
      apply nf_dS; auto with T1.
      apply is_limit_dS_not_zero; eauto with T1.

      destruct (LT_inv_strong H1).
      exists 0;  rewrite dS_tail.
      apply LT2;auto.
      { 
        apply nf_intro. 
        eauto with T1.
        apply nf_dS; eauto with T1.

        apply lt_phi0_phi0R.   
        apply lt_trans with lambda2.

        apply dS_lt; eauto with T1.
        apply lt_phi0_phi0.
        apply lt_phi0_intro with n; eauto with T1.
      }
      auto.
      auto.
      destruct H4.
      exists 0.
      rewrite dS_tail.
      apply LT3.
      auto.
      {
        apply nf_intro. 
        eauto with T1.
        apply nf_dS; eauto with T1.
        apply lt_phi0_phi0R.   
        apply lt_trans with lambda2.
        apply dS_lt; eauto with T1.
        apply lt_phi0_phi0.
        apply lt_phi0_intro with n; eauto with T1.
      }

      all: auto. 

      subst.
      assert {i: nat |  beta2 < (dS i lambda2)}.
      apply Hrec.
      {
        apply tail_LT_cons; auto.
      }

      auto.
      auto. 
      eauto with T1.
      auto.
      auto.
      destruct H4.
      exists x. 
      rewrite dS_tail.
      apply LT4.
      auto. 
      {
        apply nf_intro. 
        eauto with T1.
        apply nf_dS; eauto with T1.
        apply lt_phi0_phi0R.   
        apply lt_trans with lambda2.
        apply dS_lt; eauto with T1.
        apply lt_phi0_phi0.
        apply lt_phi0_intro with n; eauto with T1.
      }
      all: auto. 
  -   destruct s.
      destruct a.
      subst. 
      destruct (@is_limit_succ  x); auto. 
Defined.


Lemma dS_limit : forall lambda, nf lambda -> is_limit lambda = true ->
                               limit_strict (fun i => dS i lambda) lambda.
Proof.
  split.
  - intros; split.
    + now  apply nf_dS.
    +    split; trivial.
         * apply dS_lt;    auto.
           intro H1;subst; discriminate.
  - intros l' Hl';  assert (nf l').
    {  specialize (Hl' 0); eauto with T1. }
    destruct (lt_eq_lt_dec lambda l').
    + repeat split;auto.
      destruct s.
      *  auto with T1. 
      *  subst; auto with T1. 
    + assert (l' < lambda) by (split; auto).
      destruct (dS_limit_S_strong   H H0 H2).
      specialize (Hl' x).
      assert (l' < l').
      { apply LT_trans with  (dS x lambda); auto. }
      now destruct (@LT_irrefl l' ).
Qed. 

SearchAbout (dS _ _ = _).

About zero_succ_limit_dec.


(** Relational version of Ketonen Solovay function 
Our notation [dS_plus i alpha beta] corresponds to
              [alpha --- i+1 ---> beta] 


*)

Definition dS_rel (i:nat) : relation T1 :=
  fun alpha beta => beta = dS i alpha.

Definition dS_plus i := clos_trans_1n T1 (dS_rel i).

Definition d_plus i alpha beta :=
  match i with
      0 => False
    | S j => dS_plus j alpha beta
  end.

Definition dS_star i alpha beta := dS_plus i alpha beta \/
                                   alpha = beta.

Definition d_star i alpha beta := d_plus i alpha beta \/
                                   alpha = beta.



Lemma dS_plus_nf : forall n alpha beta, 
                                        dS_plus n alpha beta ->
                                        nf alpha -> nf beta.
  (* move to KS *)
  Proof.
  induction 1. 
   red in H; subst. 
    intro; apply nf_dS;auto. 
    intro; apply IHclos_trans_1n.
    red in H; subst. 
 apply nf_dS;auto. 
  Qed. 


Lemma dS_plus_LT i : forall alpha beta, alpha <> zero ->
                                         nf alpha ->
                                         dS_plus i alpha beta ->
                                         beta < alpha.
Proof.
  induction 3.
  - red in H1;subst; split.
   +   apply nf_dS;  auto.
   +   split;auto.
       apply dS_lt;auto.
  -  destruct (T1_eq_dec y zero).
    +   subst;  assert (z= zero).
        { clear H1 IHclos_trans_1n.
          rewrite <- clos_trans_t1n_iff in H2.
          rewrite clos_trans_tn1_iff in H2.
          induction H2.
          red in H1.       
          assumption.
          subst y.
          red in H1.
          assumption.
        }
        subst. 
        split;auto with T1.
        destruct x.
        destruct H;auto.
        split; auto with T1.
    +  apply LT_trans with y.
       apply IHclos_trans_1n;  auto.
       red in H1;  subst;  apply nf_dS;  auto.
       red in H1; subst;  split; auto.
       apply nf_dS;  auto.
       split; auto.
       * apply dS_lt;   auto. 
Qed.

Lemma dS_plus_zero : forall n alpha, dS_plus n zero alpha -> alpha = zero.
Proof.
  intros n alpha H;
  red in H; rewrite <- clos_trans_t1n_iff in H;
  rewrite clos_trans_tn1_iff in H;
  induction H; red in H; now subst.
Qed.


Lemma dS_plus_LE i : forall alpha beta, 
                       nf alpha ->
                       dS_plus i alpha beta ->
                       beta <= alpha.
Proof.
  intros; destruct (T1_eq_dec alpha zero).
  - subst; apply dS_plus_zero in H0.
    subst; repeat split; auto. 
  -  apply LE_r;  eapply dS_plus_LT; eauto.
Qed.
     




Lemma dS_plus_inv : forall n alpha beta, dS_plus n alpha beta ->
                                          beta = dS n alpha \/
                                          dS_plus n (dS n alpha) beta.
Proof.
  inversion 1.
  -  left; congruence.
  - right; red in H0;subst; auto.
Qed.

(** KS's proposition 2.3 b  *)
Lemma dS_plus_trans : forall n alpha beta gamma,
                         dS_plus n alpha beta ->
                         dS_plus n beta gamma ->
                         dS_plus n alpha gamma.
Proof.                    
 unfold dS_plus; intros n alpha beta gamma H H0;                       
 rewrite <-  clos_trans_t1n_iff in *;
 constructor 2 with beta;assumption.
Qed.


Lemma dS_star_trans : forall i alpha beta gamma,
                        dS_star i alpha beta -> 
                        dS_star i beta gamma ->
                        dS_star i alpha gamma .
Proof. 
  destruct 1.
  - destruct 1.
    + left.
      eapply dS_plus_trans; eauto. 
    +  subst; now left. 
  -    subst; auto. 
Qed.

Lemma dS_star_LE_2 : forall n alpha beta,  nf alpha -> nf beta ->
                                           dS_star n alpha beta ->
                                           beta <= alpha.
Proof. 
  destruct 3.                                           
  - eapply dS_plus_LE; eauto. 
  - subst; apply LE_refl; auto.
Qed. 


Lemma Proposition_2_3a:  (* KS paper, p 281 *)
  forall n alpha , nf alpha -> 
                   forall beta gamma,   gamma < beta ->
                                        dS_plus n alpha gamma ->
                                      dS_plus n alpha beta ->
                                      dS_plus n beta gamma.
Proof.
  intros n alpha .
  transfinite_induction_LT alpha.  clear alpha ;
  intros alpha Hrec. intros H beta gamma H0 H1 H2.
  destruct (dS_plus_inv H1), (dS_plus_inv H2).
  - subst; destruct (LT_irrefl H0).
  - destruct (T1_eq_dec alpha zero).
    + subst.
      assert (beta = zero).
      { eapply dS_plus_zero;eauto. }
      subst; cbn; left; red;auto.
    + subst gamma; elimtype False.
      apply dS_plus_LE in H4.
      destruct (@LT_irrefl beta).
      eapply LE_LT_trans;eauto.
      apply nf_dS;auto.

  - subst beta;  auto.
  - destruct (T1_eq_dec alpha zero).
    +  subst alpha; assert (beta = zero) .
       { eapply dS_plus_zero; eauto. }
       subst;  destruct (LT_zero H0).
    + apply (Hrec (dS n alpha)); auto.
          apply dS_LT; auto. 
   apply nf_dS;auto.
Qed.



(* KS's Theorem 2.4 *)

(**  Lemma 1 (reformulated) *)

Lemma Theorem_2_4_Lemma1 : forall i alpha n beta beta',
                  nf (ocons alpha n beta) ->
                  beta <> zero ->
                  dS_plus i beta beta' ->
                  dS_plus i (ocons alpha n beta)
                          (ocons alpha n beta').
Proof.
  induction 3.
  -  red in H1;subst; left. 
     red; now  rewrite dS_tail.
  -  red in H1;subst;  destruct (T1_eq_dec (dS i x) zero).    
     { rewrite e in *;  assert (z = zero).
       {   eapply dS_plus_zero; eauto. }
       subst z;    left;    red;  rewrite dS_tail; auto. 
       now rewrite e.
     }
     {
       constructor 2 with   (ocons alpha n (dS i x)).  
       red;  now  rewrite dS_tail.
       apply IHclos_trans_1n.   
       apply nf_LT_right with x;auto. 
       apply dS_LT; eauto with T1.
       auto.
  }
Qed.



Lemma dS_plus_first_step : forall i alpha beta,
                           dS_plus i alpha beta ->
                           { beta = dS i alpha } +
                           { dS_plus i (dS i alpha) beta}.
Proof.     
  intros; destruct (T1_eq_dec beta (dS i alpha)).
  - now left.   
  -  right;  generalize n; induction H.
   +   red in H; destruct n;auto. 
   + intros;  red in H; now  subst y. 
Qed.    


Lemma Theorem_2_4_Lemma1' : forall i alpha n beta ,
                              nf alpha -> 
                              alpha <> zero ->
                              dS_plus i (phi0 alpha)  beta ->
                              dS_plus i (ocons alpha (S n) zero)
                                      (ocons alpha n  beta).
Proof.
  intros.
  destruct (zero_succ_limit_dec H). 
  - destruct s. 
    +  now destruct H0.
    +  destruct (dS_plus_first_step H1).
       { rewrite  e; left; red;  rewrite dSSn.
         f_equal.
         auto.
       }
       { apply dS_plus_trans with (ocons alpha n (phi0 (dS i alpha))).
         { left; red; rewrite dSSn.
           - rewrite dS_lim1;auto.
           - trivial.
         }
         apply Theorem_2_4_Lemma1.
         unfold phi0;  apply nf_ocons_LT; eauto with T1.
         apply dS_LT; auto.
         apply single_nf.
         apply nf_dS; auto.
         discriminate. 
         unfold phi0 in d0; rewrite dS_lim1 in d0.
         assumption.
         auto. 
         auto. 
       }      
  - destruct s as [gamma [H2 H3]].
    destruct (dS_plus_first_step H1).
    { 
      rewrite e. 
      left;red; subst alpha; unfold phi0.
      repeat rewrite dS_succ_eqn1; trivial.
      now rewrite dS_succ_eqn2.
    }
    apply dS_plus_trans with (ocons alpha n (ocons gamma i zero)).
    left;red; subst alpha; rewrite dS_succ_eqn2; auto. 
    apply Theorem_2_4_Lemma1; auto. 
    apply nf_ocons_LT.
    subst; apply LT_succ; auto. 
    auto.
    apply single_nf; auto. 
    discriminate. 
    unfold phi0 in d0; subst alpha; rewrite  dS_succ_eqn1 in d0; auto. 
Qed.


Lemma Theorem_2_4_lemma2 (n:nat)(alpha:T1) :
  nf alpha -> dS_plus n alpha zero.
Proof.

 transfinite_induction_LT alpha.
 clear alpha ; intros alpha hrec Halpha.
 destruct (zero_succ_limit_dec Halpha).
 -  destruct s.
  +  subst alpha ; left ; reflexivity.
  +  intros;  specialize (hrec (dS n alpha)).
     constructor 2 with (dS n alpha).
     *  reflexivity.
     * apply hrec;  apply dS_LT;auto.
       intro; subst ; discriminate.
       eapply nf_dS; eauto. 
       now apply is_limit_dS_not_zero.
- destruct s as [beta [Hbeta e]]; subst.
     constructor 2 with beta; auto.
     +  red; now  rewrite dS_succ.
    + apply hrec; now apply LT_succ.
Qed. 


Lemma Theorem_2_4_lemma3_0 : forall i alpha n, nf alpha ->
               dS_plus i (ocons alpha (S n)  zero)
                       (ocons alpha n zero).
Proof.
  intros i alpha n H;  destruct (zero_succ_limit_dec H).
  - destruct s.
   +  subst;  left; red;   reflexivity.
   +  right with (dS i   (ocons alpha (S n) zero)).
      * red;  reflexivity.
      * rewrite dS_lim2;auto.
        eapply Theorem_2_4_Lemma1; auto.
        apply ocons_nf; auto.
        apply dS_lt; auto.
        intro; subst; discriminate.
        apply single_nf;  apply nf_dS; auto.
        discriminate.
        apply Theorem_2_4_lemma2.
        apply single_nf. 
        apply nf_dS; auto.
  - destruct s as [beta [Hbeta e]]; subst.
    right with (dS i   (ocons (succ beta) (S n) zero)).
    +  red;reflexivity.
    +  rewrite dS_succ_eqn2;auto.
       eapply Theorem_2_4_Lemma1; auto.
       apply ocons_nf; auto. 
       apply lt_succ; auto. 
       discriminate.
       apply Theorem_2_4_lemma2;  auto with T1.
Qed.

(* Lemma 3 of Theorem 2.4 in [K-S] *)

Lemma Theorem_2_4_lemma3 : forall i n p alpha , nf alpha -> (n < p)%nat ->
                                 dS_plus i (ocons alpha p zero)
                                          (ocons alpha n zero).
Proof.
  induction 2.
  - apply Theorem_2_4_lemma3_0; auto.
  -   apply dS_plus_trans with  (ocons alpha  m zero); auto. 
   +     apply Theorem_2_4_lemma3_0; auto.   
Qed.

Lemma Theorem_2_4_lemma4 : forall i alpha,
               nf alpha ->
               dS_plus i (phi0 (succ alpha)) (phi0 alpha).
Proof.
  destruct i.
  - intros; unfold phi0; left; red;  rewrite dS_succ_eqn1; auto.
  -  intros; unfold phi0; right with (dS (S i) (phi0 (succ alpha))).
     +    red; unfold phi0; rewrite dS_succ_eqn1; auto. 
     +    unfold phi0; rewrite dS_succ_eqn1; auto;
          apply Theorem_2_4_lemma3; auto with arith.
Qed.

Lemma Theorem_2_4_lemma5 : forall i alpha beta,
                 dS_plus i alpha beta -> nf alpha ->
                 alpha <> zero -> 
                 dS_plus i (phi0 alpha) (phi0 beta).
  induction 1.
  {
    red in H; subst y; intros.
    destruct (zero_succ_limit_dec H).
    - destruct s.
      + subst x;now destruct H0.
      + unfold phi0; left; red; now rewrite dS_lim1.
    -  destruct s as [beta [H1 e]]; subst.
       unfold phi0; rewrite dS_succ; auto.
       * now apply Theorem_2_4_lemma4.
  }
  { red in H;  subst y;  unfold phi0; intros.
    destruct (zero_succ_limit_dec H).
    - destruct s.
      + subst x; now destruct H1.
      + right with (phi0 (dS i x)).
        * unfold phi0; red;   rewrite dS_lim1; auto.
        *   apply IHclos_trans_1n.
            apply nf_dS;auto.
            apply is_limit_dS_not_zero; auto.
    -    destruct s as [beta [H2 e]]; subst x.
         destruct (T1_eq_dec beta zero).
         { subst; cbn in H0.
           assert (z=zero).
           { eapply dS_plus_zero;  eauto. }
           subst;  destruct i.
           -  left; red; rewrite dS_succ_eqn1; auto.
           - red.   right with (ocons zero (S i) zero).
             constructor.
              apply Theorem_2_4_lemma3;auto. 
              auto with arith. 
         }
         rewrite dS_succ in IHclos_trans_1n; auto. 
         apply dS_plus_trans with (phi0 beta).
         apply Theorem_2_4_lemma4; auto. 
         apply IHclos_trans_1n; auto. 
  }
Qed.    


Theorem Theorem_2_4 :
  forall lambda, nf lambda ->
                      is_limit lambda  ->
                      forall i j, (i < j)%nat ->
                                  dS_plus 0 (dS j lambda) (dS i lambda).
Proof.
  intros alpha ; transfinite_induction_LT alpha.
  clear alpha ; intros alpha Hrec Halpha H.
  destruct (zero_succ_limit_dec Halpha).
 - destruct s.
   + subst;  discriminate.
   + destruct alpha.
     * discriminate.
     *  destruct (is_limit_cases Halpha  H).
        { destruct a; subst alpha2.
          intros.
          destruct (@zero_succ_limit_dec alpha1).
          eauto with T1.    
          destruct s.
          subst; now destruct H0.
          { induction n.
            - repeat rewrite dS_lim1; auto.
              apply Theorem_2_4_lemma5. 
              apply Hrec;auto.
              split; eauto with T1.     
              split.
              apply head_lt_ocons; eauto with T1.
              auto.
              apply nf_dS;auto.
              apply is_limit_dS_not_zero; auto. 
            -  repeat rewrite dS_lim2; auto.
               apply Theorem_2_4_Lemma1; auto.
               apply ocons_nf; eauto with T1.
               apply dS_lt;eauto with T1.
               apply single_nf.   
               apply nf_dS;eauto with T1.
               discriminate.
               apply Theorem_2_4_lemma5;auto.
               apply Hrec;auto.
               split; eauto with T1.
               split.
               apply  head_lt_ocons; eauto with T1.
               apply single_nf; eauto with T1.
               apply nf_dS;eauto with T1.
               apply is_limit_dS_not_zero;  eauto with T1.
          }
          destruct s as [beta [H2 H3]];  subst; clear H i; induction n.
          {
            repeat rewrite  dS_succ_eqn1; eauto with T1.  
            apply Theorem_2_4_lemma3;auto.    
          }
          { 
            repeat rewrite  dS_succ_eqn2; eauto with T1.  
            apply Theorem_2_4_Lemma1; auto.
            apply ocons_nf.
            apply lt_succ; auto.
            apply succ_nf; eauto with T1.
            apply single_nf; eauto with T1.
            discriminate. 
            apply Theorem_2_4_lemma3;auto.
          }
        }
        intros.
        destruct a.
        repeat rewrite dS_tail; eauto with T1.
        apply Theorem_2_4_Lemma1; eauto with T1.
        apply nf_LT_right with alpha2; eauto with T1.
        apply dS_LT.
        eauto with T1.
        intro; subst; discriminate.
        apply is_limit_dS_not_zero.
        eauto with T1.
        auto.
        apply Hrec; eauto with T1.
        split;  eauto with T1. 
        split.
        apply  tail_lt_ocons; auto.
        auto. 
        intro; subst; discriminate.
        intro; subst; discriminate.
 -  destruct s as [beta [H0 H1]].
  subst; destruct (@is_limit_succ beta); auto. 
Qed.




Corollary C12 : forall alpha,  nf alpha ->
                               forall beta i n, beta < alpha -> (i < n)%nat ->
                                       dS_plus i alpha beta ->
                                       dS_plus n alpha beta.
Proof.
  intros alpha .
  pattern alpha; apply transfinite_recursor;auto.
  clear alpha ; intros alpha Hrec Halpha; intros.
  destruct (zero_succ_limit_dec   Halpha).
  - 
    destruct s.
    subst.
    destruct (LT_zero H).

    (* alpha limit *)
    
    destruct  H1.
    (* one step only *)
    { red in H1.    
      subst y.
      assert (dS_plus 0 (dS n alpha) (dS i alpha)).     
      {  apply Theorem_2_4;  auto. }
      constructor 2 with (dS n alpha).
      -              split. 
      -       apply Hrec with 0; auto.
              +  apply nf_dS;   auto. 
              + apply dS_lt; auto. 
                intro; subst ; discriminate.
              +  apply  nf_dS; auto.
              + apply dS_limit_mono; auto.
              +  apply Nat.le_lt_trans with i;  auto with arith.
    }
    {     
      (*   alpha -i->  y= dS i alpha -i-+->  z *)
      red in H1;  subst y;   assert (dS_plus n (dS i alpha) z).
      {
        apply Hrec with i;auto.
        apply nf_dS;   auto.
        apply dS_lt;auto.
        intro; subst;discriminate.
        eapply dS_plus_LT.
        apply   is_limit_dS_not_zero; auto.
        apply nf_dS; auto.
        eapply H2.
        eapply dS_plus_LT; eauto.
        eapply is_limit_dS_not_zero; eauto.
        now apply nf_dS.
      }

      {
        red; rewrite <- clos_trans_t1n_iff.
        apply t_trans with (dS n alpha).
        left.
        red.
        auto.
        right with (dS i alpha).
        rewrite  clos_trans_t1n_iff.
        apply Hrec with 0;eauto.
        now apply nf_dS.
        apply dS_lt; auto.
        intro; subst;discriminate.
        apply nf_dS;auto.
        apply  dS_limit_mono ;auto.
        apply Nat.le_lt_trans with i;auto.
        auto with arith.
        apply Theorem_2_4; auto.
        now rewrite  clos_trans_t1n_iff.
      }
     }
  - 
    destruct s.
    destruct a.
    subst. 
    { destruct H1.
      red in H1.
      subst y.
      rewrite dS_succ.
      left.
      red.
      rewrite dS_succ;auto.
      auto.
      red in H1.
      subst y.
      
      rewrite dS_succ in H3.
      right with x.
      red.
      rewrite dS_succ;auto.
      destruct (T1_eq_dec x zero). 
      {
        subst x; auto. 
        assert (z = zero).
        {
          simpl succ in H; apply LT_one in H.
          auto. 
        }
        constructor 1.
        subst. 
        red. 
        reflexivity. 
      }
      apply Hrec with i;eauto.
      apply lt_succ.
      eapply dS_plus_LT;eauto.
      auto. 
    }
Qed.


Corollary C12' : forall alpha,  nf alpha ->
                                forall beta i n, beta < alpha ->
                                                 (i <= n)%nat ->
                                                 dS_plus i alpha beta ->
                                                 dS_plus n alpha beta.
Proof.
  intros alpha H beta i n H0 H1 H2;  destruct (Lt.le_lt_or_eq _ _ H1).
  -   eapply C12;eauto.
  -   now subst.
Qed.


Corollary C12'' : forall alpha,
                    nf alpha ->
                    forall beta i n,
                      (i <= n)%nat ->
                      dS_star i alpha beta ->
                      dS_star n alpha beta.
Proof.
  intros.  
  destruct (T1_eq_dec  alpha beta).
  subst;now right. 
   destruct H1.
  assert (beta < alpha).
  apply dS_plus_LE in H1; auto.
  destruct (LE_LT_eq_dec H1);auto. 
    now destruct n0.
   left.
   apply C12' with i;auto.
   now destruct n0.
Qed. 

(* Lemma 1, Section 2.6 *)


Lemma L2_6_1 : forall alpha,  nf alpha ->
                           forall beta,
                             beta < alpha ->
                             {n:nat |  dS_plus n alpha beta}.
Proof.
  intros alpha ; transfinite_induction_LT alpha.
  intros x H Hx; destruct (zero_succ_limit_dec Hx).
  destruct s.
  subst;  intros;  not_neg H0.
  intros;
  destruct (dS_limit_S_strong Hx i H0) as [n1 H1].
  assert (dS n1 x < x). 
  { apply dS_LT; auto.
    intro; subst; discriminate. }
    assert (nf (dS n1 x)).
   eapply nf_dS;eauto. 
  destruct (H (dS n1 x) H2 H3 beta) as [n2 Hn2].
  auto.
  exists (Nat.max n1 n2).
  red;rewrite  <- clos_trans_t1n_iff.
  constructor 2  with (dS n1 x).
  rewrite  clos_trans_t1n_iff.
  apply C12' with n1; auto.
  auto with arith.   
  left.
  reflexivity. 
  rewrite  clos_trans_t1n_iff.
  apply C12' with n2; auto.
 
  auto with arith.  
  intros.
  destruct s.
  destruct a.
  subst.
  assert (Hx0: x0 < succ x0) by (apply LT_succ ;auto).
  assert ({beta < x0} + { beta = x0}).
  { apply LT_succ_le_2 in H0; eauto with T1.
 }  

  destruct H2. 
  destruct (H x0 Hx0 H1 beta l) as [n Hn].
  exists n.
  right with x0.
  red.  rewrite dS_succ. reflexivity.
  auto.
  assumption.
  exists 1.
  subst;left.
  red.
  rewrite dS_succ. reflexivity.
  auto.
Defined.


Lemma small_lemma (i:nat) : forall beta, forall alpha,
                              dS_plus i alpha beta ->
                              nf alpha -> 
                              beta <= dS i alpha .
Proof.
  intro beta; destruct (T1_eq_dec beta zero).
  -  subst; intros; apply LE_zero; auto with T1.
        apply nf_dS; auto with T1. 
  - induction 1.
   +  intros; red in H;  subst;  apply LE_refl; auto with T1. 
        apply nf_dS; auto with T1. 
   +  red in H;subst; intro; apply LE_trans with (dS i (dS i x)).
      * apply IHclos_trans_1n; auto.
        apply nf_dS; auto.
      * apply LE_r, dS_LT.
         { apply nf_dS; auto. }
         { intro H1; rewrite H1 in H0.
           apply dS_plus_zero in H0. 
           now destruct n.
         }
Qed.


Lemma dS_cons_not_zero : forall i alpha n beta,
                      alpha <> zero -> dS i (ocons alpha n beta) <> zero.
Proof. 
  destruct alpha.
  -  now destruct 1.
  -  intros; cbn.
     destruct n0, beta, alpha1, n; try discriminate. 
     all :   destruct (pred alpha2);  try discriminate. 
Qed.


Lemma L2_6_2 (p: nat)  :
  forall alpha, nf alpha ->
                forall beta, dS_plus p alpha beta ->
                             succ beta < alpha ->
                             dS_plus (S p) alpha (succ beta) .
Proof.                                                                
  intros  alpha ; transfinite_induction_LT alpha.
  clear alpha ; intros alpha HRecAlpha Halpha .
  intros;assert (beta <= dS p alpha).
  -   apply small_lemma; auto. 
  -   
    destruct (LE_LT_eq_dec H1).
    +   assert (succ beta <= dS p alpha).
     { apply   LT_succ_LE;auto. }
     destruct (LE_LT_eq_dec H2).
     *   specialize (HRecAlpha (dS p alpha) ).
         assert (dS p alpha < alpha).
         {   apply dS_LT; auto.
             intro; subst alpha; destruct (LT_zero H0).
         }
         apply dS_plus_trans  with (dS p alpha).
         { apply C12 with p; auto. 
           left;  reflexivity. }
         { apply HRecAlpha.
           apply dS_LT; auto with T1.
           intro; subst. 
           destruct (LT_zero H0).
            destruct H3;tauto.
           apply  Proposition_2_3a with alpha; eauto with T1.
           left; now red.   
           auto. 
         }
     * rewrite e;  apply C12 with p; auto. 
        apply dS_LT; auto with T1.
        intro; subst alpha. 
        destruct (LT_zero H0).
        left; now red.
    +  clear H1;  destruct alpha. 
       { destruct (LT_zero H0). }
       {  destruct n. 
          destruct alpha2.
          destruct (@zero_succ_limit_dec alpha1);  eauto with T1.
          destruct s.
          subst; cbn in H0.
          destruct (LT_irrefl H0).
          rename alpha1 into lambda.    
          rewrite dS_lim1 in e.
          assert (dS_plus  p lambda (dS p lambda)).
          left; now red.
          assert (dS_plus  (S p) lambda (succ (dS  p  lambda))).
          {
            apply HRecAlpha;auto. 
            apply head_LT_cons; eauto with T1.
            apply succ_lt_limit; auto. 
            apply dS_LT;auto with T1. 
            intro; subst; discriminate.
          }
          apply dS_plus_trans
          with (ocons (succ  (dS p lambda)) 0 zero); eauto with T1.
          apply Theorem_2_4_lemma5; auto.                                          
          intro; subst; discriminate.
          apply HRecAlpha; eauto with T1.
          apply LT2.
          apply single_nf.
          apply succ_nf.
          apply nf_dS.
          eauto with T1.
          apply single_nf; eauto with T1.
          apply succ_lt_limit; auto. 
          apply dS_LT;auto with T1. 
          intro; subst; discriminate. 
   apply succ_nf;  apply nf_dS; eauto with T1.
          subst; apply Theorem_2_4_lemma5; eauto with T1.
          left; red;   rewrite dS_succ; auto. 
          apply nf_dS; eauto with T1.
          apply succ_nf;  apply nf_dS; eauto with T1.
          apply succ_not_zero.
          rewrite e;  apply succ_lt_limit; auto. 
          apply single_nf; eauto with T1.
          apply succ_nf;  apply nf_dS; eauto with T1.
          simpl; case_eq (succ (dS p lambda)).
          intro; destruct (succ_not_zero (dS p lambda)); auto. 
          auto. 
          apply LT2; eauto with T1.
          apply single_nf; eauto with T1.
          apply nf_dS; eauto with T1.
          apply single_nf; eauto with T1.
          apply succ_nf;  apply nf_dS; eauto with T1.
          apply LT_succ.
          apply nf_dS; eauto with T1.
          eauto with T1.
          auto.
          
          (* end of the is_limit case *)
          {
            destruct s as [gamma [H1 H2]];destruct (T1_eq_dec gamma zero).
            {
              (** cas limite *)
              subst gamma alpha1; simpl in e;   subst beta;    left. 
              red;   rewrite dS_succ_eqn1;    trivial.
            }
            assert ( beta = ocons gamma p zero).
            { rewrite e, H2; now rewrite dS_succ_eqn1. }
            apply dS_plus_trans with (ocons gamma (S p) zero).
            { left;red; rewrite H2; now rewrite dS_succ_eqn1. }
            apply HRecAlpha.
            - subst alpha1; apply LT2. 
              apply single_nf; auto with T1.
              apply single_nf; eauto with T1.
              apply LT_succ; auto. 
            -     apply single_nf;auto. 
            -  rewrite H3.
               apply Theorem_2_4_lemma3_0; auto. 
            -
               apply succ_lt_limit; eauto with T1.
              cbn;  destruct gamma.
               +   now destruct n.
               +   auto. 
               +  rewrite H3; apply LT3; auto with arith.
          }
          { (** recursive case on tail :  *)
            rewrite dS_tail in e.
            - subst beta; destruct (T1_eq_dec alpha2_1 zero).
              {
                subst alpha2_1;  assert (alpha2_2 = zero).
                { 
                apply nf_inv2 in Halpha; now  apply nf_of_finite in Halpha.
                }  
                subst alpha2_2; simpl dS; destruct n.
                simpl in H0.
                destruct alpha1.
                destruct H0.    
                destruct H1.
                discriminate H1.
                destruct (LT_irrefl H0).
                simpl in H0.
                destruct alpha1.
                destruct n.
                generalize (nf_of_finite Halpha); discriminate.
                generalize (nf_of_finite Halpha); discriminate.
                destruct (LT_irrefl H0).
              }
            rewrite succ_ocons.
            + apply Theorem_2_4_Lemma1; eauto with T1.
              discriminate.
              apply HRecAlpha; eauto with T1.
              apply LT2; eauto with T1.
              split; eauto with T1.
              left; now red.
              rewrite succ_ocons in H0.
              apply remove_first_sumand in H0.
              auto.
              destruct alpha2_1.
              destruct n0; auto.
              { cbn. destruct n, alpha2_2, alpha2_1_1, n1;
                  try discriminate.
                all:  destruct (pred alpha2_1_2); try discriminate.
              }
              apply nf_LT_right with (ocons alpha2_1 n alpha2_2).
              eapply Halpha.
              apply dS_LT;  eauto with T1.
              discriminate.
            + eapply dS_cons_not_zero; eauto. 
            +  eapply nf_LT_right; eauto.
               apply dS_LT;  eauto with T1.
               discriminate. 
           -  eauto with T1. 
           -  discriminate. 
          }
          destruct (T1_eq_dec alpha2 zero).
          { 
            subst alpha2;  rewrite dSSn in e.
            destruct (T1_eq_dec alpha1 zero).
            {
              subst alpha1; simpl dS in e;  subst beta;
              simpl in H0.
              destruct (LT_irrefl H0).
            }
            rewrite e,  succ_ocons.
            apply Theorem_2_4_Lemma1'.
            eauto with T1.
            auto. 
            apply HRecAlpha.
            unfold phi0; apply LT3; auto with arith. 
            eauto with T1.
            left;red. 
            reflexivity. 
            apply succ_lt_limit; eauto with T1.
            unfold phi0.
            cbn;   destruct alpha1; auto.
            unfold phi0; apply dS_LT; eauto with T1.
            discriminate. 
            apply dS_cons_not_zero; auto.
            apply nf_intro. 
            eauto with T1.
            apply nf_dS; eauto with T1.
            apply lt_phi0_phi0R.
            destruct (@dS_LT p (ocons alpha1 0 zero)).
            eauto with T1.
            discriminate. 
            destruct H2;auto. 
            eauto with T1. 
          }
          
          rewrite dS_tail in e; auto.
          rewrite e; simpl succ; destruct alpha1.
          apply nf_of_finite in Halpha.
          now destruct n0.
          apply   Theorem_2_4_Lemma1; auto. 

          rewrite e in H0.
          cbn in H0; apply remove_first_sumand in H0.
          apply HRecAlpha; auto. 
          apply tail_LT_cons; auto. 
          eauto with T1.
          left; now red. 
       }
       Unshelve.
       exact 42.
Qed. 



Section Lemma_4_3_Proof.
  Variables alpha beta : T1.
  Hypothesis H00 : alpha <> zero.
  Hypothesis nf1 : nf alpha.
  Hypothesis nf2 : nf beta.
  Variables n0 n1 n2: nat.
  Hypothesis H0 : (n0 <=n2)%nat.
  Hypothesis H1 : (n1 <=n2)%nat.
  Hypothesis H4 : dS_star n0 alpha beta.

  Remark R4_3_1 : dS_star n2 alpha beta.
  Proof.
   destruct H4.
  - left;  apply C12' with n0; auto. 
     eapply dS_plus_LT;eauto.    
   - subst;now right. 
 Qed. 

  Remark R4_3_2 : dS_star n2 beta (dS n1 beta).
  Proof.
    destruct (T1_eq_dec beta zero).
    - subst;  cbn;  now right. 
    -  left;  apply C12' with n1; auto. 
       + apply dS_LT; auto. 
       +  left; reflexivity. 
  Qed. 

  Lemma dS_LE : forall alpha n, nf alpha -> dS n alpha <= dS (S n) alpha.
  Proof.
    intros;destruct (zero_succ_limit_dec H).
    -   destruct s. 
        +  subst; cbn. 
           apply LE_refl; auto with T1.
        + apply LE_r; eapply dS_limit_mono; auto.
    - destruct s as [gamma [Hgamma Hgamma']]; subst alpha0.
      repeat rewrite dS_succ;auto with T1.
      apply LE_refl;auto. 
  Qed. 


  Remark R4_3_3: dS_star n2 alpha (dS n1 beta).
  Proof.
   eapply dS_star_trans with beta; auto.
   - apply R4_3_1 .
   - apply R4_3_2.
  Qed. 

   Remark R4_3_4 : dS n1 beta <= dS n2 alpha.
   Proof.
     destruct (T1_eq_dec alpha beta).     
     - subst.
      clear H4 H0; induction H1.
      +  apply LE_refl.
         now apply nf_dS.
      +  apply LE_trans with  (dS m beta);  auto. 
         apply dS_LE; auto.
     - assert (beta < alpha).
        { apply  dS_star_LE_2 in H4;auto. 
          destruct H4.
          destruct H2. 
          split; auto. 
          split;auto. 
          destruct (le_lt_eq beta alpha); auto. 
          - subst.
            now destruct n.
        }
        assert (dS_plus n2 alpha (dS n1 beta)).
        {
          destruct R4_3_3.  auto.
          subst.
          destruct (@LT_irrefl beta).
          apply LT_trans with (dS n1 beta);  auto.      
          apply dS_LT   ;auto. 
          intro;subst; cbn in H00.
          now destruct H00.
        }
       inversion H2.
       red in H3; subst.
       rewrite H3;  apply LE_refl.
       apply nf_dS;auto.
       red in H3;  subst y. 
       eapply dS_plus_LE; eauto. 
       eapply nf_dS;eauto.  
   Qed. 

 Lemma Lemma_4_3_0 : dS_star n2 (dS n2 alpha) (dS n1 beta).
 Proof. 
   destruct (LE_LT_eq_dec R4_3_4).
   - left.
     eapply Proposition_2_3a with (alpha := alpha); auto.   
     destruct R4_3_3; auto. 
     rewrite <- H in l; destruct (@LT_irrefl alpha).
     apply LT_trans with (dS n2 alpha); auto. 
     apply dS_LT;auto. 
     left; red;auto. 
   - rewrite e;  now right.
   Qed. 
 
 
End Lemma_4_3_Proof.


Lemma Lemma4_3 :
  forall alpha beta : T1,
   nf alpha ->
   nf beta ->
   forall n0 n1 n2 : nat,
   (n0 <= n2)%nat ->
   (n1 <= n2)%nat ->
   dS_star n0 alpha beta -> 
   dS_star n2 (dS n2 alpha) (dS n1 beta).
Proof.
 intros.  
 destruct (T1_eq_dec alpha zero).
 subst.
  destruct H3.
 rewrite (dS_plus_zero H3).
 cbn.
 now right.
 subst.
 cbn. now right.
  eapply  Lemma_4_3_0 with n0;eauto.
Qed.


Lemma dS_plus_LT' (i : nat) (alpha beta : T1) (H:nf alpha):
  dS_plus i alpha beta -> alpha = zero \/ beta < alpha.
Proof.
  intro H0; destruct (T1_eq_dec alpha zero).
  - now left.
  - right.
    eapply dS_plus_LT; eauto.
Qed.

