(**   Pierre Casteran 
      LaBRI, Université de Bordeaux
*)

Require Import Arith Max Omega  List  Relation_Operators  Operators_Properties
Peano_dec.
Require Import Epsilon0  NaturalSum  insertion_sort Simple_LexProd.
Require Import Coq.Wellfounded.Inverse_Image Coq.Wellfounded.Inclusion.
Require Import WfVariant.
Import Relations.
Open Scope nat_scope.


(** * Data types 

 An hydra is a finitely branched tree. We use an auxiliary type : Hydrae  for
 repesenting lists of hydras.
*)

Inductive Hydra : Set :=
|  node :  Hydrae -> Hydra
with Hydrae : Set :=
| hnil : Hydrae
| hcons : Hydra -> Hydrae -> Hydrae.

Check Hydra_rect.
(*
Hydra_rect
     : forall P : Hydra -> Type,
       (forall h : Hydrae, P (node h)) -> forall h : Hydra, P h
*)

Check Hydra_ind.

(** *** Abbreviations 
 *)
Notation head := (node hnil).

(** hydra with 1, 2 or 3 daugthers 
*)
Notation hyd1 h := (node (hcons h hnil)).
Notation hyd2 h h' := (node (hcons h (hcons h' hnil))).
Notation hyd3 h h' h'' := (node (hcons h (hcons h' (hcons h'' hnil)))).

(** Adds n_copies of the same hydra h to s 
*)

Fixpoint hcons_mult (h:Hydra)(n:nat)(s:Hydrae):Hydrae :=
  match n with 
  | O=> s
  | S p => hcons h (hcons_mult h p s)
  end.


(** hydra with n equal daughters 
*)

Definition hyd_mult h n :=
  node (hcons_mult h n hnil).

(**  Hydrae of height 2 
*)

Definition flower_S i := hyd_mult head (S i).


(** ** Mutual induction scheme for types Hydra and Hydrae 
 *)

Scheme Hydra_rect2 := Induction for Hydra Sort Type
with   Hydrae_rect2 := Induction for Hydrae Sort Type.


(**  Number of nodes 
 *)

Fixpoint hsize (h:Hydra) : nat :=
  match h with node l => S (lhsize l)
  end
with lhsize l : nat :=
  match l with hnil => 0
            | hcons h hs => hsize h + lhsize hs
  end.

(** Exercise : redefine hsize as an application of Hydra_rect2 *)

About Hydra_rect2.

Definition hsize2 := Hydra_rect2 (fun _ => nat) (fun _ => nat) (fun h i => 1 + i) 0
      (fun h n s n0 => n + n0).

About hsize2.


Check hsize2 : Hydra -> nat.


Compute hsize2 (flower_S 7).
Compute hsize (flower_S 7).

(**  height 
 *)

Fixpoint hheight  (h:Hydra) : nat :=
  match h with node l => lheight l
  end
with
lheight l : nat :=
  match l with hnil => 0
            | hcons h hs => Max.max (S (hheight h)) (lheight hs)
  end.

Definition hheight2 := Hydra_rect2 (fun _ => nat) (fun _ => nat) 
                                   (fun h i => i) 0
                                   (fun h n s n0 => Max.max (S n) n0).



Example Hy := hyd3 head
                   (hyd2
                      (hyd1 
                         (hyd2 head head))
                      head) 
                   head.


(** A failed attempt 
 *)

Lemma height_lt_size (h:Hydra) :
  hheight h <= hsize h.
Proof.
  induction h as [s]. 
  induction  s as [ | h s'].
  -   cbn; auto with arith.
  -   (** No induction hypothesis on h *)
Abort.
Check Hydra_rect.
(** Let us build an auxiliary predicate on sequences of hydras 
*)

Fixpoint h_forall (P: Hydra -> Prop) (s: Hydrae) :=
  match s with
      hnil => True
    | hcons h s' => P h /\ h_forall P s'
  end.

Lemma max_le_plus (n p: nat) : Nat.max n p <= n + p.
Proof. 
  destruct (Nat.le_ge_cases n p). 
  - rewrite (max_r _ _ H); auto with arith.
  - rewrite (max_l _ _ H); auto with arith. 
Qed.    



Lemma  height_lt_size (h:Hydra) :
  hheight h < hsize h.
Proof.
  induction h using Hydra_rect2  with 
  (P0 :=  h_forall (fun h =>  hheight h < hsize h)).
  -  rename h into s.
     destruct s as [ | h s'].
     + cbn; auto with arith.
     +  simpl.  destruct IHh; assert (lheight s' <= lhsize s').
        { clear H; induction s'. 
          -     cbn; auto with arith. 
          -     simpl.  destruct (lheight s').
                + cbn in H0; destruct H0; apply IHs' in H0 .
                  red in H;  transitivity (hsize h0); auto.
                  auto with arith. 
                +  
                  cbn in H0; destruct H0. 
                  apply IHs' in H0.    clear IHs'.            
                  rewrite succ_max_distr; 
                  transitivity (S (hheight h0) + (S n) ).
                  apply max_le_plus; auto.    cbn.
                  omega.
        }
         clear H0; cbn; destruct (lheight s').
  *   omega. 
  *   specialize (max_le_plus (hheight h) n); omega.
  - reflexivity.
  -  split;auto. 
Qed. 

(** Exercise 

Write a tactic for proving statements of the form 
forall h: Hydra, P h, that automatizes the call to Hydra_rect2 
*)

Ltac h_induct h :=
  pattern h; match goal with |- ?P h =>
                             induction h using Hydra_rect2  with 
                             (P0 :=  h_forall P);
                               [ | cbn; trivial | split; auto]
                               end.


Lemma  height_lt_size' (h:Hydra) :  hheight h < hsize h.
Proof.
  h_induct h.
   simpl.
  rename h into s.   destruct s as [ | h s'].
  cbn ; auto.
  destruct IHh; assert (lheight s' <= lhsize s').
  {
    clear H; induction s'. 
    -     cbn; auto with arith. 
    -     simpl;  destruct (lheight s').
          + cbn in H0; destruct H0; apply IHs' in H0 .
            red in H;  transitivity (hsize h0); auto.
            auto with arith. 
          +  
            cbn in H0; destruct H0. 
            apply IHs' in H0.    clear IHs'.            
            rewrite succ_max_distr; 
              transitivity (S (hheight h0) + (S n) ).
            apply max_le_plus; auto.    
            omega.
  }
  clear H0; simpl; destruct (lheight s').
  +  omega. 
  +  specialize (max_le_plus (hheight h) n); omega.
Qed.

(** Technical lemmas *)

  Fixpoint hy_app (s1 s2: Hydrae) : Hydrae :=
      match s1 with
          hnil => s2
        | hcons h1 s1' => hcons h1 (hy_app s1' s2)
      end.

    Lemma hy_app_assoc :
      forall s1 s2 s3,
        hy_app s1 (hy_app s2 s3) = hy_app (hy_app s1 s2) s3.
    Proof.
      induction s1.
      - reflexivity.
      - intros s2 s3; cbn; now rewrite IHs1.
    Qed.


(** *  Hydra Battles

  We represent the rules of Hydra Battles through a binary relation 
  [round] on the [Hydra] type.
  [round h h'] holds if [h'] is obtained from h by a single turn of a battle 
 
 *)

(**  
 S0 s s' holds if s' is obtained by removing one head from s
*)

Inductive S0 :  relation Hydrae :=
| S0_first : forall s, S0  (hcons head s) s
| S0_rest : forall  h s s',  S0  s s' ->
                             S0  (hcons h s)
                                 (hcons h s').

(** 
   R1 h h' holds if h' is obtained by removing one head from h at distance 1 from
   the foot *)

Inductive R1  :  Hydra -> Hydra -> Prop :=
| R1_intro : forall s s', S0 s s' -> R1 (node s) (node s').


Example ex1 : forall h h', R1 (hyd3 h head h')
                              (hyd2 h h').
 Proof. split.  right; left. Qed.


 (** 
  S1 n s s' holds if an item h of s has been replaced by (n+1) copies of h',
   where R1 h h'.
  Thus, n is the number of new replicas of h' 
 *)
  
Inductive S1 (n:nat)  : Hydrae -> Hydrae -> Prop :=
| S1_first : forall s h h' ,
                        R1 h h' -> 
                        S1 n (hcons h s) (hcons_mult h' (S n) s)
| S1_next : forall h s s',
                       S1 n s s' ->
                       S1 n (hcons h s) (hcons h s').


Example rep0 : forall  s h h', R1 h h' -> S1 0 (hcons h s) (hcons h' s).
Proof.
  intros  s h h' H;
  change (hcons h' s) with (hcons_mult h' 1 s);
  now left.
Qed.

(**  R2 n h h' holds if some sub-hydra h0 of h has been replaced by h'0,
  where R1 n h0 h'0

*)

Inductive R2 (n:nat)  :  Hydra -> Hydra -> Prop :=
| R2_intro : forall s s', S1 n s s' -> R2 n (node s) (node s')
| R2_intro_2 : forall s s', S2 n s s' -> R2 n (node s) (node s')

with S2 (n:nat) :  Hydrae -> Hydrae -> Prop :=
|  S2_first : forall h h' s , R2 n   h h'->
                           S2  n (hcons h s) (hcons h'  s)

|  S2_next  : forall h   r r', S2 n   r r' ->
                               S2  n (hcons h r) (hcons h r').

(** Relations associated with one round *)

Definition round_n n h h' := R1 h h' \/ R2 n h h'.

Definition round h h' := exists n,  round_n n h h'.

Infix "-1->" := round (at level 60).

(** Relation associated with one or more round_plus *)

Definition round_plus := clos_trans_1n Hydra round.

Definition round_star h h' := h = h' \/ round_plus h h'.

Infix "-+->" := round_plus (at level 60).
Infix "-*->" := round_star (at level 60).

Lemma round_star_intro h h'' : forall h', h -1-> h' -> h' -*-> h'' -> h -*-> h''.
Proof. 
  destruct 2.
  - subst; right; now left .
  -  right; now right with h'.
Qed.


 Lemma S0_app : forall s1 s2 s3, S0 s2 s3 ->
                                    S0 (hy_app s1 s2)
                                       (hy_app s1 s3).
    Proof.
      induction s1.
      - cbn; auto.  
      - intros; cbn;  constructor;auto.
    Qed.

    Lemma R1_app : forall s1 s2 s3,
                     R1 (node s2) (node s3) ->
                     R1 (node (hy_app s1 s2)) (node (hy_app s1 s3)).
    Proof.
      constructor.
      inversion  H; now apply S0_app.
    Qed.


Lemma S1_app n : forall s1 s2 s3,
                       S1 n s2 s3 -> S1 n (hy_app s1 s2)
                                        (hy_app s1 s3).
    Proof.
      induction s1; simpl hy_app.
      - trivial.  
      - intros s2 s3 H; constructor; auto.
    Qed.


    
(** Induction Schemes for R2 *)

Scheme R2_ind2 := Induction for R2 Sort Prop
                  with
                  S2_ind2 := Induction for S2 Sort Prop.

    Lemma S2_app : forall  n  s1 s2 ,
                     S2 n s1 s2 ->
                     forall l,  S2 n (hy_app l s1) (hy_app l s2).
    Proof.
      intros n s1 s2 h;
      eapply  S2_ind2 with
      (P := fun h1 h2 _ => match h1, h2 with
                               node s1, node s2 =>
                               forall l, R2 n (node (hy_app l s1))
                                            (node (hy_app l s2)) 
                           end).
      - intros; constructor.
        apply S1_app; auto.
        intros;  eauto. 
      - constructor 2.
        induction l.
        + cbn; auto.
        + cbn; now constructor 2.
      -  intros h0 h' s r  H l;  destruct h0, h';  induction l.   
         + simpl;auto.
         +  simpl;auto; constructor 2;auto.
      - auto.
      -  eexact h.
    Qed.



    Lemma R2_app n : forall  s1 s2 ,
                       R2 n (node s1) (node  s2) ->
                       forall l,  R2 n (node (hy_app l s1))
                                     (node (hy_app l s2)).
    Proof.
      intros s1 s2 H;  inversion  H.
      - intros; constructor.
        apply S1_app; auto.
      -  constructor 2; now apply S2_app.
    Qed.


(** Tactics for interactive rounds *)


(* removes the n-th daughter (should be a head) *)

Ltac hremove n :=
  match goal with |- S0 ?h ?h' =>
                  match n with
                    | O => eapply S0_first
                    | S ?p => eapply S0_rest; hremove p
                  end
  end.

Ltac round_1 i :=
  match goal with |- round ?h ?h' =>
                  exists 0;  left; split; hremove i
  end.
                                                 
Ltac round_2 n  :=  match goal with
                        |- round ?h ?h' =>  exists n; eright
                    end.
                         
Ltac S2_nth n :=
  match n with
    | 0 => eleft
    | S ?p => eright ; S2_nth p
    end.


Ltac R2_up i := match goal with
                    |- R2 ?n ?h ?h' => eright; S2_nth i 
end.



Ltac S1_nth i :=
  match goal with
      |- S1 ?n ?h ?h' =>
      match i with
        | 0 => eleft
        | S ?j => eright ; S1_nth j
      end
  end.

Ltac R2_here i :=
 match goal with
      |- R2 ?n ?h ?h' =>
      eleft; S1_nth i; split
 end.

Ltac stop :=
  match goal with
      |- round_star ?h ?h' => left; reflexivity
  end.

Ltac forward :=
  match goal with
      |- round_star ?h ?h' => eapply round_star_intro
  end.


Module Examples.

  Example Hy := hyd3 head
                   (hyd2
                      (hyd1 
                         (hyd2 head head))
                      head) 
                   head.

Example Hy' := hyd2 head
                   (hyd2
                      (hyd1 
                         (hyd2 head head))
                      head).




Example ex4:  round Hy Hy'.
    Proof.  round_1 2. Qed.


    



                           
Example Hy'' := 
                  hyd2 head
                      (hyd2
                         (hyd_mult (hyd1 head) 5)
                         head).
Goal round Hy' Hy''.
round_2 4 .
R2_up 1.
R2_up 0.
R2_here 0.
hremove 1.
Qed.

Example R2_example:  R2 4 Hy' Hy''.
Proof.
  right;  right;  left;  right; left; left; left; split; left.
Qed.


Goal {h' | round Hy' h'}.
eexists.
round_2 4 .
R2_up 1.
R2_up 0.
R2_here 0.
hremove 1.
Defined. 


Example Ex5 : {h' | Hy -*-> h'}.
eexists.
forward.
round_1 2.
forward.
round_2 4.
R2_here 1.
hremove 1.
stop.
Defined.


Example Hy''' :=
  node (hcons head
              (hcons_mult (hyd1 (hyd_mult (hyd1 head) 5)) 3 hnil)).




Example hsize_bigger : (hsize Hy'' < hsize Hy''')%nat.
Proof. compute. omega.  Qed.


Example hheight_not_strictly_decreasing : hheight Hy'' = hheight Hy'''.
Proof. reflexivity.  Qed.


Example Hy_1 : R1 Hy Hy'.
Proof. 
  split ; right; right; left.
Qed.




      
Example Hy_2 : R2 4 Hy' Hy''.
Proof.
  right.  right. left.  right; left;  left; left.
  split; hremove 0%nat.
Qed.



Example ex_2 :{Hy'' | R2 4 Hy' Hy''}.
eexists.
(*  R2 4 Hy' ?Hy'' *)

unfold Hy'. (* not necessarty *)
(*  R2 4 (hyd2 head (hyd2 (hyd1 (hyd2 head head)) head)) ?Hy'' *)
(* Go to the second daugther (index 1) *)
R2_up 1.

(*   R2 4 (hyd2 (hyd1 (hyd2 head head)) head) ?Goal0 *)

(* Go to the first daugther (index 0) *)
R2_up 0. 

 (* R2 4 (hyd1 (hyd2 head head)) ?Goal1 
  stop recursion (we want to chop an head at level 1)
*)

R2_here 0.

(*  S1 4 (hcons (hyd2 head head) hnil) ?Goal0 *)

(* R1 (hyd2 head head) ?h' *)

 hremove 0%nat.
      Defined
      . 


Example Hy_3 : R2 2 Hy'' Hy'''.
Proof.
 left;right;left;split;right;left.
Qed.

Example Hy''_Hy''' : Hy'' -1-> Hy'''.
Proof.
 exists 2;  right; apply  Hy_3.
Qed.

Example Hy_Hy''' :  Hy  -+-> Hy'''.
Proof.
  eright.
  - eexists;  left; apply Hy_1.
  -  eright.
     + eexists.  right; apply Hy_2.
     + left.  eexists; right;apply Hy_3.
       Unshelve. 
       exact 0. 
Qed.

End Examples.



Lemma round_plus_trans : forall h h' h'', h -+-> h' ->
                                      h' -+-> h'' ->
                                      h -+-> h''. 
Proof.
  unfold round_plus.
  intros h h' h''; repeat rewrite <- clos_trans_t1n_iff.
  econstructor 2; eauto. 
Qed.


Lemma round_plus_height : forall h h', h -+-> h' ->
                                   hheight h' <= hheight h.
Proof. Abort. (*exercise *)

(** ** Reasoning about hydra battles 
*)




Lemma hcons_mult_S0 : forall h  i s s' , 
                             S0 s s' ->
                             S0 (hcons_mult h i s)
                                (hcons_mult h i s').
Proof.
    induction i.
    - now  simpl.
    - right; auto.
Qed.


Lemma hcons_mult_S1 n : forall h i  s s' , S1 n s s' ->
                                         S1  n (hcons_mult h i s)
                                              (hcons_mult h i s').
Proof.
  induction i.
  - now  simpl.
 -  right;auto.
Qed.

Lemma hcons_mult_S2 n : forall h  i s s' , S2 n s s' ->
                                            S2 n (hcons_mult h i s)
                                               (hcons_mult h i s').
Proof.
  induction i.
  - now  simpl.
  - right;auto.
Qed.


Lemma hcons_mult_comm :
  forall i h s,
    hcons_mult h i (hcons h s) = hcons h (hcons_mult h i s).
Proof.
   induction i; simpl; auto.
   intros; now rewrite IHi.
Qed.



Lemma head_dec : forall h, {h= head}+{h <> head}.
destruct h  as [l].
destruct l as [| h l].
- now left.
- right; discriminate. 
Defined.


(** If the hydra is not reduced to a head, there exists at least one that
    Hercules can chop  *)

Lemma liveliness n :
 forall h , h <> head -> exists (h' : Hydra),  round_n n h  h'. 
Proof.
induction h using   Hydra_rect2 with
(P0 :=
   fun hs => hs <> hnil -> 
             exists hs', S0 hs hs' \/  S1 n hs hs' \/ S2 n hs hs' ).
 - intros H; destruct h.
   + now destruct H.
   +  destruct IHh.
      * discriminate.
      * exists (node x).  
        destruct H0 as [H0 | [H0 | H0]].
        {  left ; now split. }
        { right; now left. }
        {  right; now right. }
   - now destruct 1.
   - intros _;  destruct h0 as [| h1 hs].
     + destruct (head_dec h).
       *  subst h;  exists hnil; left;  left.
       *  destruct IHh as [x Hx]; [ assumption|].
          destruct Hx.
         { exists (hcons_mult  x (S n) hnil); right;  left.
           change (hcons x hnil) with (hcons_mult x 1 hnil).
           constructor 1. auto. 
         }
         { exists (hcons_mult x 1 hnil); right;right; now  left. }
     +   destruct IHh0.
         * discriminate.
         * destruct H.
           {  exists (hcons h x); left; now right. }
           { destruct H.
              - exists (hcons h x); right; left; now right.
             -  exists (hcons h x); right;right; now right.
           }
Qed.


Definition  left_lowest_0 n :
 forall h , (h = head) + {h' : Hydra &   {R1 h h'} + {R2  n h  h'} }.
Proof.
induction h using   Hydra_rect2 with
(P0 :=
   (fun hs => (hs = hnil) +
             { hs' : Hydrae & ( S0 hs hs') + {S1 n hs hs'} + {S2 n hs hs'}})%type).
rename h into s.
- destruct s.
  + now left.
  + right.
     destruct IHh.
     * discriminate.
     * destruct s0 as [hs' H0].
        destruct H0.
        destruct s0.
        (** chose the a lowest head *)
        exists (node hs').
        left.
        now constructor.
        exists (node hs').
        right; now left.
        exists (node hs').
         right;now right.
   - now left.
   - destruct IHh.
     right.
      exists h0.
      left.
      left.
      (* chose the leftmost head *)
      subst h; left.
      right.
      destruct s as [h' [H1 | H1]].
         exists (hcons_mult h' (S n)  h0).
         left; right.      
         now constructor.           
         exists (hcons h' h0).
         right.
         now constructor.
         Defined.

Definition round_spec h n :=
  ((h = head) + {h' : Hydra |  round_n n h h'})%type.


Definition  left_lowest  n :
 forall h , round_spec h n.
Proof.
  intro h; destruct (left_lowest_0 n h).
  - now left.
  - destruct s as [h' [H | H]].
    + right; exists h'; now left.
    + right; exists h'; now right.
Defined.


Definition next_step (f : forall n h, round_spec h n) n h :=
  match f n h with
      inl _ => head
    | inr (exist _ h' _) =>  h'
  end.

Compute next_step left_lowest 5 Hy.


Definition classic_battle f t h :=
  let fix go n t h := 
      match h, t with head, _ => h
                   |   _, 0 => h
                | _, S t' => go (S n) t' (next_step f n h)
      end
  in go 1 t h.


About classic_battle.
Compute hsize (classic_battle  left_lowest 0 Hy).
Compute hsize (classic_battle  left_lowest 1 Hy).
Compute hsize (classic_battle  left_lowest 2 Hy).
Compute hsize (classic_battle  left_lowest 3 Hy).
Compute hsize (classic_battle  left_lowest 4 Hy).

Compute hsize (classic_battle  left_lowest 5 Hy).
Compute hsize (classic_battle  left_lowest 10 Hy).
Compute hsize (classic_battle  left_lowest 20 Hy).

Compute hsize (classic_battle  left_lowest 30 Hy).



Extraction Language Ocaml.
Recursive Extraction classic_battle left_lowest.


(* ** Proof of termination of all hydra battles *)


(** converting an hydra into an ordinal notation in Cantor normal form  

    Notice the use of Hessenberg natural sum [alpha o+ beta ]

  *)

Fixpoint m (h:Hydra) : T1 :=
  match h with head => zero
             | node hs => ms hs
end 
with ms (s:Hydrae) :  T1 :=
  match s with  hnil => zero
              | hcons h s' => phi0 (m h) o+  ms s'
 end.

(* for rewriting ... *)

Lemma ms_eqn2 :  forall h s, ms (hcons h s) = phi0 (m h) o+  ms s.
Proof.   reflexivity. Qed.

Lemma o_finite_mult_S_rw :
  forall n a, o_finite_mult (S n) a =  a o+ o_finite_mult n a.
Proof.   reflexivity.  Qed.

(**  The functions [m] and [ms] return well formed ordinals (less than epsilon0)
 *)

Theorem m_nf : forall h, nf (m h).
Proof.
 intro h; elim h using Hydra_rect2 
            with (P0 := fun s =>  nf (ms s)).
 -  destruct h0; simpl; auto.
 -  constructor.
 - intros;  rewrite ms_eqn2; apply oplus_nf.
  + now apply nf_phi0.
  + assumption. 
Qed.

Theorem ms_nf : forall s, nf (ms s).
Proof with auto with T1.
 induction s...
  rewrite ms_eqn2...
  apply oplus_nf...
  apply nf_phi0; now  apply m_nf.   
Qed.   

Hint Resolve m_nf nf_phi0 ms_nf : T1.

Lemma ms_eqn3 :  forall h n s,  ms (hcons_mult h  n s) =
                                    o_finite_mult n (phi0 (m h)) o+ ms s.
Proof with auto with T1.
 induction n.
 - intros; simpl (hcons_mult h  0 s); simpl o_finite_mult.
   rewrite oplus_0_b; reflexivity.
 -  intros; simpl hcons_mult;  rewrite ms_eqn2.
    rewrite o_finite_mult_S_rw, IHn, oplus_assoc ...
    apply o_finite_mult_nf ...
Qed. 

(** Monotonicity lemmas for S0, R1, etc . *)
Open Scope cnf_scope.

Lemma S0_decr_0 :
  forall s s', S0  s s' -> lt (ms s') (ms s).
Proof with auto with T1.
 induction 1.
 -  rewrite ms_eqn2; simpl phi0; apply oplus_lt2...
    compute...
 -  repeat rewrite ms_eqn2; apply oplus_strict_mono_r ...
 Qed.


Lemma S0_decr :  forall s s', S0  s s' -> ms s' < ms s.
Proof.
  repeat split; auto with T1; now apply S0_decr_0.
Qed.

Lemma R1_decr_0 : forall h h',
                  R1 h h' -> lt (m h') (m h).
Proof with auto with T1.
  destruct 1;  simpl;  destruct s, s'.
  -  inversion H.
  -  inversion H.
  -  rewrite ms_eqn2; apply lt_le_trans with (phi0 (m h))...
     +  unfold phi0;constructor.
     +  apply oplus_le...
  -  now apply   S0_decr. (* simpl in H  en V8.6 ??? *)
Qed.

Lemma R1_decr : forall h h',
                  R1 h h' -> m h' < m h.
Proof.
  repeat split; auto with T1; now apply R1_decr_0.
Qed.


Lemma S1_decr_0 n:
  forall s s', S1 n s s' -> lt (ms s') (ms s).
Proof with auto with T1.
  induction 1.
   -  rewrite  ms_eqn2.
      simpl hcons_mult.
      rewrite  ms_eqn2.
      rewrite ms_eqn3.
      rewrite oplus_assoc...
      apply oplus_strict_mono_l...
      apply oplus_nf...
      apply o_finite_mult_nf...
      rewrite <- o_finite_mult_S_rw.
      apply o_finite_mult_lt_phi0_1.
      now apply R1_decr.
      apply o_finite_mult_nf...
   -  repeat rewrite ms_eqn2 ...
      apply oplus_strict_mono_r ...
Qed.


Lemma S1_decr n:
  forall s s', S1 n s s' -> ms s' < ms s.
Proof.
  repeat split; auto with T1; now eapply S1_decr_0 with n.
Qed.

Lemma m_ms : forall s, m (node s) = ms s.
Proof. 
 destruct s; reflexivity.
Qed.

Lemma R2_decr_0 n : forall h h', R2 n h h' -> lt (m h') (m h).
Proof with auto with T1.
  induction 1 using R2_ind2 with
  (P0 := fun s s' (H : S2 n s s') => lt (ms s') (ms s))...
  - repeat rewrite m_ms.
    now  apply S1_decr with n.
  - now repeat rewrite m_ms.
  -  apply oplus_strict_mono_l ...
     apply phi0_mono_strict;  auto.
  - repeat rewrite ms_eqn2.
    apply oplus_strict_mono_r...
Qed.

Lemma R2_decr n : forall h h', R2 n h h' -> m h' <  m h.
Proof.
  repeat split; auto with T1; now eapply R2_decr_0 with n.
Qed.



Lemma round_decr : forall h h', h -1-> h' -> m h' < m h.
Proof.
destruct 1 as [n H]; destruct H. 
-  now apply R1_decr.
-  now apply R2_decr with n.
Qed.


Global Instance HydraVariant : WfVariant round LT m .
Proof. 
split.
-  apply T1_wf.
-  apply round_decr.
Qed.

About Variant_ok.
Theorem Every_battle_terminates :  terminates  round.
Proof. 
  apply (Variant_ok  HydraVariant).
Qed. 









