(** Hydra Battles *)

(* Message utilisé lorsqu'un bug est détecté – par exemple, dans les fonctions de réplication *)

let the_msg = "ouah, le bug!"

(* Une hydre est représentée sous la forme d'un arbre enraciné dont les noeuds peuvent avoir un nombre quelconque mais néanmoins fini de filles. *)

type hydra = Node of hydra list

(* Quelques abréviations  simples *)

(* Hydre à une seule tête *)
let head = Node []
let is_head h = h = head

(* Nœuds à 1, 2 ou 3 filles *)
let single h = Node [h]
let bi h1 h2 = Node [h1;h2]
let tri h1 h2 h3 = Node [h1;h2;h3]

(* Idem, lorsque les filles sont identiques *)
let bisame h = bi h h
let trisame h = tri h h h

(* Liste des filles d'un nœud *)
let les_filles (Node hs) = hs

(* Exemples d'hydres *)

let baby_hydra = single head
let very_small_hydra = bi head head
let small_hydra = single (bi head head)
let my_hydra = tri head head  (single (single (tri head head head)))
let another_hydra = single (tri head head head)
let yet_another_hydra = single small_hydra
let goodstein_hydra = tri head head (tri head head head)

(* Exemple du sujet, page 1 *)
let example_hydra = tri head (bi (single (tri head head head)) head) head
(* Exemple du sujet après 2 coups d'Hercule et réplication en surface *)
let example_shallow = bi (bi (trisame very_small_hydra) head) head
(* Exemple du sujet après 2 coups d'Hercule et réplication en profondeur *)
let example_deep =
  let one = bisame very_small_hydra in
  let two = tri one one head in
  tri two head two
(* L'hydre qu'on aurait obtenue si la duplication en profondeur avait fait 2 copies *)
let example_deep_two_copies =
  let one = trisame very_small_hydra in
  let two = Node[one; one; one; head] in
  Node [two; two; two; head]

(* Les hydres pouvant être assez grosses, il est utile de fournir quelques mesures  *)

(* Écrire une fonction donnant la taille d'une hydre (nombre total de noeuds) *)
let rec size : hydra -> int = fun h ->
  match h with
    Node [] -> 1
  | Node l -> 1 + List.fold_right (+)  (List.map size l) 0

(* Écrire une fonction donnant la hauteur d'une hydre (longueur maximale d'un  chemin partant du pied) *)
let rec height : hydra -> int = fun h ->
  match h with
    Node [] -> 0
  | Node (h'::l) -> 1 + List.fold_right (max)  (List.map height l) (height h')

let rec height' : hydra -> int = fun h ->
  match h with
    Node [] -> 0
  | Node(h'::l) ->
    let h1 = 1 + height' h' in
    let h2 = height' (Node l) in
    max h1 h2

let rec arite_max: hydra -> int = fun h ->
  match h with
    Node [] -> 0
  | Node l ->
    max (List.fold_right (max) (List.map arite_max l) 0) (List.length l)

let rec decroissant: hydra -> bool = fun h ->
  match h with
    Node [] -> true
  | Node l -> List.for_all (fun x -> List.length l >= x) (List.map (fun (Node t) -> List.length t) l)
              &&
              List.fold_right (&&) (List.map decroissant l) true


(* Écrire une fonction qui calcule l'histogramme d'une hydre, nombre de noeuds à chaque niveau *)
let fold_app l = List.fold_right ( @ ) l []

let histogram : hydra -> int list = fun h ->
  let rec aux hs =
    match hs with
      [] -> []
    | _  -> List.length hs :: aux (fold_app (List.map (fun h -> match h with Node lh -> lh) hs))
  in aux [h]

(* Écrire une fonction qui compte le nombre de têtes à chaque niveau. *)
let histogram_heads : hydra -> int list = fun h ->
  let rec aux hs =
    match hs with
      [] -> []
    | _  -> List.length (List.filter is_head hs) ::
            aux (fold_app (List.map (fun h -> match h with Node lh -> lh) hs))
  in aux [h]

(*
   Écrire une fonction qui retourne une liste d'arêtes de l'hydre, avec les contraintes
   décrites dans le sujet.
*)
let hydra_edges : hydra -> (int * int) list = fun h ->
  (* Construit la liste [i; i+1; ... ; i+n-1] *)
  let iota i n =
    let rec aux n acc =
      if n <= 0 then acc
      else aux (n-1) ((i+n-1)::acc)
    in aux n []
  in
  let rec aux l next acc =
    (* l: liste de couples (hydre,label du sommet de l'hydre)
       next: premier entier non encore utilisé comme étiquette
       acc: accumule la liste des arêtes déjà traitées.
    *)
    match l with
      [] -> acc
    | ((h,i)::reste) -> (
        match h with
          Node [] -> aux reste next acc
        | Node daughters ->
          let n = List.length daughters in
          (* Les filles sont numérotés next, next+1,...,next+n-1 *)
          let new_edges = List.map (fun x -> (i,x))(iota next n) in
          aux ((List.combine daughters (iota next n)) @ reste) (next+n) (new_edges @ acc)
      )
  in
  (* Tri de la liste pour dot *)
  List.sort
    (fun a b -> if a < b then -1 else if a = b then 0 else 1)
    (aux [(h,0)] 1 [])

let iota i n =
  let rec aux n acc =
    if n <= 0 then acc
    else aux (n-1) ((i+n-1)::acc)
  in aux n []

let rec aux l next acc =
  (* l: liste de couples (hydre,label du sommet de l'hydre)
     next: premier entier non encore utilisé comme étiquette
     acc: accumule la liste des arêtes déjà traitées.
  *)
  match l with
    [] -> acc
  | ((h,i)::reste) -> (
      match h with
        Node [] -> aux reste next acc
      | Node daughters ->
        let n = List.length daughters in
        (* Les filles sont numérotés next, next+1,...,next+n-1 *)
        let new_edges = List.map (fun x -> (i,x))(iota next n) in
        aux ((List.combine daughters (iota next n)) @ reste) (next+n) (new_edges @ acc)
    )


(*
   Affiche une hydre h.
   Prérequis : la fonction hydra_edges doit avoir été écrite.
*)
let show_hydra h =
  (* Translates the list of edges in dot format, and outputs it to filename *)
  let hydra_to_dot h filename =    
    let rec edges_to_dot edges channel =
      match edges with
        [] -> ()
      | (a,b)::r -> Printf.fprintf channel "%d -- %d\n" a b; edges_to_dot r channel
    in
    let dot_preamble = "graph hydra {\n" ^
                       "\trankdir=BT;\n" ^
                       "\tnode [label=\"\" shape=point style=filled fixedsize=true];\n"
    in
    let dot_postamble = "\n}" in
    let edges = hydra_edges h in
    let channel = open_out filename in
    Printf.fprintf channel "%s\n" dot_preamble;
    edges_to_dot edges channel;
    Printf.fprintf channel "%s\n" dot_postamble;
    close_out channel
  in
  (* Get uname of the system to properly set the png viewer *)
  let uname() =
    let (inchannel, outchannel) = Unix.open_process "uname" in
    let name = input_line inchannel in
    close_in inchannel;
    close_out outchannel;
    name
  in
  (* Set viewer to Imagemagick "display" under Linux, or "open" under OSX, otherwise fail :)  *)
  let viewer = let uname = uname() in
    if uname = "Linux" then " display "
    else if uname = "Darwin" then " open "
    else failwith "Viewer not set under windows" in 
  (* Set style to view hydra's heads *)
  let style = "{style=\"invisible\",$.shape=\"none\",height=0.2,width=0.2,image=\"head.png\",label=\"\"}" in
  (* Prepare command *)
  let command = "gvpr -c 'N[$.outdegree==0] " ^ style ^ "' tmp.dot" (* post-process dot file to set style and view hydra's heads *)
                ^ "|" ^ "dot -T png -o tmp.png "                    (* Launch dot on resulting file *)
                ^ "&&" ^ viewer ^ " tmp.png" ^ "&"                  (* Launch viewer in bg *)
  in
  let _ = hydra_to_dot h "tmp.dot" in
  Unix.system command

(*
   Pour désigner un noeud ou une tête, on utilise une notation dite "de Dewey" : le chemin d'accés à un noeud
   est une liste d'indices qui représente le chemin à suivre depuis la racine ("le pied", si on préfère).
   un 0 signifie "aller vers la fille la plus à gauche", etc.
*)

type path = int list

(* Écrire la fonction nth i l qui sélectionne l'élément i d'une liste (la numérotation des éléments commence à 0) *)

let rec nth i l =
  match i,l with
    0, a::_ -> a
  | i, _::l' when i> 0 -> nth (i-1) l'
  | _,_ -> failwith the_msg

(*
   Réactions de l'Hydre.
   Quand la tête de l'Hydre donnée par le chemin p est supprimée, l'Hydre
   effectue son algorithme de réplication.
*)

let rec repeat_concat n a l =
  if n <= 0
  then l
  else repeat_concat (n-1) a (a::l)

(* Supprime le i-ème élément de hs (si c'est une tête) *)

let rec remove_head i hs =
  match i,hs with
  | 0,(Node []) ::hs' -> hs'
  | i, h::hs' when i> 0 -> h :: remove_head (i-1) hs'
  |  _,_  -> failwith the_msg

(* Un tour de base : 
   - Hercule coupe une tête de l'Hydre h donnée par le chemin p.
   - L'Hydre se réplique n fois.
*)

type replication_fun = path -> hydra -> int -> hydra

(* Version en profondeur *)
let rec deep_replication : replication_fun = fun  p h n ->
  match p,h with
    [i], Node l -> Node (remove_head i l)
  | (_::_), Node l -> Node (deep_replication_list p l n)
  | _,_ -> failwith the_msg
and deep_replication_list p l n =
  match p,l with
    0::p', h::lh -> repeat_concat (1+n) (deep_replication p' h n) lh
  | i::p', h::lh when i> 0 -> h :: deep_replication_list (i-1::p') lh n
  | _,_ -> failwith the_msg

(* Version en surface *)
let rec shallow_replication : replication_fun = fun p h  n ->
  match p,h with
    [i], Node l -> Node (remove_head i l)
  | (_::_), Node l -> Node (shallow_replication_list p l n)
  | _,_ -> failwith the_msg
and shallow_replication_list p l n =
  match p,l with
    [0;i], Node l :: lh -> repeat_concat (1+n) (Node (remove_head  i l)) lh
  | 0::p',  h::lh -> shallow_replication p' h n :: lh
  | i::p', h::lh when i> 0 -> h :: shallow_replication_list (i-1::p') lh n
  | _,_ -> failwith the_msg

(* Les stratégies: Hercule et l'Hydre suivent chacun une stratégie *)

(*
   Une stratégie d'Hercule est, à partir d'une Hydre, de choisir une tête.
   Le programmeur qui définit une stratégie doit s'assurer qu'elle retourne
   toujours un chemin vers une tête.
*)

type hercules_strat =  hydra -> path

(*
   Suggestion: avant la fonction check_hercules_strategy, écrire une fonction sub_hydra
   telle que sub_hydra path h renvoie la sous-hydre de h donnée par le chemin path.
*)

let rec sub_hydra : path -> hydra -> hydra = fun path h ->
  match path, h with
    [],_ -> h
  | i::p',Node l -> sub_hydra p'  (nth i l)

(* Écrire la fonction suivante qui teste si une stratégie choisit bien une tête  *)
let check_hercules_strategy : hercules_strat -> hydra -> bool = fun strat  h  ->
  sub_hydra (strat h)  h = head

(* Écrire la stratégie choisissant la tête la plus à gauche *)
let leftmost_head_strat : hercules_strat = fun  h  ->
  let rec aux h =
    match h with
      Node [] -> []
    | Node (h::_) -> 0 :: aux h
  in aux h

(* Écrire la stratégie choisissant une tête de hauteur maximale *)
let rec optimize better l x =
  match l with
    [] -> x
  | y::l' -> if better y x
    then optimize better l' y
    else optimize better l' x

let highest_head_strat : hercules_strat = fun h ->
  let rec helper l i =
    match l with
      [] -> []
    | (n,p)::l' -> (n+1, i :: p):: helper l' (i+1)
  in
  let rec find_largest_path h =
    match h with
      Node [] -> 0, []
    | Node hs -> let l = List.map find_largest_path hs
      in let l' = helper l 0
      in match l' with
        p::ps -> optimize
                   (fun p1 p2 -> fst p1 > fst p2)
                   ps p
      | _ -> failwith "shouldn't happen"
  in snd (find_largest_path h)

(* Écrire une stratégie visant à choisir une tête le plus près du sol possible *)
let closest_to_ground_strat : hercules_strat = fun h  ->
  let rec helper i p hs =
    match hs with [] -> []
                | h::hs' -> (i::p,h):: helper (i+1) p hs'
  in let rec search config =
       match config with
         [] -> failwith the_msg
       | (p,Node [])::_ -> List.rev p
       | (p, Node ls):: c' -> search (c' @ helper 0 p ls)
  in search [[], h]

(* En apprenant à utiliser la bibliothèque Random, écrire une stratégie pour choisir une tête au hasard *)
let random_strat : hercules_strat = fun h ->
  failwith "A écrire"

(* Étant donnée une date, l'Hydre peut calculer un nombre de réplications >= 1 *)

type time = Time of int

type hydra_strat =  time -> int

let check_hydra_strategy : hydra_strat -> time -> bool = fun st t -> st t >= 1

(* Une stratégie classique (celle de la vidéo): à chaque tour, le nombre de réplications est incrémenté. *)

let original_hydra_strat : hydra_strat = function Time  t -> t + 1

(* Une stratégie plus amusante : attention à l'explosion de pile ! *)

let boum : hydra_strat = function (Time t) ->
  let rec exp2 i =
    if i = 0 then 1 else 2 * exp2 (i-1)
  in exp2 t

(* Genre de bataille *)

type genre_de_bataille = Battle_kind of replication_fun * hercules_strat * hydra_strat

(*  Le score final d'une bataille *)
type result =
    Hercules_wins of time       (* Nombre de tours effectués *)
  | Hercules_gives_up of hydra  (* Hydre restante *)

let r2h (r: result) =
  match r with Hercules_wins t -> head
             | Hercules_gives_up h -> h

(* Écrire la fonction de simulation *)
let simulation : genre_de_bataille -> hydra -> time -> result = fun (Battle_kind(replication,hercules_strat, hydra_strat)) initial_hydra (Time(duration)) ->
  let rec aux h time =
    if h = head (* hercules wins *)
    then  Hercules_wins (Time time)
    else if time >= duration
    then Hercules_gives_up h (* The hydra's still alive ... *)
    else let p = hercules_strat h
      in aux (replication p h (hydra_strat (Time time))) (time+1)
  in aux initial_hydra 0

(*
   Écrire une fonction make_trace telle que make_trace measure bat h_init (Time t) donne la suite 
   des valeurs de la fonction measure sur les hydres obtenues en partant de l'hydre h_init et 
   en effectuant t tours de la bataille de genre bat.
*)

let make_trace : (hydra -> 'a) -> genre_de_bataille -> hydra -> time -> 'a list
  = fun measure (Battle_kind(replication,hercules_strat, hydra_strat)) initial_hydra (Time duration) ->
  let rec aux h time acc =
    if h = head  || time >= duration
    then List.rev acc
    else let p = hercules_strat h
      in let h' = replication  p h (hydra_strat (Time time))
      in  aux  h' (time+1) (measure h'::acc)
  in aux initial_hydra 0   [measure initial_hydra]


(* Exercice :
   Définir une fonction construisant une hydre aléatoire pour une hauteur maximum et un branchement
   maximum donnés

   Utiliser cette fonction pour comparer deux couples de stratégies sur un échantillon statistique approprié

*)

(** Ames sensibles s'abstenir *)

(*  On veut définir une fonction de comparaison sur les hydres telle
    que, à chaque tour d'une bataille d'hydre, la nouvelle hydre soit "plus petite" que l'ancienne.

    Malheureusement, cette comparaison ne peut pas être simple (par exemple
    basée sur des nombres entiers, voire même des listes d'entiers, des listes de
    listes d'entiers, etc.)

    Voici une solutions ***très récursive*** à ce problème: *)

(* On définit d'abord un type pour les comparaisons *)


type comparison =
  | Lt (* plus petit *)
  | Eq (* equivalent *)
  | Gt (* plus grand *)


(* On sait alors trier des listes pour le pré-ordre associé à une
   fonction de comparaison *)

(* insertion dans une liste triee par ordre decroissant *)


let rec insert_decr  comp x l =
  match l with
    [] -> [x]
  | y::l' -> match  comp x  y with
    |Gt | Eq -> x :: l
    |Lt -> y :: insert_decr comp x l'

(*  tri par ordre decroissant *)

let rec  insertion_sort comp l =
  match l with
    [] -> []
  | x::l' -> insert_decr comp x (insertion_sort comp l')

(* exemples *)
let int_comp = fun x y -> if x < y then Lt else if x = y then Eq else Gt
let _ = insertion_sort  int_comp  [6;8;4;3;7;2]

(* comparaison de listes triées en ordre décroissant *)

let rec lexico comp l l' =
  match l, l' with
  | [], [] -> Eq
  | _ ,[] -> Gt
  | [], _ -> Lt
  | x::l1, y::l'1 -> (match comp x y with
      Lt -> Lt
    | Gt -> Gt
    | Eq ->  lexico comp l1 l'1)


let _ = lexico int_comp  [5;5;3] [5;5;2]
let _ = lexico int_comp [6;6;5;3] [6;5;5;5;5;5;5;5;5;4;4;4;4;2]
let _ = lexico int_comp [] [5;6;2]

(* Maintenant, on a tous les outils pour comparer deux hydres.
   h est "plus petite" que h' :
   signifie que h est plus proche de la défaite que h' :
   rien à voir avec la taille !
   Les matheux diraient que l'ordinal associé à h est strictement inférieur
   à l'ordinal associé à h'

   Noter que si la comparaison de deux hydres renvoie la valeur Eq, cela ne
   signifie pas que les hydres soient identiques. Elles sont seulement
   "équivalentes" (à des permutations de filles près)

*)

let rec hydra_comp h h' =
  match h, h' with
    Node l, Node l' ->
    lexico hydra_comp
      (insertion_sort hydra_comp l)
      (insertion_sort hydra_comp l')

let smaller h h' = hydra_comp h h' = Lt

let smaller_test bat_kind h n = smaller (r2h (simulation bat_kind h n)) h

(* Démos et tests *)
(* statistiques sur une hydre *)

let _ = histogram my_hydra
let _ = size my_hydra
let _ = height my_hydra

(* fonction sub_hydra *)
let _ = sub_hydra [2;0;0] my_hydra

(* Pour comprendre les deux algorithmes de réplication *)

(* A small hydra story *)

let h0 = tri head (single (single (tri head head head))) head
let h1 = shallow_replication [0] h0 0
let h2 = shallow_replication [1] h1 0
let h3 = shallow_replication [0;0;0;2] h2 1
let h4 = shallow_replication [0;0;1;1] h3 2
let h5 = shallow_replication [0;0;3;0] h4 3
let h6 = shallow_replication [0;0;0;1] h5 6
let h7 = shallow_replication [0;0;12] h6 7
let _ = histogram h7

let h'3 = deep_replication [0;0;0;2] h2 1
let h'4 = deep_replication [0;0;1;1] h'3 2
let h'5 = deep_replication [0;0;3;0] h'4 3
let h'6 = deep_replication [0;0;0;1] h'5 6
let h'7 = deep_replication [0;0;12] h'6 7
let _ = histogram h'7

(* tests de stratégie (pour Hercule) *)

let _ = closest_to_ground_strat (single (single (bi (single head) head)))
let _ = highest_head_strat (bi head  (single (bi (single head) head)))
let _ = closest_to_ground_strat yet_another_hydra
let _ = closest_to_ground_strat my_hydra

let _ = check_hercules_strategy highest_head_strat (bi head  (single (bi (single head) head)))
let _ = check_hercules_strategy leftmost_head_strat my_hydra
let _ = check_hercules_strategy highest_head_strat my_hydra
let _ = check_hercules_strategy closest_to_ground_strat my_hydra
let _ = check_hercules_strategy closest_to_ground_strat yet_another_hydra



(* simulations (diverses hydres et stratégies *)
(* inventez vos propres exemples ! *)



let bat1 = Battle_kind(deep_replication, leftmost_head_strat, original_hydra_strat)
let bat2 = Battle_kind(shallow_replication, leftmost_head_strat, original_hydra_strat)
let bat3 = Battle_kind(shallow_replication, highest_head_strat, original_hydra_strat)
let bat5 = Battle_kind(shallow_replication, closest_to_ground_strat, original_hydra_strat)

(* beware of stack overflows ! *)

let bat4 = Battle_kind(shallow_replication, highest_head_strat, boum)


let sc1  = simulation bat1  my_hydra (Time 25)
let _ = size (r2h sc1)

let _ = size (r2h (simulation bat2  my_hydra (Time 3)))
let _ = size (r2h (simulation bat2  my_hydra (Time 25)))

let _ = smaller_test bat2 my_hydra (Time 25)
let _ = smaller_test bat5 goodstein_hydra (Time 125)

let _ = histogram (r2h (simulation bat5 another_hydra (Time 12000)))


let histo_g4 t = histogram (r2h (simulation bat5 goodstein_hydra  (Time t)))
let histo'_g4 t = histogram_heads  (r2h (simulation bat5 goodstein_hydra  (Time t)))
let _ = histo_g4 5
(* [1; 13; 10] *)
let _ = histo'_g4 4
let _ = histo'_g4 11
let _ = histo'_g4 25
let _ = histo'_g4 53
let _ = histo'_g4 109
let _ = histo'_g4 110
let _ = histo'_g4 221
let _ = histo'_g4 447
let _ = histo'_g4 448
let _ = histo'_g4 (448  + 449)
let _ = histo'_g4 (448  + 449 + 1)
let _ = histo'_g4 5000

let test_higher t = simulation bat3 goodstein_hydra  (Time t)


let _ = test_higher  5000

let stats =  make_trace size  bat5 goodstein_hydra
let histos =  make_trace histogram  bat5 goodstein_hydra



let sizes = stats  (Time 800)
let tr  = histos (Time 800)
let _ = List.nth tr 777
let _ = List.nth sizes 777


(* more experiments *)


(* NE PAS DIFFUSER *)
let  decompose  (Node l) =
  let rec decompose_help x acc l' =
    match l' with
      [] -> [(x,acc)]
    | y :: l' -> if x = y then decompose_help x (acc+1) l'
      else (x,acc)::decompose_help y 1 l'
  in match l with
    [] -> []
  | x::l' -> decompose_help x 1 l'

let test t =decompose (r2h (simulation bat5 goodstein_hydra t))

let _ = test (Time 3);;

