(* Pierre Casteran, LaBRI *)


(* Ordinals of the first and second class *)

Set Implicit Arguments.

Require Import Relations  Classical  Well_Orders  Lub.
Require Import Denumerable  Compare_dec  Coq.Sets.Image  PartialFun.
Require Import Classical_sets  PartialFix.
Import MoreIota.

Hint Unfold In.
Arguments Included [U] _ _.
Arguments denumerable [U] _.
Arguments Same_set [U] _ _.

Delimit Scope ord_scope with ord.
Open Scope ord_scope.


(* The following declaration is AX1 axiom from Schutte   ? *)

Parameter OT : Type.
Parameter ordinal: Ensemble OT.
Parameter lt : relation OT.
Notation ON := ordinal (only parsing).
Axiom AX1 : WO lt ordinal.

Axiom inh_OT : inhabited OT.
Instance InH_OT : InH OT.
Proof. split; exact inh_OT. Qed.

Hint Resolve inh_OT : schutte.

Definition le := Le lt ordinal.
Notation "a < b" := (lt a b): ord_scope.
Notation "a <= b" := (le a b):ord_scope.

Ltac ordi := match goal with
                 [ |- ordinal _  ] => eauto with schutte
               |               _     =>  idtac 
             end.

Definition Unbounded (X : Ensemble OT) :=
  forall x, ordinal  x -> exists y, In X y /\ x < y.

Axiom AX2 : forall X: Ensemble OT, Included X ordinal ->
                                   (exists a, ordinal a /\
                                              (forall y, X y -> y < a)) ->
                                   denumerable X.


Axiom AX3 : forall X : Ensemble OT,
              denumerable X -> (Included X ordinal) ->
              exists a, ordinal a /\ (forall y, X y -> y < a).


(** Definitions  *)


Definition the_least  (X: Ensemble OT)  : OT :=
  the (is_least_member  (ordinal) lt X ).

Definition zero := the_least ordinal.

Definition succ (alpha : OT) := the_least  (fun beta => alpha < beta).

Reserved Notation "'F' n" (at level 29) .

Fixpoint finite (i:nat) : OT :=
  match i with 0 => zero
            | S i=> succ (F i)
  end
where "'F' i" := (finite i)  : ord_scope.

Definition is_finite  := seq_range finite.

Definition sup_spec X l := is_lub  (ordinal) lt X l.

Definition sup (X: Ensemble OT) : OT  := the  (sup_spec X).

Notation "'|_|' X" := (sup X) (at level 29) : ord_scope.

Definition omega_limit (f:nat->OT) : OT 
  := |_| (seq_range f).

Definition omega := omega_limit finite.

Definition is_succ (a:OT) := exists b, ordinal b /\ a = succ b.

Definition is_limit (a:OT) := ordinal a /\ a <> zero /\ ~ is_succ a.

Definition members (a:OT) := Intersection _ ordinal (fun b => b < a).

Definition set_eq (X Y: OT -> Prop) := forall a, ordinal a -> (X a <-> Y a).


Definition progressive (P : OT -> Prop) : Prop :=
  forall a, ordinal a -> (forall b, ordinal b -> b < a -> P b) -> P a.


Definition closed (B : Ensemble OT) : Prop := 
  Included B ordinal /\
  (forall M, Included M B -> Inhabited _ M -> denumerable M ->
                            In B (|_| M )).


Definition  continuous (f:OT->OT)(A B : Ensemble OT) : Prop :=
  fun_codomain A B f /\
  closed A /\
  (forall U, Included U A -> Inhabited _ U -> 
             denumerable U -> |_| (image U f) = f (|_| U)).

(** Proofs, proofs, proofs *)


Lemma Unbounded_not_denumerable : forall X,
                                    Included X ordinal ->
                                    Unbounded X -> not (denumerable X).
Proof.
  red;intros X H H0 H1; case (AX3 H1 H); intros x (Hx,H'x).
  case (H0 _ Hx); intros y (Hy,H'y);
  case (Lt_irreflexive x); 
  apply (Lt_trans  x y x); auto.
Qed.

Lemma denumerable_not_Unbounded : forall X, Included X ordinal ->
                                            denumerable  X -> not (Unbounded X).
Proof.
  red;intros; apply Unbounded_not_denumerable with X;auto.
Qed.


Lemma Progressive_Acc: progressive (Acc lt).
Proof.
  split; intros; apply H0; auto; eapply Lt_inc_D1;eauto.
Qed.


Theorem Induction : forall (P: OT->Prop), progressive P ->
                                          forall a, ordinal a -> P a.
Proof.
  intros P Hp; case (classic  (forall a : OT, ordinal a -> P a));
  [trivial|idtac].
  intro H; case (not_all_ex_not _ (fun a => ordinal a -> P a) H).
  intros x Hx; case (imply_to_and _ _ Hx); intros Ox P'x.
  elimtype False.
  {  case (well_order  (fun z => ~ P z) x Ox P'x).
     intros x0 H0; assert (forall b, ordinal b -> b < x0 -> P b).
     {  intros.
        case H0; intros H3 (H4,H5); apply NNPP.
        intro H6; case (H5 _ H1 H6).
        intro;subst x0.
        case (Lt_irreflexive b);auto.
        intro; case (Lt_irreflexive  b).
        eapply Lt_trans ;eauto.
     }
     assert (~ (P x0)) by (case H0; unfold In; tauto).
     apply H2; apply Hp;auto.
     case H0;tauto.
  }
Qed.


Lemma le_refl : forall alpha, ordinal alpha -> alpha <= alpha.
Proof.  left;auto. Qed.

Hint Resolve le_refl : schutte.

Lemma eq_le : forall a b, ordinal a -> a = b -> a <= b.
Proof. left;auto. Qed.

Lemma lt_le : forall a b, a < b -> a <= b.
Proof. right;auto. Qed.


Lemma lt_trans : forall a b c, a < b ->  b < c -> a < c.
Proof Lt_trans.

Lemma le_lt_trans : forall a b c, a <= b -> b < c -> a < c.
Proof. eapply Le_Lt_trans;  exact AX1. Qed.

Lemma lt_le_trans : forall a b c, a < b -> b <= c -> a < c.
Proof (Lt_Le_trans AX1).

Lemma le_trans : forall a b c, a <= b -> b <= c -> a <= c.
Proof (Le_trans AX1).

Lemma lt_irr : forall a, ~ (a < a).
Proof Lt_irreflexive.

Lemma le_antisym : forall a b, a <= b -> b <= a -> a = b.
Proof (Le_antisym AX1).

Lemma le_eq_or_lt : forall a b, a <= b -> a = b \/ a < b.
Proof.  intros a b [(_,H)|H];auto. Qed.


Lemma le_not_gt : forall a b, a <= b -> ~ b < a.
Proof.
  intros a b H; case (le_eq_or_lt H).
  -  intro;subst b;apply lt_irr.
  -  intros H0 H1;case (@lt_irr a);eapply lt_trans;eauto.
Qed.

Hint Resolve eq_le lt_le lt_trans le_trans le_lt_trans 
     lt_le_trans lt_irr le_not_gt:
  schutte.


Lemma lt_ordinal : forall a b, a < b -> ordinal a.
Proof.
  intros; generalize (Lt_inc_D1  a);eauto.
  intro; eapply H0; eauto.
Qed.

Lemma gt_ordinal :  forall a b, a < b -> ordinal b.
Proof.
  intros; generalize  (Lt_inc_D2  a); unfold In; auto.
Qed.


Lemma le_ordinal : forall a b, a <= b -> ordinal a.
Proof.
  destruct 1.
  - tauto.
  -  eapply lt_ordinal; eauto.  
Qed.


Lemma ge_ordinal :  forall a b, a <= b -> ordinal b.
Proof.
  destruct 1.
  -  case H;intros;subst b;auto.
  -  eapply gt_ordinal; eauto.  
Qed.


Lemma le_disj : forall alpha beta, alpha <= beta -> alpha = beta \/ alpha < beta.
Proof.
  intros alpha beta H; destruct H; tauto.
Qed.


Hint Resolve lt_ordinal gt_ordinal le_ordinal ge_ordinal : schutte.


Lemma all_ord_acc : forall alpha, ordinal alpha -> Acc lt alpha.
Proof.
  intros alpha Halpha;  pattern alpha;  apply Induction; auto.
  apply Progressive_Acc; auto.  
Qed.


Lemma trichotomy : forall a b, ordinal a -> ordinal b ->
                               a < b \/ a = b \/ b < a.
Proof (Lt_connect AX1).

Ltac tricho t u Hname := case (@trichotomy t u);
    [auto with schutte | 
     auto with schutte | 
     intro Hname |
     intros [Hname|Hname]].

Lemma lt_or_ge : forall a b, ordinal a -> ordinal b ->  a < b \/ b <= a.
Proof.
  intros a b Ha Hb; tricho a b H;auto with schutte.
Qed.


Lemma not_gt_le : forall a b, ordinal a -> ordinal b -> ~ b < a  -> a <= b.
Proof.
  intros a b H H0 H1; tricho a b H2; auto with schutte;  contradiction.  
Qed.

Hint Unfold Included.

Theorem Non_denum : ~ denumerable ordinal.
Proof.
  red; intro H; case (AX3  H); auto.
  intros x (H1,H2); case (@lt_irr x);auto.
Qed.

Lemma  the_least_unicity (X: Ensemble OT)
        (HX: Inhabited _ X /\ Included X ordinal) 
: exists! l , is_least_member  (ordinal) lt X l.
Proof.
  destruct HX as [H H0]; destruct H as [x Hx].
  destruct (well_order   X x (H0 x Hx) Hx);intros;auto.
  exists x0; split; auto.
  intros; eapply least_member_unicity;eauto.
  apply AX1.
Qed.

Lemma Inh_ord : Inhabited _ ordinal.
Proof. 
  destruct (@AX3 (Empty_set  OT)).
    apply denumerable_empty.
    intros x H; destruct H.
    exists x;tauto.
Qed.

Theorem ordinal_zero : ordinal zero.
Proof.
  unfold zero, the_least, the; apply iota_ind.
  -  apply the_least_unicity.
     split; auto.  apply Inh_ord.
  - now destruct 1 as [[H _] _].
Qed.

Hint Resolve ordinal_zero : schutte.


Lemma le_zero : forall a, ordinal a -> zero <= a.
Proof.
  unfold zero, the_least, the; apply iota_ind.
  -  apply the_least_unicity.
     split; auto;  apply Inh_ord.
  -  destruct 1 as [[H [H0 H1]] _].
     intros a0 H2; destruct (H1 _ H2);auto.
     + now left.  
     + now right.
Qed.

Lemma not_lt_zero_0 : forall a, ordinal a -> ~ a < zero.
Proof.
  intros;apply le_not_gt; apply le_zero;auto.
Qed.

Lemma not_lt_zero : forall a,  ~ a < zero.
Proof.
  intros alpha H;  case (not_lt_zero_0  (lt_ordinal  H)  H).
Qed.

Lemma zero_or_greater : forall a, ordinal a -> 
                                  a = zero \/ exists b, ordinal b /\ b < a.
Proof.
  intros a Ha; tricho zero a t;auto with schutte.
  -  right;exists zero;auto with schutte.
  -  case (not_lt_zero   t).
Qed.

Lemma zero_or_positive : forall a, ordinal a -> 
                                   a = zero \/ zero < a.
Proof.
  intros a Ha; tricho zero a t;auto with schutte;  case (not_lt_zero   t).
Qed.


(* the following proof should be simplified with some lemmas on
  denumerability *)

Theorem unbounded : forall alpha, ordinal alpha ->
                                  exists beta, ordinal beta /\
                                               alpha < beta. 
Proof.
  intros alpha H; pose (X0 := fun z =>  z < alpha /\ ordinal z).
  assert (denumerable X0). 
  { apply AX2.
    unfold X0, Included, In; intuition.
    exists alpha; split;auto.
    unfold X0;intuition.
  }
  pose (X:= Add  _ X0 alpha).
  assert (denumerable X). 
  {  unfold X; unfold Add; apply denumerable_union.
     auto.
     apply denumerable_singleton.
  } case (AX3  H1).
  unfold X; intros x Hx.
  case Hx.
  -     unfold In, X0;tauto.
  - inversion 1.
    subst x0; unfold In;auto.
  - intros beta (Hb,H'b).
    exists beta; split;auto.
    apply H'b;  unfold X;auto.
    right;split. 
Qed.





Lemma the_least_ok : forall X, Inhabited OT X -> Included X ordinal -> 
                               forall a, X a -> the_least X <= a.
Proof.
  intros;   pattern X, (the_least X).
  unfold the_least, the;apply iota_ind.
  -  apply the_least_unicity.
     split;auto.
  - intros a0 H2;  destruct H2.
    destruct H2.
    destruct H4.
    destruct (H5 _ (H0 _ H1));  auto.
    + subst;   now apply le_refl.   
    + right;auto.
Qed. 


(** The successor *)

(** (succ alpha) is defined as the leat ordinal grater than alpha *)

Definition succ_spec (alpha:OT) := is_least_member  ordinal lt (fun z => alpha < z).

Lemma succ_ok : forall alpha, In ordinal alpha ->
                              succ_spec alpha  (succ alpha).
Proof.
  intros; unfold succ, the_least,the;  apply iota_spec.
  destruct (@AX3 (Singleton _ alpha)).
  apply denumerable_singleton.
    now destruct 1.  
     unfold succ_spec; apply the_least_unicity; split.
     exists x; intuition.
     eauto with schutte.
Qed.     


Remark Rsucc : forall a, ordinal a ->
                         exists ! x : OT,
                           is_least_member ordinal lt (fun beta : OT => a < beta) x.
Proof.
  intros.
  eapply the_least_unicity.
  split.  
  +  destruct  (unbounded H).
     exists x;auto.
     red;auto.
     tauto.
  +   red;eauto with schutte.
Qed.


Lemma lt_succ : forall a, ordinal a -> a < succ a.
Proof.
  intros a Ha; destruct  (succ_ok  Ha);  tauto.
Qed.

Hint Resolve lt_succ : schutte.



Lemma ordinal_succ :  forall a , ordinal a ->  ordinal  (succ a).
Proof.
  intros a Ha; destruct  (succ_ok  Ha);  tauto.
Qed.


Hint Resolve ordinal_succ :  schutte.


Lemma lt_succ_le : forall a b, 
                     a < b -> succ a <= b.
Proof with eauto with schutte.
  intros a b H;  pattern (succ a).
  apply the_least_ok ... 
  exists (succ a); red;apply lt_succ ...
Qed.

Lemma lt_succ_le_2 : forall a b,
                       ordinal b ->
                       a < succ b ->
                       a <= b.
Proof.
  intros;  tricho a b t;eauto with schutte.
  case (@lt_irr a).
  apply lt_le_trans with (succ b).
  - assumption.
  - apply lt_succ_le;eauto with schutte.
Qed.


Lemma succ_mono : forall a b, ordinal a -> ordinal b ->
                              a < b -> succ a < succ b.
Proof.
  intros; eapply le_lt_trans. 
  - eapply lt_succ_le;eauto.
  - auto with schutte.
Qed.

Arguments succ_mono [ a b].

Lemma succ_monoR : forall a b, ordinal a -> ordinal b ->
                               succ a < succ b -> a < b.
Proof.
  intros a b Ha Hb H; tricho a b t; auto with schutte.
  -  subst b; case (@lt_irr (succ a)); eauto with schutte.
  - case (@lt_irr  (succ a)); eapply lt_trans;eauto.
    apply succ_mono;auto.
Qed.


Lemma succ_injection : forall a b, ordinal a -> ordinal b ->
                                   succ a = succ b -> a = b.
Proof.
  intros a b Ha Hb H; tricho a b t;auto.
  - generalize (succ_mono Ha Hb t).
    rewrite H; intro; case (@lt_irr (succ b));auto.
  -generalize (succ_mono Hb Ha  t).
   rewrite H; intro; case (@lt_irr (succ b));auto.
Qed. 

Lemma succ_zero_diff : forall a , ordinal a -> succ a <> zero.
Proof.
  intros a Ha H0; destruct (@not_lt_zero a).
  rewrite <- H0; auto with schutte.
Qed.

Lemma zero_lt_succ : forall alpha, ordinal alpha -> zero < succ alpha.
Proof.
  intros a Ha; apply le_lt_trans with a;eauto with schutte;  apply le_zero;auto.
Qed.

Hint Resolve zero_lt_succ le_zero : schutte.




Lemma ordinal_finite : forall i, ordinal (F i).
Proof.
  induction i; simpl; auto with schutte.
Qed.

Hint Resolve ordinal_finite : schutte.


Lemma finite_lt_inv : forall i o, (* ordinal o -> *) o < F i -> 
                                                     exists j:nat , (j<i)%nat /\ o = F j.
Proof.
  induction i.
  - simpl (F 0); intros ; destruct  (not_lt_zero  H ).
  - simpl (F (S i));intros o H; destruct (@lt_succ_le_2 o (F i));auto with schutte.
    +  exists i.
       destruct H0;subst o; auto with arith.
    + destruct (IHi o  H0) as [x [H1 H2]].
      exists x; case H1;auto with arith.
Qed.

Lemma finite_mono : forall i j, (i<j)%nat -> F i < F j.
Proof.
  induction 1.
  -  simpl (finite (S i)); auto with schutte.
  -  apply lt_trans  with (finite m);auto.
     simpl (finite (S m)); apply lt_succ; auto with schutte.
Qed.

Hint Resolve finite_mono : schutte.

Lemma finite_inj : forall i j, F i = F j -> i = j.
Proof.
  intros i j H; case (lt_eq_lt_dec i j). 
  -  destruct 1.
     +  case (@lt_irr (F i)).
        *  pattern (F i) at 2 ; rewrite H. 
           apply finite_mono;auto.
     + auto.
  -  intro;  case (@lt_irr (F i)).
     pattern (F i) at 1 ; rewrite H; apply finite_mono;auto.
Qed.

(* Building limits *)

(* the following proof should be simplified *)

Lemma sup_exists : forall X, denumerable  X ->
                             Included X ordinal  ->
                             ex (sup_spec  X).
Proof.
  unfold Included, In; intros X H H0.
  case (AX3 H); auto.
  intros x (Hx,H'x).
  assert (H''x : forall y : OT, X y -> y=x \/ lt y x) by  intuition.
  generalize  (well_order   
                 (fun z => ordinal z /\ 
                           (forall y,  ordinal y -> X y -> y = z \/ y < z))
                 x Hx ).
  intros H1; case H1;intros; auto.
  exists x0; unfold is_least_member in H2; repeat split.
  -  case H2; auto.
  - case H2; intros; intuition.
    red; intros; case H2;  intros; apply H11;auto.
  -  intros; decompose [and] H2; apply H8; auto.
Qed.

Lemma sup_unicity : forall X l l',
                      (denumerable  X /\ forall x, X x -> ordinal x) ->
                      sup_spec X l -> sup_spec X l' -> l = l'.
Proof.
  intros X l l' _ H H0; apply le_antisym.
  destruct H, H0; intuition.  
  case (H4 l' H0);auto with schutte.
  destruct H, H0; intuition.  
  case (H5 l H);auto with schutte.
Qed.

Lemma sup_spec_unicity (X:OT->Prop)(HX : denumerable X /\ Included X ordinal ) :
  exists! u, sup_spec X u.
Proof.
  destruct HX as [H1 H2].
  destruct (sup_exists  H1 H2).
  exists x;split;auto.
  intros.  apply sup_unicity with   X;split;auto.
  destruct H. auto.
   red in H. red in H. split. destruct H. tauto.
  destruct H.
  destruct H3.
  auto.
  destruct H0. auto.
  red in H0. red in H0. split. destruct H0. tauto.
  destruct H0.
  destruct H3.
 auto.
Qed.


Lemma sup_ok1 (X:OT->Prop)(HX : denumerable X /\ Included X ordinal ) :
  sup_spec X (sup X).
  unfold sup, the;  apply iota_spec. 
  now apply sup_spec_unicity.
Qed.  


Lemma lt_succ_lt : forall a b,  is_limit b -> 
                                a < b -> succ a < b.
Proof.
  intros a b (Ob,(Zb,Sb)) H .
  case (lt_succ_le  H).
  intros (_,H') ; subst b.
  case Sb.
  exists a;eauto with schutte.
  auto.
Qed.

Lemma sup_upper_bound :
  forall (X: OT -> Prop) x,
    denumerable X ->
    (forall x, X x -> ordinal x) ->
    X x -> x <= sup  X.
Proof.
  intros.
  
  generalize (sup_ok1   (conj H H0)).
  intro H2; red in H2.
  red in H2.
  destruct H2.
  destruct H3.
  red in H3.
  destruct (H3 x (H0 x H1)); auto.
  left.  
  split;auto.
  right;auto.
Qed. 

Lemma sup_least_upper_bound : forall (X: OT->Prop) x,
                                denumerable X ->
                                (forall y, X y -> ordinal y) -> ordinal x ->
                                (forall y, X y -> y <= x) -> sup  X <= x.
Proof.
   intros.
    
    generalize (sup_ok1   (conj H H0)).
    intro H3; red in H3.
    red in H3.
    destruct H3.
    destruct H4.
     specialize (H5 x H1 ).
    destruct H5.
 red.
   red in H4.
    intros.
   destruct (H2 x0).
    auto.
   destruct H7;subst;auto.
   now right.
  subst;left.
  split;auto.
   right;auto.
Qed.

Lemma sup_ordinal : forall X, denumerable X -> 
                              (forall x, X x -> ordinal x) ->
                              ordinal (sup X).
Proof.
  intros X HX H. 
   generalize (sup_ok1   (conj HX H)).
  intro H0; red in H0.
  now destruct H0.
Qed.



Hint Resolve sup_ordinal : schutte.


Lemma sup_of_leq : forall a, ordinal a -> a = |_| (fun x => ordinal x /\
                                                            x <= a).
Proof.
  intros.
  assert (DD :denumerable (fun x : OT => ordinal x /\ x <= a)).
  {
    apply AX2.
    unfold Included;auto.
    intuition.
    red in H0.
    tauto.
    exists (succ a).
    split;auto with schutte.
    destruct 1.
    apply le_lt_trans with a;auto.
    now apply lt_succ.
  }
  apply le_antisym.
  apply sup_upper_bound;auto.
  tauto.
  split;auto with schutte.
  apply sup_least_upper_bound; auto.
  tauto.
  tauto.
Qed.



Lemma sup_mono : forall (X Y : Ensemble OT),
                   denumerable X ->
                   denumerable Y ->
                   (Included X ordinal)  ->
                   (Included Y ordinal)  ->
                   (forall x, In X x -> exists y, In Y y /\ x <= y) ->
|_| X <= |_| Y.
Proof.
  intros.
  apply sup_least_upper_bound; auto with schutte.
  intros x Hx; case (H3 x Hx); intros y (Hy,Hy'). 
  apply le_trans  with y;auto with schutte.
  apply sup_upper_bound;auto with schutte.
Qed.

Lemma sup_eq_intro :  forall (X Y : Ensemble OT),
                        denumerable X ->
                        denumerable Y ->
                        (Included X ordinal) ->
                        (Included Y ordinal) ->
                        (Included X Y) ->
                        (Included Y X) ->
|_| X = |_| Y.
Proof.
  unfold Included, In; intros; apply le_antisym; apply sup_mono;auto with schutte;
  intro x;exists x; unfold In;auto with schutte.
Qed.



Lemma lt_sup_exists_leq : forall X, denumerable X ->
                                    (Included X ordinal)  ->
                                    forall y, ordinal y -> y < sup X ->
                                              exists x, In X x /\ y <= x.
Proof.
  intros.
  case (not_all_not_ex _ (fun x =>  X x /\ y <= x)).
  2 : intros x Hx; exists x;auto.
  red;intro.

  unfold Included,In in H0.
  
  assert (forall n, X n -> n < y).
  intro.
  intro.
  apply NNPP.
  red;intro.
  assert (y < n).
  case (@trichotomy n y).
  
  apply H0;auto.
  auto.
  intro; case H5;auto.
  destruct 1.
  subst n.
  case (H3 y).
  auto with schutte.
  auto.
  case (H3 n).
  split;auto with schutte.
  clear H3.
  case ( @lt_irr y).
  eapply lt_le_trans with (|_|X);eauto.
  apply sup_least_upper_bound;auto with schutte.
Qed.




Lemma lt_sup_exists_lt : forall X, denumerable X ->
                                   Included X ordinal ->
                                   forall y, ordinal y -> y < sup X ->
                                             exists x, In X x /\ y < x.
Proof.
  
  intros.
  unfold Included, In in H0.
  case (not_all_not_ex _ (fun x =>  X x /\ y < x)).
  2 : intros x Hx; exists x;auto.
  red;intro.
  assert (forall n, X n -> n <= y).
  intro.
  intro.
  apply NNPP.
  red;intro.
  assert (y < n).
  case (@trichotomy n y).
  apply H0;auto.
  auto.
  intro.
  case H5;auto with schutte.  
  destruct 1.
  case H5;auto with schutte.
  auto.
  case (H3 n).
  split;auto with schutte.
  assert (sup X <= y). 
  apply sup_least_upper_bound.
  auto.
  auto.
  auto.
  auto.
  case (Lt_irreflexive  y).
  eapply Lt_Le_trans;eauto.
  apply AX1.
Qed.




Lemma members_eq : forall a b, ordinal a -> ordinal b ->
                               members a = members b -> a = b.
Proof.
  intros.
  tricho a b H2.
  assert (members b a).
  split;auto.
  assert (~(members a a)).
  red; inversion_clear 1.
  case (@lt_irr a);auto.
  rewrite H1 in H4.
  case H4;auto.
  auto.
  assert (members a b).
  split;auto.
  assert (~(members b b)).
  red;inversion_clear 1.
  case (@lt_irr b);auto.
  rewrite <- H1 in H4.
  case H4;auto.
Qed.


Lemma sup_members_succ : forall a, ordinal a ->
                                   (sup (members a)) < a ->
                                   a = succ (|_| (members a)).
Proof.
  intros a H;pose (A := sup (members a)).
  fold A.
  generalize (@lt_succ_le  A a). 
  intros.
  assert (ordinal A). 
  unfold A;apply sup_ordinal.  
  apply AX2.
  unfold Included, In, members. destruct 1;tauto.
  exists a;unfold members. 
  (* The inductive flavour of Union, Intersection, etc ... prevents us 
     from using tauto, intuition *)
  split;auto;destruct 1.
  auto.
  unfold members.
  destruct 1.
  auto.
  case (le_eq_or_lt  (H0  H1)).
  symmetry;auto.
  intro H3.
  assert (succ A <= A). 
  unfold A;apply sup_upper_bound. 
  apply AX2.
  unfold Included. destruct 1; auto. 
  exists a;unfold members. 
  split; auto. 
  destruct 1;auto.
  unfold members;destruct 1;auto.
  unfold members at 1.
  split.
  fold A.
  unfold In; apply ordinal_succ.
  auto.
  unfold In. 
  auto.
  case (@lt_irr A). 
  eapply lt_le_trans. 
  eapply lt_succ;eauto.
  auto.
Qed.




Lemma sup_members_not_succ : forall a b, ordinal a -> ordinal b ->
                                         a = sup (members a) -> a <> succ b.
Proof.
  intros a b Ha Hb e e1.
  generalize (sup_of_leq  Hb).
  intro.
  assert (sup (members a) = b).
  unfold members;rewrite H.
  apply sup_eq_intro.
  apply AX2.
  auto.
  apply Intersection_decreases_l.
  exists a. split;auto.
  destruct 1.
  auto.
  apply AX2.
  unfold Included. destruct 1.
  auto.
  exists (succ b).
  split;auto with schutte.
  intros y (Hy,H'y).
  eauto with schutte.
  apply Intersection_decreases_l.
  unfold Included. destruct 1.
  auto.
  unfold Included. destruct 1.
  split; auto.
  unfold In in H1.
  rewrite e1 in H1.
  apply lt_succ_le_2;auto.
  unfold Included, In.
  intros x (H1,H2).
  split;auto with schutte.
  unfold In.
  
  
  rewrite e1.
  eapply le_lt_trans;eauto with schutte.
  case (@lt_irr a).
  pattern a at 2;rewrite e1.
  rewrite e;rewrite H0.
  auto with schutte.
Qed.





Definition seq_mono (f:nat -> OT) :=
  forall i j, (i<j)%nat -> f i < f j.

Lemma seq_mono_intro : forall f,
                         (forall i, f i <  f (S i)) -> seq_mono f.
Proof.
  intros f H; unfold seq_mono; induction 1.
  auto.
  apply (Lt_trans ) with (f m);auto.
Qed.


Lemma seq_mono_inj : forall f,
                       seq_mono f -> injective _ _ f. 
Proof.
  unfold injective;intros f H i j H0.
  case (lt_eq_lt_dec i j).
  destruct 1.
  generalize (H _ _ l).
  rewrite H0;intro;case (Lt_irreflexive  ( f j));auto.
  auto.
  intro l;generalize (H _ _ l).
  rewrite H0;intro;case (Lt_irreflexive  ( f j));auto.
Qed.





Lemma seq_range_denumerable : forall f,
                                seq_mono f -> denumerable (seq_range f).
Proof.
  intros f _.
  apply Denumerable.seq_range_denumerable.
Qed.

Hint Resolve seq_range_denumerable seq_mono_intro : schutte.


Lemma lt_omega_limit : forall f , (forall i, ordinal (f i)) ->
                                  seq_mono f ->
                                  forall i, f i <  omega_limit f.
Proof.
  unfold omega_limit;intros.
  apply lt_le_trans with (f (S i)).
  eapply H0.
  auto.
  apply sup_upper_bound;auto with schutte.
  destruct 1.
  subst x;auto.
  exists (S i);auto.
Qed.



Lemma omega_limit_least : forall f , (forall i, ordinal (f i)) ->
                                     seq_mono f ->
                                     forall y, ordinal y ->
                                               (forall i, f i < y) ->
                                               omega_limit f <=  y.
Proof.
  intros.
  unfold omega_limit;apply sup_least_upper_bound;auto with schutte.
  destruct 1.
  subst y0;auto.
  destruct 1.
  subst y0;auto with schutte.
Qed.


Lemma finite_lt_omega : forall i, F i < omega.
Proof.
  unfold omega ; intros.
  apply lt_omega_limit; auto with schutte.
Qed.



Lemma lt_omega_finite : forall a,  a < omega ->
                                   exists i:nat, a = F i.
Proof.
  unfold omega;intros a H0.
  assert (H:ordinal a).
  eauto with schutte.
  unfold omega_limit in H0.
  generalize (lt_sup_exists_leq (X:=(seq_range finite))  ).
  intros.
  assert (denumerable (seq_range finite)).
  apply seq_range_denumerable.
  unfold seq_mono;intros;apply finite_mono;auto.
  generalize (H1 H2);clear H1;intros.
  assert (forall y : OT, seq_range finite y -> ordinal y).
  destruct 1.
  subst y;auto with schutte.
  generalize (H1 H3);clear H1;intros.
  case (H1 a H H0).
  destruct 1.
  case H5.
  intros (_,H6);subst a.
  case H4.
  intros.
  exists x0;auto.
  case H4;intros.
  subst x.
  case (finite_lt_inv x0 H7);auto.
  intros x (H8,H9);exists x;auto.
Qed.


Lemma ordinal_omega_limit : forall s, (forall i, s i < s (S i)) ->
                                      ordinal (omega_limit s).
Proof.
  unfold omega_limit.
  intros;apply sup_ordinal.
  apply seq_range_denumerable;auto.
  red; apply seq_mono_intro; auto.
  destruct 1.
  subst x;auto with schutte.
  eauto with schutte.
Qed.


Lemma ordinal_omega : ordinal omega.
  unfold omega; apply ordinal_omega_limit.  
  auto with schutte arith.
Qed.


Lemma lt_omega_limit_lt_exists_lt :  forall alpha s, 
                                       (forall i, s i < s (S i)) ->
                                       alpha < omega_limit s ->
                                       exists j, alpha < s j.
Proof.
  intros.
  unfold omega_limit in H0.
  assert (exists y, In (seq_range s) y /\ alpha <  y).
  apply lt_sup_exists_lt; auto with schutte.
  red;red.
  destruct 1;auto.
  subst x;eauto with schutte.
  eauto with schutte.
  eauto with schutte.
  case H1; intros y Hy.
  case Hy;destruct 1.
  subst y;exists x;auto.
Qed.


Hint Resolve ordinal_omega ordinal_omega_limit : schutte.

Lemma is_limit_omega : is_limit omega.

Proof.
  unfold is_limit.
  repeat split; auto with schutte.
  red;intro.
  absurd (F 1 < zero).
  apply not_lt_zero;auto with schutte.
  rewrite <- H.
  apply finite_lt_omega.
  red;intro.
  case H.
  intros x (Ox,Ex).
  assert (x < omega).
  rewrite Ex;auto with schutte.
  case (lt_omega_finite    H0).
  intros; subst x.
  assert (omega = F (S x0)).
  simpl;auto.
  case ( @lt_irr omega);auto with schutte.
  pattern omega at 1;rewrite H1;auto with schutte.
  apply finite_lt_omega;auto with schutte.
Qed.



(* about zero, is_succ and is_limit *)

Lemma not_is_succ_zero : ~ (is_succ zero).
Proof.
  red. intros (b, (Hb,Hb')).  
  case (succ_zero_diff  Hb).  
  auto.
Qed.


Lemma not_is_limit_zero : ~ (is_limit zero).
Proof.
  unfold is_limit.
  tauto.
Qed.

Lemma not_is_limit_succ : forall a, ordinal a -> is_succ a -> ~ (is_limit a).
Proof.
  unfold is_limit;tauto.
Qed.

Lemma not_is_succ_limit : forall a, ordinal a -> is_limit a -> ~(is_succ a).
Proof.
  unfold is_limit;tauto.
Qed.

Lemma is_limit_ordinal : forall a, is_limit a -> ordinal a.
Proof.
  unfold is_limit;tauto.
Qed.

Lemma is_succ_ordinal : forall a, is_succ a -> ordinal a.
Proof.
  unfold is_succ.
  intros a (b,(Hb,H'b)).
  subst a; auto with schutte.
Qed.

Hint Resolve is_limit_ordinal is_succ_ordinal : schutte.

Lemma denumerable_members : forall a, ordinal a -> denumerable (members a).
Proof.
  intros a Ha.
  apply AX2.
  unfold members.
  apply Intersection_decreases_l.
  exists a; split ; auto.
  destruct 1. auto. 
Qed.

Hint Resolve denumerable_members : schutte.

Lemma le_sup_members : forall a, ordinal a -> |_| (members a) <= a.
Proof.
  intros; apply sup_least_upper_bound; auto with schutte.
  destruct 1; auto. 
  destruct 1;auto. 
  right; auto. 
Qed.



Lemma is_limit_sup_members : forall a, is_limit a -> a = |_| (members a).
Proof.
  intros.
  case (@le_sup_members a);auto with schutte.
  intuition.

  intro; generalize (@sup_members_succ a).

  intuition.

  generalize (H1 (is_limit_ordinal H) H0).
  intro.
  case (not_is_limit_succ (a:=a));auto with schutte.
  exists (|_|members a).
  split;auto.
  apply sup_ordinal;auto with schutte.
  destruct 1;auto.
Qed.




Lemma sup_members_disj : forall a, ordinal a -> a = sup (members a) ->
                                   a = zero \/ is_limit a.
Proof.
  intros.
  unfold is_limit.
  case (classic (a=zero)).
  auto.
  right;repeat split;auto.
  red;intro H2.
  case H2;intros x (H3,H4).
  case (sup_members_not_succ  H H3 H0).
  auto.
Qed.


Lemma ordinal_sup_members : forall a, ordinal a ->
                                      ordinal (sup (members a)).
Proof.
  intros; apply sup_ordinal; auto with schutte.
  destruct 1;auto.
Qed.

Theorem classification : forall a, ordinal a -> a = zero \/
                                                is_succ a \/ 
                                                is_limit a.
Proof.
  intros a Ha.
  case (le_sup_members  Ha).
  intro;case (sup_members_disj  Ha);auto with schutte.
  case H;auto.
  right.
  left.
  exists (sup (members a)).
  split.
  apply ordinal_sup_members;auto.
  apply sup_members_succ;auto.
Qed.





(* Segments *)

(* extensional equalities on sets of ordinals *)







Lemma members_omega :  Same_set (members omega) is_finite.
Proof.
  unfold Same_set;split.
  red.  unfold members, In.
  
  destruct 1.
  
  case (lt_omega_finite  H0).
  intros;exists x0;auto.
  unfold  Included, members.
  
  destruct 1.
  subst x;split;red;auto with schutte.
  apply finite_lt_omega.
Qed.


(* remplacer par un ET *)






Lemma Not_Unbounded_bounded : forall X, Included X ordinal ->
                                        not (Unbounded X) ->
                                        exists beta,
                                          ordinal beta /\
                                          forall x, In X x -> x < beta.
Proof.
  intros.
  apply NNPP.
  red;intro.
  assert (forall beta, ordinal beta ->
                       exists x, In X x /\ beta <= x).
 
  intro.
  intros.
  change (exists x:OT, (fun y => In X y /\ beta <= y) x).
  apply not_all_not_ex.
  red;intro.
  case H1.
  exists beta; split;auto.
  intros.
  tricho x beta H5;auto with schutte.
  apply H;auto.
  
  case (H3 x).
  split;auto with schutte.
  case (H3 x);split;auto with schutte.
 
  {  clear H1.
  case H0.
  red;intro.
  intros.
  case (H2 (succ x)).
  auto with schutte.
  intros.
  exists x0;split;auto with schutte.
  case H3;auto.
  case H3;intros.
  apply lt_le_trans with (succ x).
  auto with schutte.
  
  auto.
  }
Qed.




Lemma Not_Unbounded_denumerable : forall X, Included X ordinal ->
                                            not (Unbounded X) -> denumerable X. 
  intros;eapply AX2;eauto.
  case (Not_Unbounded_bounded H);auto.
  intros.
  exists x;auto.
Qed.

Lemma not_denumerable_unbounded : forall X, Included X ordinal ->
                                            not (denumerable X) -> Unbounded X.
Proof.
  intros.
  apply NNPP.
  red;intro.
  apply H0;apply Not_Unbounded_denumerable;auto.
Qed.


Lemma le_alpha_zero : forall alpha, alpha <= zero -> alpha = zero.
Proof.
  intros.
  apply le_antisym;auto.
  apply le_zero.
  eauto with schutte.
Qed.


 
 
 
 







 



 
 
 
 








 
 


 
 
 
 










 
 
 

 
 
 
 
