(** Polymorphic versions of exponentiation functions *)

Require Import Arith ZArith MulTree.

(** 
 Let us define an exponentaition function in a context that 
 declares a type A, a binary function mult over A, and a neutral
 element one

*)
Section Definitions.

 Variables (A : Type)
           (mult : A -> A -> A)
           (one : A).

 Local Infix "*" := mult.
Local Notation "1" := one.

Fixpoint power (x:A)(n:nat) : A :=
  match n with 0%nat => 1
             | S p =>   x * x ^ p
  end
where "x ^ n" := (power x n).



Fixpoint binary_power_mult (x a:A)(p:positive) : A 
  :=
  match p with
    | xH =>  a * x
    | xO q => binary_power_mult  (x * x) a q
    | xI q =>  binary_power_mult  (x * x) (a * x) q
  end.

Fixpoint Pos_bpow   (x:A)(p:positive) :=
 match p with
  | xH => x
  | xO q => Pos_bpow  (x * x) q
  | xI q => binary_power_mult   (x * x) x q
end.


(** ***
  It is straightforward to adapt [Pos_bpow]
 for accepting exponents of type [N] :
*)

Definition N_bpow  x (n:N) := 
  match n with 
  | 0%N => 1
  | Npos p => Pos_bpow x p
  end.



 (** *** 

The  following function decomposes the exponent [p]
  into $2 ^ k \times q$ , then calls [binary_power_mult]
   with $x^{2^k}$ and $q$.

*)

End Definitions.

Arguments N_bpow  {A} mult one x n.
Arguments power  {A} mult one x n.

Compute power Z.mul 1%Z 2%Z 10.

Compute N_bpow Z.mul 1%Z 2%Z 10.

Require Import String.
Open Scope string_scope.

Compute N_bpow append  "" "ab"  12.


Module M2.
Section M2_Definitions.
  
  Variables (A: Type)
           (zero one : A) 
           (plus mult  : A -> A -> A).
  
  Variable rt : semi_ring_theory  zero one plus mult   (@eq A).
  Add Ring Aring : rt.

  Notation "0" := zero.  
  Notation "1" := one.
  Notation "x + y" := (plus x y).  
  Notation "x * y " := (mult x y).
  
  Structure t : Type := mat{c00 : A;  c01 : A;
                            c10 : A;  c11 : A}.
  
  Definition Id2 : t := mat 1 0 0 1.
  
  Definition M2_mult (M M':t) : t :=
    mat (c00 M * c00 M' + c01 M * c10 M')
        (c00 M * c01 M' + c01 M * c11 M')
        (c10 M * c00 M' + c11 M * c10 M')
        (c10 M * c01 M' + c11 M * c11 M').

End M2_Definitions.
End M2.

Import M2.

Arguments M2_mult {A} plus mult  _  _.
Arguments mat {A} _ _ _ _.
Arguments Id2 {A}  _ _.

Definition fibonacci (n:N) :=
 c00 N  (N_bpow  (M2_mult Nplus Nmult) (Id2  0%N 1%N)(mat  1 1 1 0)%N n).

 Compute fibonacci 20.



Definition power_t := forall (A:Type)
                             (mult : A -> A -> A)
                             (one:A)
                             (x:A)
                             (n:N), A.


Module Wrong.

Definition correct_expt (f : power_t) : Prop :=
  forall A (mult : A -> A -> A) (one:A)
         (x:A) (n:N), power mult one x (N.to_nat n) = f A mult one x n.


Lemma correct_exp_too_strong : ~ correct_expt (@N_bpow).
Proof.
  intro H; specialize (H (MulTree N)  Mul_node (One_node) (A_node 2%N) 7%N);
  discriminate.    
Qed.

End Wrong.
