(** ** Some useful instances of Monoid classes 

File %\href{../Powers/SRC/Monoid_instances.v}{Powers/SRC/Monoid\_instances}
defines various instances of [Monoid] and [EMonoid].
*)

(* begin hide *)

Require Export Monoid_def.
Require Import RelationClasses Morphisms.

Require Import ZArith PArith.
Open Scope Z_scope.

(* end hide *)

(** *** Multiplicative monoid on [nat] *)
Require Import Arith.

Instance nat_mult_op : Binop nat | 5 := Nat.mul.

Instance  Natmult : Monoid nat_mult_op  1%nat | 5.
Proof.
   split;unfold nat_mult_op, bin_op; intros; ring.
Qed.


(** *** Additive monoid on [nat] 

The following monoid is useful for proving the correctness of complex
exponentiation algorithms. In effect, the $n$-th "power" of $1$ is
equal to $n$. See Sect.%~\ref{chains-exponent}.
*)

Instance nat_plus_op : Binop nat | 12 := Nat.add.

Instance Natplus : Monoid nat_plus_op  0%nat | 12.
Proof.
   split;unfold nat_plus_op, bin_op; intros; ring.
Qed.

(** *** Multiplicative monoid on [Z] *)

Instance Z_mult_op : Binop Z := Z.mul.

Instance ZMult : Monoid  Z_mult_op 1.
Proof. 
  split;unfold Z_mult_op, bin_op;intros;ring.
Qed.

Instance ZMult_Abelian : Abelian_Monoid ZMult.
Proof.
  split; exact Zmult_comm.
Qed.


(* end  hide *)

(** Multiplicative Monoid on [positive] 
*)

Instance P_mult_op : Binop positive | 5  := Pos.mul .
Instance P_plus_op : Binop positive | 12 := Pos.add.

Instance PMult : Monoid P_mult_op xH | 5.
(* begin hide *)
Proof.
 split;unfold P_mult_op, Binop;intros.  
 -  now rewrite Pos.mul_assoc.
 - reflexivity.
 - now rewrite Pos.mul_1_r.
Qed.


Require Import NArith.
Open Scope N_scope.
(* end hide *)


Instance N_mult_op  : Binop N | 5 := N.mul.

Instance NMult : Monoid N_mult_op 1 | 5.
Proof.
  split;unfold N_mult_op, bin_op; intros; ring.
Qed.


Instance N_plus_op  : Binop N | 12 := N.add.

Instance NPlus : Monoid N_plus_op 0 | 12.
Proof.
  split;unfold N_plus_op, bin_op; intros; ring.
Qed.




(* begin hide *)
Require Import Ring31.
Import Int31.
Open Scope int31_scope.
(* end hide *)


(** ***  Multiplicative monoid on 31-bits integers 

Cyclic numeric types are  good candidates for testing exponentiations
with big exponents, since the size of data is bounded.


The type [int31] is defined in Standard Library in Module
[Coq.Numbers.Cyclic.Int31.Int31].
*)

Instance int31_mult_op : Binop int31 := mul31.

Instance  Int31mult : Monoid int31_mult_op  1.
Proof.
   split;unfold int31_mult_op, bin_op; intros; ring.
Qed.

(* begin hide *)
Close Scope int31_scope.
(* end hide *)

(** *** Integers modulo m

The following instance of [EMonoid] describes the set of integers modulo
$m$, where $m$ is some integer greater or equal than $2$.
For simplicity's sake, we represent such values using the type [N],
and consider "equivalence modulo $m$" instead of equality.
*)


Section Nmodulo.
  Variable m : N.
  Hypothesis m_gt_1 : 1 < m.
  
  (* begin hide *)
  
  Remark m_neq_0 : m <> 0.
    intro H;subst m. discriminate. 
  Qed.
  
  Hint Resolve m_neq_0.
  
  (* end hide *)
  
  Definition mult_mod ( x y : N) := (x * y) mod m.
  Definition mod_eq ( x y: N) := x mod m = y mod m.
  
  Global Instance mod_equiv : Equiv N := mod_eq.

  Global Instance mod_op : Binop N := mult_mod.
  Global Instance mod_Equiv : Equivalence mod_equiv.
  (* begin hide *)
  Proof.
    split; unfold mod_equiv, equiv.
    - intros x; now red.
    - intros x y H; now symmetry.  
    - intros x y z Hxy Hyz; transitivity (y mod m) ; auto.
  Qed.
  (* end hide *)
  
  Local  Open Scope M_scope.
  Global Instance mult_mod_proper : Proper (mod_equiv ==> mod_equiv ==> mod_equiv) 
                                           mod_op.
  (* begin hide *)
  Proof.
    unfold mod_equiv, mod_op, bin_op, equiv.
    intros x y Hxy z t Hzt.
    unfold mod_eq, mult_mod in *; unfold mod_eq, mult_mod.
    repeat rewrite N.mod_mod; auto.
    rewrite (N.mul_mod x z);auto.
    rewrite (N.mul_mod y t);auto.
    rewrite Hxy, Hzt;auto.
  Qed.
  (* end hide *)
  


  Lemma mult_mod_associative :  forall x y z,
                                  x * (y * z) = x * y * z.
  (* begin hide *)
  Proof.
    intros  x y z.
    unfold mod_op, bin_op, mult_mod.
    rewrite N.mul_mod_idemp_r;auto.
    rewrite N.mul_mod_idemp_l;auto.
    now rewrite  N.mul_assoc.
  Qed.
  (* end hide *)
  
  Lemma one_mod_neutral_l  : forall x, 1 * x == x.
  (* begin hide *)
  Proof.
     unfold mod_equiv, equiv, mod_eq, mod_op, bin_op, mult_mod.
    intro x;    rewrite N.mul_1_l.
    rewrite N.mod_mod;auto.
  Qed.
  (* end hide *)
  
  Lemma one_mod_neutral_r  : forall x, x * 1 == x.
  (* begin hide *)
  Proof.
    unfold mod_equiv, equiv, mod_eq, mod_op, bin_op, mult_mod.
    intro x;     rewrite N.mul_1_r.
    rewrite N.mod_mod;auto.
  Qed.
  (* end hide *)
  
  Global Instance Nmod_Monoid : EMonoid  mod_op 1 mod_equiv.
  (* begin hide *)
  Proof.
    unfold mod_equiv, equiv, mod_eq, mod_op, bin_op, mult_mod.
    split.
    - apply mod_Equiv.
    -  apply mult_mod_proper.
    - intros;rewrite mult_mod_associative.
      reflexivity.
    - intro;rewrite one_mod_neutral_l; reflexivity.
    - intro;rewrite one_mod_neutral_r; reflexivity.
  Defined.
(* end hide *)
End Nmodulo.

Section S256.

Let mod256 :=  mod_op 256.

Local Existing Instance mod256 | 1.

Compute (211 * 67)%M.

(** = 57 : N *)
End S256.

Compute (211 * 67)%M.
(** = 14137 : N *)



(** *** Monoid of 2x2 matrices 

Let $A$ be some type, provided with a ring structure. We define the multiplication 
of $2\times 2$-matrices, the coefficients of which have type $A$.

*)

Section M2_def.
Variables (A:Type)
           (zero one : A) 
           (plus mult  : A -> A -> A).
 Notation "0" := zero.  
 Notation "1" := one.
 Notation "x + y" := (plus x y).  
 Notation "x * y " := (mult x y).

 Variable rt : semi_ring_theory  zero one plus mult  (@eq A).
 Add  Ring Aring : rt.


Structure M2 : Type := {c00 : A;  c01 : A;
                        c10 : A;  c11 : A}.

Definition Id2 : M2 := Build_M2 1 0 0 1.

Definition M2_mult (m m':M2) : M2 :=
 Build_M2 (c00 m * c00 m' + c01 m * c10 m')
          (c00 m * c01 m' + c01 m * c11 m')
          (c10 m * c00 m' + c11 m * c10 m')
          (c10 m * c01 m' + c11 m * c11 m').


(* begin hide *)

Lemma M2_eq_intros : forall a b c d a' b' c' d',
  a=a' -> b=b' -> c=c' -> d=d' ->
   Build_M2 a b c d = Build_M2 a' b' c' d'.
Proof. 
 intros; now f_equal.
Qed.

(* end hide *)

Global Instance M2_op : Binop M2 := M2_mult.

Global Instance M2_Monoid : Monoid   M2_op Id2.
(* begin hide *)
Proof. 
 unfold M2_op, bin_op; split.
 - destruct x;destruct y;destruct z;simpl.
   unfold M2_mult; apply M2_eq_intros; simpl; ring.
 - destruct x;simpl;
   unfold M2_mult; apply M2_eq_intros; simpl; ring. 
 - destruct x;simpl;
   unfold M2_mult;apply M2_eq_intros;simpl;ring. 
Qed.
(* end hide *)
End M2_def.

Arguments M2_Monoid {A zero one plus mult} rt.
Arguments Build_M2 {A} _ _ _ _.

(** Matrices over N *)
Definition M2N := M2_Monoid Nth.

(* begin hide *)
Close Scope N_scope.
Close Scope positive_scope.
(* end hide *)

Check NMult : EMonoid  N.mul 1%N eq.

