(** *  Exponentiation Schemes 

Pierre Cast\'eran
*)


Require Import Chains.
Import Monoid_instances Pow Omega List .
Generalizable All Variables.
Open Scope M_scope.

(** ** Star chains

   Star chains are chains in which if $x^i$ is computed as the
product $x^j \times x^k$, then $x^j$ is the previous computed power 
of $x$. Thus the execution of a star-chain uses a "register" x
that is implicit in the coding of a computation. 

  Consequently, the arguments of  a multiplication 
  are + the implicit register
      + the "other argument"
      + a continuation that depends on whether the register will be
        used in further computation (see function star_eval)

  
*)



Inductive Star_computation {A:Type}  : Type :=
| star_return 
| Mult_by (y : A) (k1 :  Star_computation )
| Square (k1 :  Star_computation )
| Save (k1 : A -> Star_computation ).


Fixpoint star_inclusion {A:Type}(a:A)(c : Star_computation) :
   Computation :=
match c with
 star_return => Return a
| Mult_by y k1 => Mult a y (fun z => star_inclusion z k1)
| Square k1 => Mult a a (fun z => star_inclusion z k1)
| Save k1 =>  star_inclusion a (k1 a)
end. 

Definition star_chain := forall A:Type, A ->  @Star_computation A.

Notation "'SQR'  ';' e2 " := (Square  e2)
                               (right associativity, at level 60).
Notation "'R*=' y ';' e2 " :=
        (Mult_by y  e2) (right associativity, at level 60).

Notation "y  '<---R' ';' e2 " := (Save  (fun y =>   e2))
                               (right associativity, at level 60).

Example Star7 : star_chain :=
  fun A (x:A) => 
    SQR ; R*= x ; SQR ;  R*= x; star_return .




Example Star10 : star_chain :=
  fun A (x:A) =>
  SQR ;  R*= x; x3 <---R ; SQR ;  R*= x ; R*= x3 ; star_return.


Compute   fun A a => star_inclusion a (Star10 A a).

Fixpoint star_computation_execute  `{M:@EMonoid A E_op E_one E_eq}  (m : Star_computation) (x:A) : A :=
match m with 
| star_return  => x
| Square k1 => star_computation_execute k1 ( x * x)
| Mult_by  y k1 => star_computation_execute k1 (x * y)
| Save  k1 => star_computation_execute (k1 x) x
end.


Definition star_chain_execute (c:star_chain)`{M:@EMonoid A E_op E_one E_eq} (a:A) :=
  star_computation_execute (c A a)  a.

Compute star_chain_execute Star7 2%Z.

Compute star_chain_execute Star10 2%Z.

Definition star_chain_for_87 : star_chain :=
 fun A (x : A)=>
  SQR  (** x^2 *) ;
  R*= x ; x3 <---R ;
  SQR  (** x^6 *) ; 
  R*= x; x7 <---R ;
  R*= x3 (** x^10 *); 
  SQR ; SQR ; SQR  (** x ^ 80 *); 
  R*= x7 (** x ^87 *) ;
  star_return. 


Compute  Pos.eqb (star_chain_execute star_chain_for_87  2%positive)
                 (chain_apply C87 2%positive).


(** Generated by Marc Lasson's plug-in *)

Inductive
Star_computation_R (A1 A2 : Type) (A_R : A1 -> A2 -> Type)
  : Star_computation -> Star_computation -> Type :=
    star_return_R : Star_computation_R A1 A2 A_R star_return star_return
  | Mult_by_R : forall (y1 : A1) (y2 : A2),
                A_R y1 y2 ->
                forall k1 k2 : Star_computation,
                Star_computation_R A1 A2 A_R k1 k2 ->
                Star_computation_R A1 A2 A_R (R*= y1; k1) (R*= y2; k2)
  | Square_R : forall k1 k2 : Star_computation,
               Star_computation_R A1 A2 A_R k1 k2 ->
               Star_computation_R A1 A2 A_R (SQR ; k1) (SQR ; k2)
  | Save_R : forall (k1 : A1 -> Star_computation)
               (k2 : A2 -> Star_computation),
             (forall (a1 : A1) (a2 : A2),
              A_R a1 a2 -> Star_computation_R A1 A2 A_R (k1 a1) (k2 a2)) ->
             Star_computation_R A1 A2 A_R (y <---R ; k1 y) (y <---R ; k2 y).








(** ** Computation schemes as data 
*)




